#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import json


MOCKED_EXCHANGE_SYMBOL_DETAILS = {
    "bybit": [{'id': 'BLZUSDT', 'symbol': 'BLZ/USDT:USDT', 'base': 'BLZ', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'BLZ', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 1.0, 'price': 1e-05}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 1.0, 'max': 98500.0}, 'price': {'min': 1e-05, 'max': 19.99998}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'BLZUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'BLZ', 'quoteCoin': 'USDT', 'launchTime': '1654497673000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '5', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.00001', 'maxPrice': '19.99998', 'tickSize': '0.00001'}, 'lotSizeFilter': {'maxOrderQty': '98500', 'minOrderQty': '1', 'qtyStep': '1', 'postOnlyMaxOrderQty': '985000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'BNBPERP', 'symbol': 'BNB/USD:USDC', 'base': 'BNB', 'quote': 'USD', 'settle': 'USDC', 'baseId': 'BNB', 'quoteId': 'USD', 'settleId': 'USDC', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.01, 'price': 0.05}, 'limits': {'leverage': {'min': 1.0, 'max': 50.0}, 'amount': {'min': 0.01, 'max': 250.0}, 'price': {'min': 0.05, 'max': 99999.9}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'BNBPERP', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'BNB', 'quoteCoin': 'USD', 'launchTime': '1664459242000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '2', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '50.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.05', 'maxPrice': '99999.90', 'tickSize': '0.05'}, 'lotSizeFilter': {'maxOrderQty': '250.00', 'minOrderQty': '0.01', 'qtyStep': '0.01', 'postOnlyMaxOrderQty': '1250.00'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDC', 'copyTrading': 'none'}}, {'id': 'BNBUSDT', 'symbol': 'BNB/USDT:USDT', 'base': 'BNB', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'BNB', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.01, 'price': 0.05}, 'limits': {'leverage': {'min': 1.0, 'max': 50.0}, 'amount': {'min': 0.01, 'max': 1300.0}, 'price': {'min': 0.05, 'max': 99999.9}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'BNBUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'BNB', 'quoteCoin': 'USDT', 'launchTime': '0', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '2', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '50.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.05', 'maxPrice': '99999.90', 'tickSize': '0.05'}, 'lotSizeFilter': {'maxOrderQty': '1300.00', 'minOrderQty': '0.01', 'qtyStep': '0.01', 'postOnlyMaxOrderQty': '13000.00'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'normalOnly'}}, {'id': 'BNXUSDT', 'symbol': 'BNX/USDT:USDT', 'base': 'BNX', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'BNX', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.1, 'price': 0.0005}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 0.1, 'max': 35000.0}, 'price': {'min': 0.0005, 'max': 999.999}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'BNXUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'BNX', 'quoteCoin': 'USDT', 'launchTime': '1677564318000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '4', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.0005', 'maxPrice': '999.9990', 'tickSize': '0.0005'}, 'lotSizeFilter': {'maxOrderQty': '35000.0', 'minOrderQty': '0.1', 'qtyStep': '0.1', 'postOnlyMaxOrderQty': '175000.0'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'BOBAUSDT', 'symbol': 'BOBA/USDT:USDT', 'base': 'BOBA', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'BOBA', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.1, 'price': 5e-05}, 'limits': {'leverage': {'min': 1.0, 'max': 12.5}, 'amount': {'min': 0.1, 'max': 9900.0}, 'price': {'min': 5e-05, 'max': 99.9999}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'BOBAUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'BOBA', 'quoteCoin': 'USDT', 'launchTime': '1649231761000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '5', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '12.50', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.00005', 'maxPrice': '99.99990', 'tickSize': '0.00005'}, 'lotSizeFilter': {'maxOrderQty': '9900.0', 'minOrderQty': '0.1', 'qtyStep': '0.1', 'postOnlyMaxOrderQty': '99000.0'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'BSVUSDT', 'symbol': 'BSV/USDT:USDT', 'base': 'BSV', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'BSV', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.01, 'price': 0.05}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 0.01, 'max': 180.0}, 'price': {'min': 0.05, 'max': 99999.9}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'BSVUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'BSV', 'quoteCoin': 'USDT', 'launchTime': '1639105708000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '2', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.05', 'maxPrice': '99999.90', 'tickSize': '0.05'}, 'lotSizeFilter': {'maxOrderQty': '180.00', 'minOrderQty': '0.01', 'qtyStep': '0.01', 'postOnlyMaxOrderQty': '1800.00'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'BSWUSDT', 'symbol': 'BSW/USDT:USDT', 'base': 'BSW', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'BSW', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 1.0, 'price': 0.0001}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 1.0, 'max': 19500.0}, 'price': {'min': 0.0001, 'max': 199.9998}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'BSWUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'BSW', 'quoteCoin': 'USDT', 'launchTime': '1648017126000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '4', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.0001', 'maxPrice': '199.9998', 'tickSize': '0.0001'}, 'lotSizeFilter': {'maxOrderQty': '19500', 'minOrderQty': '1', 'qtyStep': '1', 'postOnlyMaxOrderQty': '195000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'BTC-04AUG23', 'symbol': 'BTC/USD:USDC-230804', 'base': 'BTC', 'quote': 'USD', 'settle': 'USDC', 'baseId': 'BTC', 'quoteId': 'USD', 'settleId': 'USDC', 'type': 'future', 'spot': False, 'margin': None, 'swap': False, 'future': True, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': 1691136000000, 'expiryDatetime': '2023-08-04T08:00:00.000Z', 'strike': None, 'optionType': None, 'precision': {'amount': 0.001, 'price': 0.5}, 'limits': {'leverage': {'min': 1.0, 'max': 50.0}, 'amount': {'min': 0.001, 'max': 10.0}, 'price': {'min': 0.5, 'max': 999999.0}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'BTC-04AUG23', 'contractType': 'LinearFutures', 'status': 'Trading', 'baseCoin': 'BTC', 'quoteCoin': 'USD', 'launchTime': '1689321600000', 'deliveryTime': '1691136000000', 'deliveryFeeRate': '0', 'priceScale': '2', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '50.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.50', 'maxPrice': '999999.00', 'tickSize': '0.50'}, 'lotSizeFilter': {'maxOrderQty': '10.000', 'minOrderQty': '0.001', 'qtyStep': '0.001', 'postOnlyMaxOrderQty': '50.000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDC', 'copyTrading': 'none'}}, {'id': 'BTC-11AUG23', 'symbol': 'BTC/USD:USDC-230811', 'base': 'BTC', 'quote': 'USD', 'settle': 'USDC', 'baseId': 'BTC', 'quoteId': 'USD', 'settleId': 'USDC', 'type': 'future', 'spot': False, 'margin': None, 'swap': False, 'future': True, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': 1691740800000, 'expiryDatetime': '2023-08-11T08:00:00.000Z', 'strike': None, 'optionType': None, 'precision': {'amount': 0.001, 'price': 0.5}, 'limits': {'leverage': {'min': 1.0, 'max': 50.0}, 'amount': {'min': 0.001, 'max': 10.0}, 'price': {'min': 0.5, 'max': 999999.0}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'BTC-11AUG23', 'contractType': 'LinearFutures', 'status': 'Trading', 'baseCoin': 'BTC', 'quoteCoin': 'USD', 'launchTime': '1689926400000', 'deliveryTime': '1691740800000', 'deliveryFeeRate': '0', 'priceScale': '2', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '50.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.50', 'maxPrice': '999999.00', 'tickSize': '0.50'}, 'lotSizeFilter': {'maxOrderQty': '10.000', 'minOrderQty': '0.001', 'qtyStep': '0.001', 'postOnlyMaxOrderQty': '50.000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDC', 'copyTrading': 'none'}}, {'id': 'BTC-25AUG23', 'symbol': 'BTC/USD:USDC-230825', 'base': 'BTC', 'quote': 'USD', 'settle': 'USDC', 'baseId': 'BTC', 'quoteId': 'USD', 'settleId': 'USDC', 'type': 'future', 'spot': False, 'margin': None, 'swap': False, 'future': True, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': 1692950400000, 'expiryDatetime': '2023-08-25T08:00:00.000Z', 'strike': None, 'optionType': None, 'precision': {'amount': 0.001, 'price': 0.5}, 'limits': {'leverage': {'min': 1.0, 'max': 50.0}, 'amount': {'min': 0.001, 'max': 10.0}, 'price': {'min': 0.5, 'max': 999999.0}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'BTC-25AUG23', 'contractType': 'LinearFutures', 'status': 'Trading', 'baseCoin': 'BTC', 'quoteCoin': 'USD', 'launchTime': '1686297600000', 'deliveryTime': '1692950400000', 'deliveryFeeRate': '0', 'priceScale': '2', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '50.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.50', 'maxPrice': '999999.00', 'tickSize': '0.50'}, 'lotSizeFilter': {'maxOrderQty': '10.000', 'minOrderQty': '0.001', 'qtyStep': '0.001', 'postOnlyMaxOrderQty': '50.000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDC', 'copyTrading': 'none'}}, {'id': 'BTC-28JUL23', 'symbol': 'BTC/USD:USDC-230728', 'base': 'BTC', 'quote': 'USD', 'settle': 'USDC', 'baseId': 'BTC', 'quoteId': 'USD', 'settleId': 'USDC', 'type': 'future', 'spot': False, 'margin': None, 'swap': False, 'future': True, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': 1690531200000, 'expiryDatetime': '2023-07-28T08:00:00.000Z', 'strike': None, 'optionType': None, 'precision': {'amount': 0.001, 'price': 0.5}, 'limits': {'leverage': {'min': 1.0, 'max': 50.0}, 'amount': {'min': 0.001, 'max': 10.0}, 'price': {'min': 0.5, 'max': 999999.0}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'BTC-28JUL23', 'contractType': 'LinearFutures', 'status': 'Trading', 'baseCoin': 'BTC', 'quoteCoin': 'USD', 'launchTime': '1683273600000', 'deliveryTime': '1690531200000', 'deliveryFeeRate': '0', 'priceScale': '2', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '50.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.50', 'maxPrice': '999999.00', 'tickSize': '0.50'}, 'lotSizeFilter': {'maxOrderQty': '10.000', 'minOrderQty': '0.001', 'qtyStep': '0.001', 'postOnlyMaxOrderQty': '50.000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDC', 'copyTrading': 'none'}}, {'id': 'BTC-29DEC23', 'symbol': 'BTC/USD:USDC-231229', 'base': 'BTC', 'quote': 'USD', 'settle': 'USDC', 'baseId': 'BTC', 'quoteId': 'USD', 'settleId': 'USDC', 'type': 'future', 'spot': False, 'margin': None, 'swap': False, 'future': True, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': 1703836800000, 'expiryDatetime': '2023-12-29T08:00:00.000Z', 'strike': None, 'optionType': None, 'precision': {'amount': 0.001, 'price': 0.5}, 'limits': {'leverage': {'min': 1.0, 'max': 50.0}, 'amount': {'min': 0.001, 'max': 10.0}, 'price': {'min': 0.5, 'max': 999999.0}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'BTC-29DEC23', 'contractType': 'LinearFutures', 'status': 'Trading', 'baseCoin': 'BTC', 'quoteCoin': 'USD', 'launchTime': '1680854400000', 'deliveryTime': '1703836800000', 'deliveryFeeRate': '0', 'priceScale': '2', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '50.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.50', 'maxPrice': '999999.00', 'tickSize': '0.50'}, 'lotSizeFilter': {'maxOrderQty': '10.000', 'minOrderQty': '0.001', 'qtyStep': '0.001', 'postOnlyMaxOrderQty': '50.000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDC', 'copyTrading': 'none'}}, {'id': 'BTC-29MAR24', 'symbol': 'BTC/USD:USDC-240329', 'base': 'BTC', 'quote': 'USD', 'settle': 'USDC', 'baseId': 'BTC', 'quoteId': 'USD', 'settleId': 'USDC', 'type': 'future', 'spot': False, 'margin': None, 'swap': False, 'future': True, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': 1711699200000, 'expiryDatetime': '2024-03-29T08:00:00.000Z', 'strike': None, 'optionType': None, 'precision': {'amount': 0.001, 'price': 0.5}, 'limits': {'leverage': {'min': 1.0, 'max': 50.0}, 'amount': {'min': 0.001, 'max': 10.0}, 'price': {'min': 0.5, 'max': 999999.0}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'BTC-29MAR24', 'contractType': 'LinearFutures', 'status': 'Trading', 'baseCoin': 'BTC', 'quoteCoin': 'USD', 'launchTime': '1688716800000', 'deliveryTime': '1711699200000', 'deliveryFeeRate': '0', 'priceScale': '2', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '50.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.50', 'maxPrice': '999999.00', 'tickSize': '0.50'}, 'lotSizeFilter': {'maxOrderQty': '10.000', 'minOrderQty': '0.001', 'qtyStep': '0.001', 'postOnlyMaxOrderQty': '50.000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDC', 'copyTrading': 'none'}}, {'id': 'BTC-29SEP23', 'symbol': 'BTC/USD:USDC-230929', 'base': 'BTC', 'quote': 'USD', 'settle': 'USDC', 'baseId': 'BTC', 'quoteId': 'USD', 'settleId': 'USDC', 'type': 'future', 'spot': False, 'margin': None, 'swap': False, 'future': True, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': 1695974400000, 'expiryDatetime': '2023-09-29T08:00:00.000Z', 'strike': None, 'optionType': None, 'precision': {'amount': 0.001, 'price': 0.5}, 'limits': {'leverage': {'min': 1.0, 'max': 50.0}, 'amount': {'min': 0.001, 'max': 10.0}, 'price': {'min': 0.5, 'max': 999999.0}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'BTC-29SEP23', 'contractType': 'LinearFutures', 'status': 'Trading', 'baseCoin': 'BTC', 'quoteCoin': 'USD', 'launchTime': '1679644800000', 'deliveryTime': '1695974400000', 'deliveryFeeRate': '0', 'priceScale': '2', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '50.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.50', 'maxPrice': '999999.00', 'tickSize': '0.50'}, 'lotSizeFilter': {'maxOrderQty': '10.000', 'minOrderQty': '0.001', 'qtyStep': '0.001', 'postOnlyMaxOrderQty': '50.000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDC', 'copyTrading': 'none'}}, {'id': 'BTCPERP', 'symbol': 'BTC/USD:USDC', 'base': 'BTC', 'quote': 'USD', 'settle': 'USDC', 'baseId': 'BTC', 'quoteId': 'USD', 'settleId': 'USDC', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.001, 'price': 0.1}, 'limits': {'leverage': {'min': 1.0, 'max': 125.0}, 'amount': {'min': 0.001, 'max': 20.0}, 'price': {'min': 0.1, 'max': 199999.8}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'BTCPERP', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'BTC', 'quoteCoin': 'USD', 'launchTime': '1610604231000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '2', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '125.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.10', 'maxPrice': '199999.80', 'tickSize': '0.10'}, 'lotSizeFilter': {'maxOrderQty': '20.000', 'minOrderQty': '0.001', 'qtyStep': '0.001', 'postOnlyMaxOrderQty': '100.000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDC', 'copyTrading': 'none'}}, {'id': 'BTCUSDT', 'symbol': 'BTC/USDT:USDT', 'base': 'BTC', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'BTC', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.001, 'price': 0.1}, 'limits': {'leverage': {'min': 1.0, 'max': 100.0}, 'amount': {'min': 0.001, 'max': 100.0}, 'price': {'min': 0.1, 'max': 199999.8}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'BTCUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'BTC', 'quoteCoin': 'USDT', 'launchTime': '1584230400000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '2', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '100.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.10', 'maxPrice': '199999.80', 'tickSize': '0.10'}, 'lotSizeFilter': {'maxOrderQty': '100.000', 'minOrderQty': '0.001', 'qtyStep': '0.001', 'postOnlyMaxOrderQty': '1000.000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'normalOnly'}}, {'id': 'BUSDUSDT', 'symbol': 'BUSD/USDT:USDT', 'base': 'BUSD', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'BUSD', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.1, 'price': 0.0001}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 0.1, 'max': 100000.0}, 'price': {'min': 0.0001, 'max': 199.9998}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'BUSDUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'BUSD', 'quoteCoin': 'USDT', 'launchTime': '1676340857000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '4', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.0001', 'maxPrice': '199.9998', 'tickSize': '0.0001'}, 'lotSizeFilter': {'maxOrderQty': '100000.0', 'minOrderQty': '0.1', 'qtyStep': '0.1', 'postOnlyMaxOrderQty': '500000.0'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'C98USDT', 'symbol': 'C98/USDT:USDT', 'base': 'C98', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'C98', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.1, 'price': 0.0005}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 0.1, 'max': 31200.0}, 'price': {'min': 0.0005, 'max': 999.999}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'C98USDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'C98', 'quoteCoin': 'USDT', 'launchTime': '1635120000000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '4', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.0005', 'maxPrice': '999.9990', 'tickSize': '0.0005'}, 'lotSizeFilter': {'maxOrderQty': '31200.0', 'minOrderQty': '0.1', 'qtyStep': '0.1', 'postOnlyMaxOrderQty': '312000.0'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'CEEKUSDT', 'symbol': 'CEEK/USDT:USDT', 'base': 'CEEK', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'CEEK', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 1.0, 'price': 5e-05}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 1.0, 'max': 18200.0}, 'price': {'min': 5e-05, 'max': 99.9999}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'CEEKUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'CEEK', 'quoteCoin': 'USDT', 'launchTime': '1658127698000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '5', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.00005', 'maxPrice': '99.99990', 'tickSize': '0.00005'}, 'lotSizeFilter': {'maxOrderQty': '18200', 'minOrderQty': '1', 'qtyStep': '1', 'postOnlyMaxOrderQty': '182000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'CELOUSDT', 'symbol': 'CELO/USDT:USDT', 'base': 'CELO', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'CELO', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.1, 'price': 0.001}, 'limits': {'leverage': {'min': 1.0, 'max': 12.5}, 'amount': {'min': 0.1, 'max': 17400.0}, 'price': {'min': 0.001, 'max': 1999.998}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'CELOUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'CELO', 'quoteCoin': 'USDT', 'launchTime': '1641367582000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '3', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '12.50', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.001', 'maxPrice': '1999.998', 'tickSize': '0.001'}, 'lotSizeFilter': {'maxOrderQty': '17400.0', 'minOrderQty': '0.1', 'qtyStep': '0.1', 'postOnlyMaxOrderQty': '174000.0'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'CELRUSDT', 'symbol': 'CELR/USDT:USDT', 'base': 'CELR', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'CELR', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 1.0, 'price': 5e-05}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 1.0, 'max': 775600.0}, 'price': {'min': 5e-05, 'max': 99.9999}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'CELRUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'CELR', 'quoteCoin': 'USDT', 'launchTime': '1635120000000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '5', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.00005', 'maxPrice': '99.99990', 'tickSize': '0.00005'}, 'lotSizeFilter': {'maxOrderQty': '775600', 'minOrderQty': '1', 'qtyStep': '1', 'postOnlyMaxOrderQty': '7756000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'CFXUSDT', 'symbol': 'CFX/USDT:USDT', 'base': 'CFX', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'CFX', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 1.0, 'price': 1e-05}, 'limits': {'leverage': {'min': 1.0, 'max': 50.0}, 'amount': {'min': 1.0, 'max': 500000.0}, 'price': {'min': 1e-05, 'max': 19.99998}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'CFXUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'CFX', 'quoteCoin': 'USDT', 'launchTime': '1676883925000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '5', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '50.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.00001', 'maxPrice': '19.99998', 'tickSize': '0.00001'}, 'lotSizeFilter': {'maxOrderQty': '500000', 'minOrderQty': '1', 'qtyStep': '1', 'postOnlyMaxOrderQty': '2500000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'normalOnly'}}, {'id': 'CHRUSDT', 'symbol': 'CHR/USDT:USDT', 'base': 'CHR', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'CHR', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.1, 'price': 0.0001}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 0.1, 'max': 50200.0}, 'price': {'min': 0.0001, 'max': 199.9998}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'CHRUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'CHR', 'quoteCoin': 'USDT', 'launchTime': '1637193600000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '4', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.0001', 'maxPrice': '199.9998', 'tickSize': '0.0001'}, 'lotSizeFilter': {'maxOrderQty': '50200.0', 'minOrderQty': '0.1', 'qtyStep': '0.1', 'postOnlyMaxOrderQty': '502000.0'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'CHZUSDT', 'symbol': 'CHZ/USDT:USDT', 'base': 'CHZ', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'CHZ', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 1.0, 'price': 5e-05}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 1.0, 'max': 700000.0}, 'price': {'min': 5e-05, 'max': 99.9999}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'CHZUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'CHZ', 'quoteCoin': 'USDT', 'launchTime': '1633996800000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '5', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.00005', 'maxPrice': '99.99990', 'tickSize': '0.00005'}, 'lotSizeFilter': {'maxOrderQty': '700000', 'minOrderQty': '1', 'qtyStep': '1', 'postOnlyMaxOrderQty': '7000000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'normalOnly'}}, {'id': 'CKBUSDT', 'symbol': 'CKB/USDT:USDT', 'base': 'CKB', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'CKB', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 10.0, 'price': 5e-06}, 'limits': {'leverage': {'min': 1.0, 'max': 12.5}, 'amount': {'min': 10.0, 'max': 740000.0}, 'price': {'min': 5e-06, 'max': 9.99999}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'CKBUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'CKB', 'quoteCoin': 'USDT', 'launchTime': '1644901148000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '6', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '12.50', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.000005', 'maxPrice': '9.999990', 'tickSize': '0.000005'}, 'lotSizeFilter': {'maxOrderQty': '740000', 'minOrderQty': '10', 'qtyStep': '10', 'postOnlyMaxOrderQty': '7400000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'COMBOUSDT', 'symbol': 'COMBO/USDT:USDT', 'base': 'COMBO', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'COMBO', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.1, 'price': 0.0005}, 'limits': {'leverage': {'min': 1.0, 'max': 12.5}, 'amount': {'min': 0.1, 'max': 18000.0}, 'price': {'min': 0.0005, 'max': 999.999}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'COMBOUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'COMBO', 'quoteCoin': 'USDT', 'launchTime': '1686210792000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '4', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '12.50', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.0005', 'maxPrice': '999.9990', 'tickSize': '0.0005'}, 'lotSizeFilter': {'maxOrderQty': '18000.0', 'minOrderQty': '0.1', 'qtyStep': '0.1', 'postOnlyMaxOrderQty': '90000.0'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'COMPUSDT', 'symbol': 'COMP/USDT:USDT', 'base': 'COMP', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'COMP', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.01, 'price': 0.05}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 0.01, 'max': 640.0}, 'price': {'min': 0.05, 'max': 99999.9}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'COMPUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'COMP', 'quoteCoin': 'USDT', 'launchTime': '1628726400000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '2', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.05', 'maxPrice': '99999.90', 'tickSize': '0.05'}, 'lotSizeFilter': {'maxOrderQty': '640.00', 'minOrderQty': '0.01', 'qtyStep': '0.01', 'postOnlyMaxOrderQty': '6400.00'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'COREUSDT', 'symbol': 'CORE/USDT:USDT', 'base': 'CORE', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'CORE', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.1, 'price': 0.0005}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 0.1, 'max': 10000.0}, 'price': {'min': 0.0005, 'max': 999.999}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'COREUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'CORE', 'quoteCoin': 'USDT', 'launchTime': '1675927514000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '4', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.0005', 'maxPrice': '999.9990', 'tickSize': '0.0005'}, 'lotSizeFilter': {'maxOrderQty': '10000.0', 'minOrderQty': '0.1', 'qtyStep': '0.1', 'postOnlyMaxOrderQty': '50000.0'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'COTIUSDT', 'symbol': 'COTI/USDT:USDT', 'base': 'COTI', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'COTI', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 1.0, 'price': 1e-05}, 'limits': {'leverage': {'min': 1.0, 'max': 16.67}, 'amount': {'min': 1.0, 'max': 182000.0}, 'price': {'min': 1e-05, 'max': 19.99998}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'COTIUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'COTI', 'quoteCoin': 'USDT', 'launchTime': '1638445117000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '5', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '16.67', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.00001', 'maxPrice': '19.99998', 'tickSize': '0.00001'}, 'lotSizeFilter': {'maxOrderQty': '182000', 'minOrderQty': '1', 'qtyStep': '1', 'postOnlyMaxOrderQty': '1820000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'CROUSDT', 'symbol': 'CRO/USDT:USDT', 'base': 'CRO', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'CRO', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 1.0, 'price': 5e-05}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 1.0, 'max': 90000.0}, 'price': {'min': 5e-05, 'max': 99.9999}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'CROUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'CRO', 'quoteCoin': 'USDT', 'launchTime': '1637625600000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '5', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.00005', 'maxPrice': '99.99990', 'tickSize': '0.00005'}, 'lotSizeFilter': {'maxOrderQty': '90000', 'minOrderQty': '1', 'qtyStep': '1', 'postOnlyMaxOrderQty': '900000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'CRVUSDT', 'symbol': 'CRV/USDT:USDT', 'base': 'CRV', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'CRV', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.1, 'price': 0.001}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 0.1, 'max': 70000.0}, 'price': {'min': 0.001, 'max': 1999.998}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'CRVUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'CRV', 'quoteCoin': 'USDT', 'launchTime': '1635120000000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '3', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.001', 'maxPrice': '1999.998', 'tickSize': '0.001'}, 'lotSizeFilter': {'maxOrderQty': '70000.0', 'minOrderQty': '0.1', 'qtyStep': '0.1', 'postOnlyMaxOrderQty': '700000.0'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'normalOnly'}}, {'id': 'CTCUSDT', 'symbol': 'CTC/USDT:USDT', 'base': 'CTC', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'CTC', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 1.0, 'price': 5e-05}, 'limits': {'leverage': {'min': 1.0, 'max': 12.5}, 'amount': {'min': 1.0, 'max': 3500.0}, 'price': {'min': 5e-05, 'max': 99.9999}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'CTCUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'CTC', 'quoteCoin': 'USDT', 'launchTime': '1652774270000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '5', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '12.50', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.00005', 'maxPrice': '99.99990', 'tickSize': '0.00005'}, 'lotSizeFilter': {'maxOrderQty': '3500', 'minOrderQty': '1', 'qtyStep': '1', 'postOnlyMaxOrderQty': '35000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'CTKUSDT', 'symbol': 'CTK/USDT:USDT', 'base': 'CTK', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'CTK', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.1, 'price': 0.0001}, 'limits': {'leverage': {'min': 1.0, 'max': 16.67}, 'amount': {'min': 0.1, 'max': 17200.0}, 'price': {'min': 0.0001, 'max': 199.9998}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'CTKUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'CTK', 'quoteCoin': 'USDT', 'launchTime': '1638444847000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '4', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '16.67', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.0001', 'maxPrice': '199.9998', 'tickSize': '0.0001'}, 'lotSizeFilter': {'maxOrderQty': '17200.0', 'minOrderQty': '0.1', 'qtyStep': '0.1', 'postOnlyMaxOrderQty': '172000.0'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'CTSIUSDT', 'symbol': 'CTSI/USDT:USDT', 'base': 'CTSI', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'CTSI', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 1.0, 'price': 5e-05}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 1.0, 'max': 156000.0}, 'price': {'min': 5e-05, 'max': 99.9999}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'CTSIUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'CTSI', 'quoteCoin': 'USDT', 'launchTime': '1648018511000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '5', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.00005', 'maxPrice': '99.99990', 'tickSize': '0.00005'}, 'lotSizeFilter': {'maxOrderQty': '156000', 'minOrderQty': '1', 'qtyStep': '1', 'postOnlyMaxOrderQty': '1560000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'CVCUSDT', 'symbol': 'CVC/USDT:USDT', 'base': 'CVC', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'CVC', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 1.0, 'price': 1e-05}, 'limits': {'leverage': {'min': 1.0, 'max': 12.5}, 'amount': {'min': 1.0, 'max': 91800.0}, 'price': {'min': 1e-05, 'max': 19.99998}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'CVCUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'CVC', 'quoteCoin': 'USDT', 'launchTime': '1638285651000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '5', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '12.50', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.00001', 'maxPrice': '19.99998', 'tickSize': '0.00001'}, 'lotSizeFilter': {'maxOrderQty': '91800', 'minOrderQty': '1', 'qtyStep': '1', 'postOnlyMaxOrderQty': '918000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'CVXUSDT', 'symbol': 'CVX/USDT:USDT', 'base': 'CVX', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'CVX', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.01, 'price': 0.005}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 0.01, 'max': 1200.0}, 'price': {'min': 0.005, 'max': 9999.99}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'CVXUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'CVX', 'quoteCoin': 'USDT', 'launchTime': '1650440282000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '3', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.005', 'maxPrice': '9999.990', 'tickSize': '0.005'}, 'lotSizeFilter': {'maxOrderQty': '1200.00', 'minOrderQty': '0.01', 'qtyStep': '0.01', 'postOnlyMaxOrderQty': '12000.00'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'DARUSDT', 'symbol': 'DAR/USDT:USDT', 'base': 'DAR', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'DAR', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.1, 'price': 0.0005}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 0.1, 'max': 27525.0}, 'price': {'min': 0.0005, 'max': 999.999}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'DARUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'DAR', 'quoteCoin': 'USDT', 'launchTime': '1651467113000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '4', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.0005', 'maxPrice': '999.9990', 'tickSize': '0.0005'}, 'lotSizeFilter': {'maxOrderQty': '27525.0', 'minOrderQty': '0.1', 'qtyStep': '0.1', 'postOnlyMaxOrderQty': '275250.0'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'DASHUSDT', 'symbol': 'DASH/USDT:USDT', 'base': 'DASH', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'DASH', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.01, 'price': 0.05}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 0.01, 'max': 410.0}, 'price': {'min': 0.05, 'max': 99999.9}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'DASHUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'DASH', 'quoteCoin': 'USDT', 'launchTime': '1633996800000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '2', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.05', 'maxPrice': '99999.90', 'tickSize': '0.05'}, 'lotSizeFilter': {'maxOrderQty': '410.00', 'minOrderQty': '0.01', 'qtyStep': '0.01', 'postOnlyMaxOrderQty': '4100.00'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'DENTUSDT', 'symbol': 'DENT/USDT:USDT', 'base': 'DENT', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'DENT', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 100.0, 'price': 1e-06}, 'limits': {'leverage': {'min': 1.0, 'max': 12.5}, 'amount': {'min': 100.0, 'max': 9800000.0}, 'price': {'min': 1e-06, 'max': 1.999998}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'DENTUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'DENT', 'quoteCoin': 'USDT', 'launchTime': '1642068928000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '6', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '12.50', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.000001', 'maxPrice': '1.999998', 'tickSize': '0.000001'}, 'lotSizeFilter': {'maxOrderQty': '9800000', 'minOrderQty': '100', 'qtyStep': '100', 'postOnlyMaxOrderQty': '98000000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'DGBUSDT', 'symbol': 'DGB/USDT:USDT', 'base': 'DGB', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'DGB', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 10.0, 'price': 1e-05}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 10.0, 'max': 351400.0}, 'price': {'min': 1e-05, 'max': 19.99998}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'DGBUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'DGB', 'quoteCoin': 'USDT', 'launchTime': '1648452331000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '5', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.00001', 'maxPrice': '19.99998', 'tickSize': '0.00001'}, 'lotSizeFilter': {'maxOrderQty': '351400', 'minOrderQty': '10', 'qtyStep': '10', 'postOnlyMaxOrderQty': '3514000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'DODOUSDT', 'symbol': 'DODO/USDT:USDT', 'base': 'DODO', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'DODO', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 1.0, 'price': 5e-05}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 1.0, 'max': 46000.0}, 'price': {'min': 5e-05, 'max': 99.9999}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'DODOUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'DODO', 'quoteCoin': 'USDT', 'launchTime': '1650457800000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '5', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.00005', 'maxPrice': '99.99990', 'tickSize': '0.00005'}, 'lotSizeFilter': {'maxOrderQty': '46000', 'minOrderQty': '1', 'qtyStep': '1', 'postOnlyMaxOrderQty': '460000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'DOGEUSDT', 'symbol': 'DOGE/USDT:USDT', 'base': 'DOGE', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'DOGE', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 1.0, 'price': 1e-05}, 'limits': {'leverage': {'min': 1.0, 'max': 75.0}, 'amount': {'min': 1.0, 'max': 10000000.0}, 'price': {'min': 1e-05, 'max': 19.99998}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'DOGEUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'DOGE', 'quoteCoin': 'USDT', 'launchTime': '0', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '5', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '75.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.00001', 'maxPrice': '19.99998', 'tickSize': '0.00001'}, 'lotSizeFilter': {'maxOrderQty': '10000000', 'minOrderQty': '1', 'qtyStep': '1', 'postOnlyMaxOrderQty': '100000000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'normalOnly'}}, {'id': 'DOTUSDT', 'symbol': 'DOT/USDT:USDT', 'base': 'DOT', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'DOT', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.1, 'price': 0.001}, 'limits': {'leverage': {'min': 1.0, 'max': 50.0}, 'amount': {'min': 0.1, 'max': 32900.0}, 'price': {'min': 0.001, 'max': 1999.998}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'DOTUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'DOT', 'quoteCoin': 'USDT', 'launchTime': '0', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '3', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '50.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.001', 'maxPrice': '1999.998', 'tickSize': '0.001'}, 'lotSizeFilter': {'maxOrderQty': '32900.0', 'minOrderQty': '0.1', 'qtyStep': '0.1', 'postOnlyMaxOrderQty': '329000.0'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'normalOnly'}}, {'id': 'DUSKUSDT', 'symbol': 'DUSK/USDT:USDT', 'base': 'DUSK', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'DUSK', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 1.0, 'price': 1e-05}, 'limits': {'leverage': {'min': 1.0, 'max': 12.5}, 'amount': {'min': 1.0, 'max': 52500.0}, 'price': {'min': 1e-05, 'max': 19.99998}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'DUSKUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'DUSK', 'quoteCoin': 'USDT', 'launchTime': '1639673130000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '5', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '12.50', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.00001', 'maxPrice': '19.99998', 'tickSize': '0.00001'}, 'lotSizeFilter': {'maxOrderQty': '52500', 'minOrderQty': '1', 'qtyStep': '1', 'postOnlyMaxOrderQty': '525000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'DYDXUSDT', 'symbol': 'DYDX/USDT:USDT', 'base': 'DYDX', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'DYDX', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.1, 'price': 0.001}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 0.1, 'max': 45000.0}, 'price': {'min': 0.001, 'max': 1999.998}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'DYDXUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'DYDX', 'quoteCoin': 'USDT', 'launchTime': '1633910400000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '3', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.001', 'maxPrice': '1999.998', 'tickSize': '0.001'}, 'lotSizeFilter': {'maxOrderQty': '45000.0', 'minOrderQty': '0.1', 'qtyStep': '0.1', 'postOnlyMaxOrderQty': '450000.0'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'normalOnly'}}, {'id': 'EDUUSDT', 'symbol': 'EDU/USDT:USDT', 'base': 'EDU', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'EDU', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.1, 'price': 0.0001}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 0.1, 'max': 20000.0}, 'price': {'min': 0.0001, 'max': 199.9998}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'EDUUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'EDU', 'quoteCoin': 'USDT', 'launchTime': '1683273334000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '4', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.0001', 'maxPrice': '199.9998', 'tickSize': '0.0001'}, 'lotSizeFilter': {'maxOrderQty': '20000.0', 'minOrderQty': '0.1', 'qtyStep': '0.1', 'postOnlyMaxOrderQty': '100000.0'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'normalOnly'}}, {'id': 'EGLDUSDT', 'symbol': 'EGLD/USDT:USDT', 'base': 'EGLD', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'EGLD', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.01, 'price': 0.05}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 0.01, 'max': 1300.0}, 'price': {'min': 0.05, 'max': 99999.9}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'EGLDUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'EGLD', 'quoteCoin': 'USDT', 'launchTime': '1637596800000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '2', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.05', 'maxPrice': '99999.90', 'tickSize': '0.05'}, 'lotSizeFilter': {'maxOrderQty': '1300.00', 'minOrderQty': '0.01', 'qtyStep': '0.01', 'postOnlyMaxOrderQty': '13000.00'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'ENJUSDT', 'symbol': 'ENJ/USDT:USDT', 'base': 'ENJ', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'ENJ', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.1, 'price': 0.0005}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 0.1, 'max': 116000.0}, 'price': {'min': 0.0005, 'max': 999.999}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'ENJUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'ENJ', 'quoteCoin': 'USDT', 'launchTime': '1635696000000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '4', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.0005', 'maxPrice': '999.9990', 'tickSize': '0.0005'}, 'lotSizeFilter': {'maxOrderQty': '116000.0', 'minOrderQty': '0.1', 'qtyStep': '0.1', 'postOnlyMaxOrderQty': '1160000.0'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'ENSUSDT', 'symbol': 'ENS/USDT:USDT', 'base': 'ENS', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'ENS', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.1, 'price': 0.01}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 0.1, 'max': 1600.0}, 'price': {'min': 0.01, 'max': 19999.98}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'ENSUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'ENS', 'quoteCoin': 'USDT', 'launchTime': '1636675200000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '2', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.01', 'maxPrice': '19999.98', 'tickSize': '0.01'}, 'lotSizeFilter': {'maxOrderQty': '1600.0', 'minOrderQty': '0.1', 'qtyStep': '0.1', 'postOnlyMaxOrderQty': '16000.0'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'normalOnly'}}, {'id': 'EOSUSDT', 'symbol': 'EOS/USDT:USDT', 'base': 'EOS', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'EOS', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.1, 'price': 0.001}, 'limits': {'leverage': {'min': 1.0, 'max': 50.0}, 'amount': {'min': 0.1, 'max': 55000.0}, 'price': {'min': 0.001, 'max': 1999.998}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'EOSUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'EOS', 'quoteCoin': 'USDT', 'launchTime': '0', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '3', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '50.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.001', 'maxPrice': '1999.998', 'tickSize': '0.001'}, 'lotSizeFilter': {'maxOrderQty': '55000.0', 'minOrderQty': '0.1', 'qtyStep': '0.1', 'postOnlyMaxOrderQty': '550000.0'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'normalOnly'}}, {'id': 'ETCPERP', 'symbol': 'ETC/USD:USDC', 'base': 'ETC', 'quote': 'USD', 'settle': 'USDC', 'baseId': 'ETC', 'quoteId': 'USD', 'settleId': 'USDC', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.1, 'price': 0.005}, 'limits': {'leverage': {'min': 1.0, 'max': 50.0}, 'amount': {'min': 0.1, 'max': 500.0}, 'price': {'min': 0.005, 'max': 9999.99}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'ETCPERP', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'ETC', 'quoteCoin': 'USD', 'launchTime': '1661990400000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '3', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '50.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.005', 'maxPrice': '9999.990', 'tickSize': '0.005'}, 'lotSizeFilter': {'maxOrderQty': '500.0', 'minOrderQty': '0.1', 'qtyStep': '0.1', 'postOnlyMaxOrderQty': '2500.0'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDC', 'copyTrading': 'none'}}, {'id': 'ETCUSDT', 'symbol': 'ETC/USDT:USDT', 'base': 'ETC', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'ETC', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.1, 'price': 0.005}, 'limits': {'leverage': {'min': 1.0, 'max': 50.0}, 'amount': {'min': 0.1, 'max': 10000.0}, 'price': {'min': 0.005, 'max': 9999.99}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'ETCUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'ETC', 'quoteCoin': 'USDT', 'launchTime': '0', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '3', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '50.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.005', 'maxPrice': '9999.990', 'tickSize': '0.005'}, 'lotSizeFilter': {'maxOrderQty': '10000.0', 'minOrderQty': '0.1', 'qtyStep': '0.1', 'postOnlyMaxOrderQty': '100000.0'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'normalOnly'}}, {'id': 'ETH-04AUG23', 'symbol': 'ETH/USD:USDC-230804', 'base': 'ETH', 'quote': 'USD', 'settle': 'USDC', 'baseId': 'ETH', 'quoteId': 'USD', 'settleId': 'USDC', 'type': 'future', 'spot': False, 'margin': None, 'swap': False, 'future': True, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': 1691136000000, 'expiryDatetime': '2023-08-04T08:00:00.000Z', 'strike': None, 'optionType': None, 'precision': {'amount': 0.01, 'price': 0.05}, 'limits': {'leverage': {'min': 1.0, 'max': 50.0}, 'amount': {'min': 0.01, 'max': 150.0}, 'price': {'min': 0.05, 'max': 99999.9}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'ETH-04AUG23', 'contractType': 'LinearFutures', 'status': 'Trading', 'baseCoin': 'ETH', 'quoteCoin': 'USD', 'launchTime': '1689321600000', 'deliveryTime': '1691136000000', 'deliveryFeeRate': '0', 'priceScale': '2', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '50.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.05', 'maxPrice': '99999.90', 'tickSize': '0.05'}, 'lotSizeFilter': {'maxOrderQty': '150.00', 'minOrderQty': '0.01', 'qtyStep': '0.01', 'postOnlyMaxOrderQty': '750.00'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDC', 'copyTrading': 'none'}}, {'id': 'ETH-11AUG23', 'symbol': 'ETH/USD:USDC-230811', 'base': 'ETH', 'quote': 'USD', 'settle': 'USDC', 'baseId': 'ETH', 'quoteId': 'USD', 'settleId': 'USDC', 'type': 'future', 'spot': False, 'margin': None, 'swap': False, 'future': True, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': 1691740800000, 'expiryDatetime': '2023-08-11T08:00:00.000Z', 'strike': None, 'optionType': None, 'precision': {'amount': 0.01, 'price': 0.05}, 'limits': {'leverage': {'min': 1.0, 'max': 50.0}, 'amount': {'min': 0.01, 'max': 150.0}, 'price': {'min': 0.05, 'max': 99999.9}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'ETH-11AUG23', 'contractType': 'LinearFutures', 'status': 'Trading', 'baseCoin': 'ETH', 'quoteCoin': 'USD', 'launchTime': '1689926400000', 'deliveryTime': '1691740800000', 'deliveryFeeRate': '0', 'priceScale': '2', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '50.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.05', 'maxPrice': '99999.90', 'tickSize': '0.05'}, 'lotSizeFilter': {'maxOrderQty': '150.00', 'minOrderQty': '0.01', 'qtyStep': '0.01', 'postOnlyMaxOrderQty': '750.00'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDC', 'copyTrading': 'none'}}, {'id': 'ETH-25AUG23', 'symbol': 'ETH/USD:USDC-230825', 'base': 'ETH', 'quote': 'USD', 'settle': 'USDC', 'baseId': 'ETH', 'quoteId': 'USD', 'settleId': 'USDC', 'type': 'future', 'spot': False, 'margin': None, 'swap': False, 'future': True, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': 1692950400000, 'expiryDatetime': '2023-08-25T08:00:00.000Z', 'strike': None, 'optionType': None, 'precision': {'amount': 0.01, 'price': 0.05}, 'limits': {'leverage': {'min': 1.0, 'max': 50.0}, 'amount': {'min': 0.01, 'max': 150.0}, 'price': {'min': 0.05, 'max': 99999.9}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'ETH-25AUG23', 'contractType': 'LinearFutures', 'status': 'Trading', 'baseCoin': 'ETH', 'quoteCoin': 'USD', 'launchTime': '1686297600000', 'deliveryTime': '1692950400000', 'deliveryFeeRate': '0', 'priceScale': '2', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '50.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.05', 'maxPrice': '99999.90', 'tickSize': '0.05'}, 'lotSizeFilter': {'maxOrderQty': '150.00', 'minOrderQty': '0.01', 'qtyStep': '0.01', 'postOnlyMaxOrderQty': '750.00'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDC', 'copyTrading': 'none'}}, {'id': 'ETH-28JUL23', 'symbol': 'ETH/USD:USDC-230728', 'base': 'ETH', 'quote': 'USD', 'settle': 'USDC', 'baseId': 'ETH', 'quoteId': 'USD', 'settleId': 'USDC', 'type': 'future', 'spot': False, 'margin': None, 'swap': False, 'future': True, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': 1690531200000, 'expiryDatetime': '2023-07-28T08:00:00.000Z', 'strike': None, 'optionType': None, 'precision': {'amount': 0.01, 'price': 0.05}, 'limits': {'leverage': {'min': 1.0, 'max': 50.0}, 'amount': {'min': 0.01, 'max': 150.0}, 'price': {'min': 0.05, 'max': 99999.9}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'ETH-28JUL23', 'contractType': 'LinearFutures', 'status': 'Trading', 'baseCoin': 'ETH', 'quoteCoin': 'USD', 'launchTime': '1683273600000', 'deliveryTime': '1690531200000', 'deliveryFeeRate': '0', 'priceScale': '2', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '50.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.05', 'maxPrice': '99999.90', 'tickSize': '0.05'}, 'lotSizeFilter': {'maxOrderQty': '150.00', 'minOrderQty': '0.01', 'qtyStep': '0.01', 'postOnlyMaxOrderQty': '750.00'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDC', 'copyTrading': 'none'}}, {'id': 'ETH-29DEC23', 'symbol': 'ETH/USD:USDC-231229', 'base': 'ETH', 'quote': 'USD', 'settle': 'USDC', 'baseId': 'ETH', 'quoteId': 'USD', 'settleId': 'USDC', 'type': 'future', 'spot': False, 'margin': None, 'swap': False, 'future': True, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': 1703836800000, 'expiryDatetime': '2023-12-29T08:00:00.000Z', 'strike': None, 'optionType': None, 'precision': {'amount': 0.01, 'price': 0.05}, 'limits': {'leverage': {'min': 1.0, 'max': 50.0}, 'amount': {'min': 0.01, 'max': 150.0}, 'price': {'min': 0.05, 'max': 99999.9}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'ETH-29DEC23', 'contractType': 'LinearFutures', 'status': 'Trading', 'baseCoin': 'ETH', 'quoteCoin': 'USD', 'launchTime': '1680854400000', 'deliveryTime': '1703836800000', 'deliveryFeeRate': '0', 'priceScale': '2', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '50.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.05', 'maxPrice': '99999.90', 'tickSize': '0.05'}, 'lotSizeFilter': {'maxOrderQty': '150.00', 'minOrderQty': '0.01', 'qtyStep': '0.01', 'postOnlyMaxOrderQty': '750.00'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDC', 'copyTrading': 'none'}}, {'id': 'ETH-29MAR24', 'symbol': 'ETH/USD:USDC-240329', 'base': 'ETH', 'quote': 'USD', 'settle': 'USDC', 'baseId': 'ETH', 'quoteId': 'USD', 'settleId': 'USDC', 'type': 'future', 'spot': False, 'margin': None, 'swap': False, 'future': True, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': 1711699200000, 'expiryDatetime': '2024-03-29T08:00:00.000Z', 'strike': None, 'optionType': None, 'precision': {'amount': 0.01, 'price': 0.05}, 'limits': {'leverage': {'min': 1.0, 'max': 50.0}, 'amount': {'min': 0.01, 'max': 150.0}, 'price': {'min': 0.05, 'max': 99999.9}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'ETH-29MAR24', 'contractType': 'LinearFutures', 'status': 'Trading', 'baseCoin': 'ETH', 'quoteCoin': 'USD', 'launchTime': '1688716800000', 'deliveryTime': '1711699200000', 'deliveryFeeRate': '0', 'priceScale': '2', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '50.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.05', 'maxPrice': '99999.90', 'tickSize': '0.05'}, 'lotSizeFilter': {'maxOrderQty': '150.00', 'minOrderQty': '0.01', 'qtyStep': '0.01', 'postOnlyMaxOrderQty': '750.00'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDC', 'copyTrading': 'none'}}, {'id': 'ETH-29SEP23', 'symbol': 'ETH/USD:USDC-230929', 'base': 'ETH', 'quote': 'USD', 'settle': 'USDC', 'baseId': 'ETH', 'quoteId': 'USD', 'settleId': 'USDC', 'type': 'future', 'spot': False, 'margin': None, 'swap': False, 'future': True, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': 1695974400000, 'expiryDatetime': '2023-09-29T08:00:00.000Z', 'strike': None, 'optionType': None, 'precision': {'amount': 0.01, 'price': 0.05}, 'limits': {'leverage': {'min': 1.0, 'max': 50.0}, 'amount': {'min': 0.01, 'max': 150.0}, 'price': {'min': 0.05, 'max': 99999.9}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'ETH-29SEP23', 'contractType': 'LinearFutures', 'status': 'Trading', 'baseCoin': 'ETH', 'quoteCoin': 'USD', 'launchTime': '1679644800000', 'deliveryTime': '1695974400000', 'deliveryFeeRate': '0', 'priceScale': '2', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '50.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.05', 'maxPrice': '99999.90', 'tickSize': '0.05'}, 'lotSizeFilter': {'maxOrderQty': '150.00', 'minOrderQty': '0.01', 'qtyStep': '0.01', 'postOnlyMaxOrderQty': '750.00'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDC', 'copyTrading': 'none'}}, {'id': 'ETHPERP', 'symbol': 'ETH/USD:USDC', 'base': 'ETH', 'quote': 'USD', 'settle': 'USDC', 'baseId': 'ETH', 'quoteId': 'USD', 'settleId': 'USDC', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.01, 'price': 0.01}, 'limits': {'leverage': {'min': 1.0, 'max': 66.67}, 'amount': {'min': 0.01, 'max': 400.0}, 'price': {'min': 0.01, 'max': 19999.98}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'ETHPERP', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'ETH', 'quoteCoin': 'USD', 'launchTime': '1659657600000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '2', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '66.67', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.01', 'maxPrice': '19999.98', 'tickSize': '0.01'}, 'lotSizeFilter': {'maxOrderQty': '400.00', 'minOrderQty': '0.01', 'qtyStep': '0.01', 'postOnlyMaxOrderQty': '2000.00'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDC', 'copyTrading': 'none'}}, {'id': 'ETHUSDT', 'symbol': 'ETH/USDT:USDT', 'base': 'ETH', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'ETH', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.01, 'price': 0.01}, 'limits': {'leverage': {'min': 1.0, 'max': 100.0}, 'amount': {'min': 0.01, 'max': 1500.0}, 'price': {'min': 0.01, 'max': 19999.98}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'ETHUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'ETH', 'quoteCoin': 'USDT', 'launchTime': '1615766400000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '2', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '100.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.01', 'maxPrice': '19999.98', 'tickSize': '0.01'}, 'lotSizeFilter': {'maxOrderQty': '1500.00', 'minOrderQty': '0.01', 'qtyStep': '0.01', 'postOnlyMaxOrderQty': '15000.00'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'normalOnly'}}, {'id': 'ETHWUSDT', 'symbol': 'ETHW/USDT:USDT', 'base': 'ETHW', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'ETHW', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.01, 'price': 0.001}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 0.01, 'max': 7500.0}, 'price': {'min': 0.001, 'max': 1999.998}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'ETHWUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'ETHW', 'quoteCoin': 'USDT', 'launchTime': '1663286400000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '3', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.001', 'maxPrice': '1999.998', 'tickSize': '0.001'}, 'lotSizeFilter': {'maxOrderQty': '7500.00', 'minOrderQty': '0.01', 'qtyStep': '0.01', 'postOnlyMaxOrderQty': '75000.00'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'normalOnly'}}, {'id': 'FETUSDT', 'symbol': 'FET/USDT:USDT', 'base': 'FET', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'FET', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 1.0, 'price': 5e-05}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 1.0, 'max': 250000.0}, 'price': {'min': 5e-05, 'max': 99.9999}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'FETUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'FET', 'quoteCoin': 'USDT', 'launchTime': '1675759851000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '5', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.00005', 'maxPrice': '99.99990', 'tickSize': '0.00005'}, 'lotSizeFilter': {'maxOrderQty': '250000', 'minOrderQty': '1', 'qtyStep': '1', 'postOnlyMaxOrderQty': '1250000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'FILUSDT', 'symbol': 'FIL/USDT:USDT', 'base': 'FIL', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'FIL', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.1, 'price': 0.001}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 0.1, 'max': 45000.0}, 'price': {'min': 0.001, 'max': 1999.998}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'FILUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'FIL', 'quoteCoin': 'USDT', 'launchTime': '0', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '3', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.001', 'maxPrice': '1999.998', 'tickSize': '0.001'}, 'lotSizeFilter': {'maxOrderQty': '45000.0', 'minOrderQty': '0.1', 'qtyStep': '0.1', 'postOnlyMaxOrderQty': '450000.0'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'normalOnly'}}, {'id': 'FITFIUSDT', 'symbol': 'FITFI/USDT:USDT', 'base': 'FITFI', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'FITFI', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 1.0, 'price': 5e-06}, 'limits': {'leverage': {'min': 1.0, 'max': 50.0}, 'amount': {'min': 1.0, 'max': 200000.0}, 'price': {'min': 5e-06, 'max': 9.99999}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'FITFIUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'FITFI', 'quoteCoin': 'USDT', 'launchTime': '1652153037000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '6', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '50.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.000005', 'maxPrice': '9.999990', 'tickSize': '0.000005'}, 'lotSizeFilter': {'maxOrderQty': '200000', 'minOrderQty': '1', 'qtyStep': '1', 'postOnlyMaxOrderQty': '2000000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'FLMUSDT', 'symbol': 'FLM/USDT:USDT', 'base': 'FLM', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'FLM', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 1.0, 'price': 5e-05}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 1.0, 'max': 72500.0}, 'price': {'min': 5e-05, 'max': 99.9999}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'FLMUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'FLM', 'quoteCoin': 'USDT', 'launchTime': '1651467612000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '5', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.00005', 'maxPrice': '99.99990', 'tickSize': '0.00005'}, 'lotSizeFilter': {'maxOrderQty': '72500', 'minOrderQty': '1', 'qtyStep': '1', 'postOnlyMaxOrderQty': '725000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'FLOWUSDT', 'symbol': 'FLOW/USDT:USDT', 'base': 'FLOW', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'FLOW', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.1, 'price': 0.0005}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 0.1, 'max': 23000.0}, 'price': {'min': 0.0005, 'max': 999.999}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'FLOWUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'FLOW', 'quoteCoin': 'USDT', 'launchTime': '1637856000000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '4', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.0005', 'maxPrice': '999.9990', 'tickSize': '0.0005'}, 'lotSizeFilter': {'maxOrderQty': '23000.0', 'minOrderQty': '0.1', 'qtyStep': '0.1', 'postOnlyMaxOrderQty': '230000.0'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'FLRUSDT', 'symbol': 'FLR/USDT:USDT', 'base': 'FLR', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'FLR', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 10.0, 'price': 1e-05}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 10.0, 'max': 379680.0}, 'price': {'min': 1e-05, 'max': 19.99998}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'FLRUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'FLR', 'quoteCoin': 'USDT', 'launchTime': '1678953508000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '5', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.00001', 'maxPrice': '19.99998', 'tickSize': '0.00001'}, 'lotSizeFilter': {'maxOrderQty': '379680', 'minOrderQty': '10', 'qtyStep': '10', 'postOnlyMaxOrderQty': '1898400'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}, {'id': 'FTMUSDT', 'symbol': 'FTM/USDT:USDT', 'base': 'FTM', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'FTM', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 1.0, 'price': 0.0001}, 'limits': {'leverage': {'min': 1.0, 'max': 75.0}, 'amount': {'min': 1.0, 'max': 360000.0}, 'price': {'min': 0.0001, 'max': 199.9998}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'FTMUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'FTM', 'quoteCoin': 'USDT', 'launchTime': '1632355200000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '4', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '75.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.0001', 'maxPrice': '199.9998', 'tickSize': '0.0001'}, 'lotSizeFilter': {'maxOrderQty': '360000', 'minOrderQty': '1', 'qtyStep': '1', 'postOnlyMaxOrderQty': '3600000'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'normalOnly'}}, {'id': 'FXSUSDT', 'symbol': 'FXS/USDT:USDT', 'base': 'FXS', 'quote': 'USDT', 'settle': 'USDT', 'baseId': 'FXS', 'quoteId': 'USDT', 'settleId': 'USDT', 'type': 'swap', 'spot': False, 'margin': None, 'swap': True, 'future': False, 'option': False, 'active': True, 'contract': True, 'linear': True, 'inverse': False, 'taker': 0.0006, 'maker': 0.0001, 'contractSize': 1.0, 'expiry': None, 'expiryDatetime': None, 'strike': None, 'optionType': None, 'precision': {'amount': 0.01, 'price': 0.0005}, 'limits': {'leverage': {'min': 1.0, 'max': 25.0}, 'amount': {'min': 0.01, 'max': 2400.0}, 'price': {'min': 0.0005, 'max': 999.999}, 'cost': {'min': None, 'max': None}}, 'info': {'symbol': 'FXSUSDT', 'contractType': 'LinearPerpetual', 'status': 'Trading', 'baseCoin': 'FXS', 'quoteCoin': 'USDT', 'launchTime': '1649044820000', 'deliveryTime': '0', 'deliveryFeeRate': '', 'priceScale': '4', 'leverageFilter': {'minLeverage': '1', 'maxLeverage': '25.00', 'leverageStep': '0.01'}, 'priceFilter': {'minPrice': '0.0005', 'maxPrice': '999.9990', 'tickSize': '0.0005'}, 'lotSizeFilter': {'maxOrderQty': '2400.00', 'minOrderQty': '0.01', 'qtyStep': '0.01', 'postOnlyMaxOrderQty': '24000.00'}, 'unifiedMarginTrade': True, 'fundingInterval': '480', 'settleCoin': 'USDT', 'copyTrading': 'none'}}],
    "binanceus": json.loads('[{"id": "BTCUSD4", "lowercaseId": "btcusd4", "symbol": "BTC/USD4", "base": "BTC", "quote": "USD4", "settle": null, "baseId": "BTC", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1e-06, "price": 0.01, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1e-06, "max": 9000.0}, "price": {"min": 0.01, "max": 100000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 60.25597724}}, "created": null, "info": {"symbol": "BTCUSD4", "status": "BREAK", "baseAsset": "BTC", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "2", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0100", "maxPrice": "100000.0000", "tickSize": "0.0100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00000100", "maxQty": "9000.00000000", "stepSize": "0.00000100"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "60.25597724", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ETHUSD4", "lowercaseId": "ethusd4", "symbol": "ETH/USD4", "base": "ETH", "quote": "USD4", "settle": null, "baseId": "ETH", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1e-05, "price": 0.01, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1e-05, "max": 9000.0}, "price": {"min": 0.01, "max": 100000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 514.38625342}}, "created": null, "info": {"symbol": "ETHUSD4", "status": "BREAK", "baseAsset": "ETH", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0100", "maxPrice": "100000.0000", "tickSize": "0.0100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00001000", "maxQty": "9000.00000000", "stepSize": "0.00001000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "514.38625342", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "XRPUSD", "lowercaseId": "xrpusd", "symbol": "XRP/USD4", "base": "XRP", "quote": "USD4", "settle": null, "baseId": "XRP", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 0.0001, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 9222449.0}}, "created": null, "info": {"symbol": "XRPUSD", "status": "BREAK", "baseAsset": "XRP", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "10000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "9222449.00000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BCHUSD4", "lowercaseId": "bchusd4", "symbol": "BCH/USD4", "base": "BCH", "quote": "USD4", "settle": null, "baseId": "BCH", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1e-05, "price": 0.01, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1e-05, "max": 9000.0}, "price": {"min": 0.01, "max": 100000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 598.64156564}}, "created": null, "info": {"symbol": "BCHUSD4", "status": "BREAK", "baseAsset": "BCH", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "2", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0100", "maxPrice": "100000.0000", "tickSize": "0.0100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00001000", "maxQty": "9000.00000000", "stepSize": "0.00001000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "598.64156564", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LTCUSD4", "lowercaseId": "ltcusd4", "symbol": "LTC/USD4", "base": "LTC", "quote": "USD4", "settle": null, "baseId": "LTC", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1e-05, "price": 0.01, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1e-05, "max": 9000.0}, "price": {"min": 0.01, "max": 100000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 5234.97036374}}, "created": null, "info": {"symbol": "LTCUSD4", "status": "BREAK", "baseAsset": "LTC", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "2", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0100", "maxPrice": "100000.0000", "tickSize": "0.0100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00001000", "maxQty": "9000.00000000", "stepSize": "0.00001000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "5234.97036374", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "USDTUSD4", "lowercaseId": "usdtusd4", "symbol": "USDT/USD4", "base": "USDT", "quote": "USD4", "settle": null, "baseId": "USDT", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 10368912.79569097}}, "created": null, "info": {"symbol": "USDTUSD4", "status": "BREAK", "baseAsset": "USDT", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "9000000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "10368912.79569097", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BTCUSDT", "lowercaseId": "btcusdt", "symbol": "BTC/USDT", "base": "BTC", "quote": "USDT", "settle": null, "baseId": "BTC", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1e-05, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1e-05, "max": 9000.0}, "price": {"min": 0.01, "max": 1000000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 14.41875058}}, "created": null, "info": {"symbol": "BTCUSDT", "status": "TRADING", "baseAsset": "BTC", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "1000000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00001000", "maxQty": "9000.00000000", "stepSize": "0.00001000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "14.41875058", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ETHUSDT", "lowercaseId": "ethusdt", "symbol": "ETH/USDT", "base": "ETH", "quote": "USDT", "settle": null, "baseId": "ETH", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.0001, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.0001, "max": 9000.0}, "price": {"min": 0.01, "max": 1000000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 94.50597208}}, "created": null, "info": {"symbol": "ETHUSDT", "status": "TRADING", "baseAsset": "ETH", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "1000000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00010000", "maxQty": "9000.00000000", "stepSize": "0.00010000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "94.50597208", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "XRPUSDT", "lowercaseId": "xrpusdt", "symbol": "XRP/USDT", "base": "XRP", "quote": "USDT", "settle": null, "baseId": "XRP", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 100000.0}, "price": {"min": 0.0001, "max": 100.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 253273.41666666}}, "created": null, "info": {"symbol": "XRPUSDT", "status": "TRADING", "baseAsset": "XRP", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "100.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "100000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "253273.41666666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BCHUSDT", "lowercaseId": "bchusdt", "symbol": "BCH/USDT", "base": "BCH", "quote": "USDT", "settle": null, "baseId": "BCH", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.1, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 90000.0}, "price": {"min": 0.1, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 38.928825}}, "created": null, "info": {"symbol": "BCHUSDT", "status": "TRADING", "baseAsset": "BCH", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.10000000", "maxPrice": "100000.00000000", "tickSize": "0.10000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "90000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "38.92882500", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LTCUSDT", "lowercaseId": "ltcusdt", "symbol": "LTC/USDT", "base": "LTC", "quote": "USDT", "settle": null, "baseId": "LTC", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 90000.0}, "price": {"min": 0.01, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 619.93581666}}, "created": null, "info": {"symbol": "LTCUSDT", "status": "TRADING", "baseAsset": "LTC", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "100000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "90000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "619.93581666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BNBUSD4", "lowercaseId": "bnbusd4", "symbol": "BNB/USD4", "base": "BNB", "quote": "USD4", "settle": null, "baseId": "BNB", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 900000.0}, "price": {"min": 0.0001, "max": 100000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 1951.71282905}}, "created": null, "info": {"symbol": "BNBUSD4", "status": "BREAK", "baseAsset": "BNB", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "100000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "900000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1951.71282905", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BNBUSDT", "lowercaseId": "bnbusdt", "symbol": "BNB/USDT", "base": "BNB", "quote": "USDT", "settle": null, "baseId": "BNB", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.1, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 900000.0}, "price": {"min": 0.1, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 345.98359583}}, "created": null, "info": {"symbol": "BNBUSDT", "status": "TRADING", "baseAsset": "BNB", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.10000000", "maxPrice": "100000.00000000", "tickSize": "0.10000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "900000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "345.98359583", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ETHBTC", "lowercaseId": "ethbtc", "symbol": "ETH/BTC", "base": "ETH", "quote": "BTC", "settle": null, "baseId": "ETH", "quoteId": "BTC", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.0001, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.0001, "max": 100000.0}, "price": {"min": 1e-05, "max": 922327.0}, "cost": {"min": 0.0001, "max": null}, "market": {"min": 0.0, "max": 79.316735}}, "created": null, "info": {"symbol": "ETHBTC", "status": "TRADING", "baseAsset": "ETH", "baseAssetPrecision": "8", "quoteAsset": "BTC", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "922327.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00010000", "maxQty": "100000.00000000", "stepSize": "0.00010000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "0.00010000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "79.31673500", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "XRPBTC", "lowercaseId": "xrpbtc", "symbol": "XRP/BTC", "base": "XRP", "quote": "BTC", "settle": null, "baseId": "XRP", "quoteId": "BTC", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-08, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 90000000.0}, "price": {"min": 1e-08, "max": 1000.0}, "cost": {"min": 0.0001, "max": null}, "market": {"min": 0.0, "max": 42853300.0}}, "created": null, "info": {"symbol": "XRPBTC", "status": "BREAK", "baseAsset": "XRP", "baseAssetPrecision": "8", "quoteAsset": "BTC", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000001", "maxPrice": "1000.00000000", "tickSize": "0.00000001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "90000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "0.00010000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "42853300.00000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BNBBTC", "lowercaseId": "bnbbtc", "symbol": "BNB/BTC", "base": "BNB", "quote": "BTC", "settle": null, "baseId": "BNB", "quoteId": "BTC", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 1e-06, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 100000.0}, "price": {"min": 1e-06, "max": 100000.0}, "cost": {"min": 0.0001, "max": null}, "market": {"min": 0.0, "max": 67.81212916}}, "created": null, "info": {"symbol": "BNBBTC", "status": "TRADING", "baseAsset": "BNB", "baseAssetPrecision": "8", "quoteAsset": "BTC", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000100", "maxPrice": "100000.00000000", "tickSize": "0.00000100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "100000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "0.00010000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "67.81212916", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LTCBTC", "lowercaseId": "ltcbtc", "symbol": "LTC/BTC", "base": "LTC", "quote": "BTC", "settle": null, "baseId": "LTC", "quoteId": "BTC", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 1e-06, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 100000.0}, "price": {"min": 1e-06, "max": 100000.0}, "cost": {"min": 0.0001, "max": null}, "market": {"min": 0.0, "max": 187.60317991}}, "created": null, "info": {"symbol": "LTCBTC", "status": "TRADING", "baseAsset": "LTC", "baseAssetPrecision": "8", "quoteAsset": "BTC", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000100", "maxPrice": "100000.00000000", "tickSize": "0.00000100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "100000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "0.00010000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "187.60317991", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BCHBTC", "lowercaseId": "bchbtc", "symbol": "BCH/BTC", "base": "BCH", "quote": "BTC", "settle": null, "baseId": "BCH", "quoteId": "BTC", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 10000000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 0.0001, "max": null}, "market": {"min": 0.0, "max": 708.32305486}}, "created": null, "info": {"symbol": "BCHBTC", "status": "BREAK", "baseAsset": "BCH", "baseAssetPrecision": "8", "quoteAsset": "BTC", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "10000000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "0.00010000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "708.32305486", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ADAUSD4", "lowercaseId": "adausd4", "symbol": "ADA/USD4", "base": "ADA", "quote": "USD4", "settle": null, "baseId": "ADA", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 872196.53409722}}, "created": null, "info": {"symbol": "ADAUSD4", "status": "BREAK", "baseAsset": "ADA", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "872196.53409722", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BATUSD4", "lowercaseId": "batusd4", "symbol": "BAT/USD4", "base": "BAT", "quote": "USD4", "settle": null, "baseId": "BAT", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 337996.54613542}}, "created": null, "info": {"symbol": "BATUSD4", "status": "BREAK", "baseAsset": "BAT", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "337996.54613542", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ETCUSD4", "lowercaseId": "etcusd4", "symbol": "ETC/USD4", "base": "ETC", "quote": "USD4", "settle": null, "baseId": "ETC", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 7672.11976042}}, "created": null, "info": {"symbol": "ETCUSD4", "status": "BREAK", "baseAsset": "ETC", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "7672.11976042", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "XLMUSD4", "lowercaseId": "xlmusd4", "symbol": "XLM/USD4", "base": "XLM", "quote": "USD4", "settle": null, "baseId": "XLM", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 1789500.0799861}}, "created": null, "info": {"symbol": "XLMUSD4", "status": "BREAK", "baseAsset": "XLM", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1789500.07998610", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ZRXUSD4", "lowercaseId": "zrxusd4", "symbol": "ZRX/USD4", "base": "ZRX", "quote": "USD4", "settle": null, "baseId": "ZRX", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 231944.61956597}}, "created": null, "info": {"symbol": "ZRXUSD4", "status": "BREAK", "baseAsset": "ZRX", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "231944.61956597", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ADAUSDT", "lowercaseId": "adausdt", "symbol": "ADA/USDT", "base": "ADA", "quote": "USDT", "settle": null, "baseId": "ADA", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 342019.86208333}}, "created": null, "info": {"symbol": "ADAUSDT", "status": "TRADING", "baseAsset": "ADA", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "342019.86208333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BATUSDT", "lowercaseId": "batusdt", "symbol": "BAT/USDT", "base": "BAT", "quote": "USDT", "settle": null, "baseId": "BAT", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 12205.42323651}}, "created": null, "info": {"symbol": "BATUSDT", "status": "TRADING", "baseAsset": "BAT", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "900000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "12205.42323651", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ETCUSDT", "lowercaseId": "etcusdt", "symbol": "ETC/USDT", "base": "ETC", "quote": "USDT", "settle": null, "baseId": "ETC", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 827.05704166}}, "created": null, "info": {"symbol": "ETCUSDT", "status": "TRADING", "baseAsset": "ETC", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "10000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "827.05704166", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "XLMUSDT", "lowercaseId": "xlmusdt", "symbol": "XLM/USDT", "base": "XLM", "quote": "USDT", "settle": null, "baseId": "XLM", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 210862.62916666}}, "created": null, "info": {"symbol": "XLMUSDT", "status": "TRADING", "baseAsset": "XLM", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "210862.62916666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ZRXUSDT", "lowercaseId": "zrxusdt", "symbol": "ZRX/USDT", "base": "ZRX", "quote": "USDT", "settle": null, "baseId": "ZRX", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 11701.79166666}}, "created": null, "info": {"symbol": "ZRXUSDT", "status": "TRADING", "baseAsset": "ZRX", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "900000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "11701.79166666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LINKUSD4", "lowercaseId": "linkusd4", "symbol": "LINK/USD4", "base": "LINK", "quote": "USD4", "settle": null, "baseId": "LINK", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 39973.88115358}}, "created": null, "info": {"symbol": "LINKUSD4", "status": "BREAK", "baseAsset": "LINK", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "39973.88115358", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "RVNUSD4", "lowercaseId": "rvnusd4", "symbol": "RVN/USD4", "base": "RVN", "quote": "USD4", "settle": null, "baseId": "RVN", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 3164694.50375261}}, "created": null, "info": {"symbol": "RVNUSD4", "status": "BREAK", "baseAsset": "RVN", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3164694.50375261", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "DASHUSD4", "lowercaseId": "dashusd4", "symbol": "DASH/USD4", "base": "DASH", "quote": "USD4", "settle": null, "baseId": "DASH", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1e-05, "price": 0.01, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1e-05, "max": 900000.0}, "price": {"min": 0.01, "max": 100000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 1187.00316815}}, "created": null, "info": {"symbol": "DASHUSD4", "status": "BREAK", "baseAsset": "DASH", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "2", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0100", "maxPrice": "100000.0000", "tickSize": "0.0100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00001000", "maxQty": "900000.00000000", "stepSize": "0.00001000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1187.00316815", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ZECUSD4", "lowercaseId": "zecusd4", "symbol": "ZEC/USD4", "base": "ZEC", "quote": "USD4", "settle": null, "baseId": "ZEC", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1e-05, "price": 0.01, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1e-05, "max": 900000.0}, "price": {"min": 0.01, "max": 100000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 3010.12044466}}, "created": null, "info": {"symbol": "ZECUSD4", "status": "BREAK", "baseAsset": "ZEC", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "2", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0100", "maxPrice": "100000.0000", "tickSize": "0.0100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00001000", "maxQty": "900000.00000000", "stepSize": "0.00001000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3010.12044466", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ALGOUSD4", "lowercaseId": "algousd4", "symbol": "ALGO/USD4", "base": "ALGO", "quote": "USD4", "settle": null, "baseId": "ALGO", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 900000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 1014693.87050868}}, "created": null, "info": {"symbol": "ALGOUSD4", "status": "BREAK", "baseAsset": "ALGO", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "10000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "900000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1014693.87050868", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "IOTAUSD4", "lowercaseId": "iotausd4", "symbol": "IOTA/USD4", "base": "IOTA", "quote": "USD4", "settle": null, "baseId": "IOTA", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 146719.39888464}}, "created": null, "info": {"symbol": "IOTAUSD4", "status": "BREAK", "baseAsset": "IOTA", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "146719.39888464", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BUSDUSD4", "lowercaseId": "busdusd4", "symbol": "BUSD/USD4", "base": "BUSD", "quote": "USD4", "settle": null, "baseId": "BUSD", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 9000000.0}, "price": {"min": 0.0001, "max": 100000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 4652432.407625}}, "created": null, "info": {"symbol": "BUSDUSD4", "status": "BREAK", "baseAsset": "BUSD", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "100000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "9000000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "4652432.40762500", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BTCBUSD", "lowercaseId": "btcbusd", "symbol": "BTC/BUSD", "base": "BTC", "quote": "BUSD", "settle": null, "baseId": "BTC", "quoteId": "BUSD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1e-05, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1e-05, "max": 9000.0}, "price": {"min": 0.01, "max": 1000000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 0.06097176}}, "created": null, "info": {"symbol": "BTCBUSD", "status": "BREAK", "baseAsset": "BTC", "baseAssetPrecision": "8", "quoteAsset": "BUSD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "2", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "1000000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00001000", "maxQty": "9000.00000000", "stepSize": "0.00001000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "0.06097176", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "DOGEUSDT", "lowercaseId": "dogeusdt", "symbol": "DOGE/USDT", "base": "DOGE", "quote": "USDT", "settle": null, "baseId": "DOGE", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 2231593.90416666}}, "created": null, "info": {"symbol": "DOGEUSDT", "status": "TRADING", "baseAsset": "DOGE", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "2231593.90416666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "WAVESUSD4", "lowercaseId": "wavesusd4", "symbol": "WAVES/USD4", "base": "WAVES", "quote": "USD4", "settle": null, "baseId": "WAVES", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 13445.78859375}}, "created": null, "info": {"symbol": "WAVESUSD4", "status": "BREAK", "baseAsset": "WAVES", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "13445.78859375", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ATOMUSDT", "lowercaseId": "atomusdt", "symbol": "ATOM/USDT", "base": "ATOM", "quote": "USDT", "settle": null, "baseId": "ATOM", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 4160.98645833}}, "created": null, "info": {"symbol": "ATOMUSDT", "status": "TRADING", "baseAsset": "ATOM", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "4160.98645833", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ATOMUSD4", "lowercaseId": "atomusd4", "symbol": "ATOM/USD4", "base": "ATOM", "quote": "USD4", "settle": null, "baseId": "ATOM", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 900000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 23542.84988472}}, "created": null, "info": {"symbol": "ATOMUSD4", "status": "BREAK", "baseAsset": "ATOM", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "10000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "900000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "23542.84988472", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "NEOUSDT", "lowercaseId": "neousdt", "symbol": "NEO/USDT", "base": "NEO", "quote": "USDT", "settle": null, "baseId": "NEO", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 829.71008333}}, "created": null, "info": {"symbol": "NEOUSDT", "status": "TRADING", "baseAsset": "NEO", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "10000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "829.71008333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "NEOUSD4", "lowercaseId": "neousd4", "symbol": "NEO/USD4", "base": "NEO", "quote": "USD4", "settle": null, "baseId": "NEO", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 900000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 6970.17727589}}, "created": null, "info": {"symbol": "NEOUSD4", "status": "BREAK", "baseAsset": "NEO", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "10000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "900000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "6970.17727589", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "VETUSDT", "lowercaseId": "vetusdt", "symbol": "VET/USDT", "base": "VET", "quote": "USDT", "settle": null, "baseId": "VET", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 9000000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 1262297.12579166}}, "created": null, "info": {"symbol": "VETUSDT", "status": "TRADING", "baseAsset": "VET", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "9000000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1262297.12579166", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "QTUMUSDT", "lowercaseId": "qtumusdt", "symbol": "QTUM/USDT", "base": "QTUM", "quote": "USDT", "settle": null, "baseId": "QTUM", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 90000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 1600.81041666}}, "created": null, "info": {"symbol": "QTUMUSDT", "status": "TRADING", "baseAsset": "QTUM", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "90000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1600.81041666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "QTUMUSD4", "lowercaseId": "qtumusd4", "symbol": "QTUM/USD4", "base": "QTUM", "quote": "USD4", "settle": null, "baseId": "QTUM", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 900000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 27528.80557847}}, "created": null, "info": {"symbol": "QTUMUSD4", "status": "BREAK", "baseAsset": "QTUM", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "10000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "900000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "27528.80557847", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "NANOUSD", "lowercaseId": "nanousd", "symbol": "NANO/USD4", "base": "NANO", "quote": "USD4", "settle": null, "baseId": "NANO", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 63577.06601807}}, "created": null, "info": {"symbol": "NANOUSD", "status": "BREAK", "baseAsset": "NANO", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "63577.06601807", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ICXUSD4", "lowercaseId": "icxusd4", "symbol": "ICX/USD4", "base": "ICX", "quote": "USD4", "settle": null, "baseId": "ICX", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 209894.29939541}}, "created": null, "info": {"symbol": "ICXUSD4", "status": "BREAK", "baseAsset": "ICX", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "209894.29939541", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ENJUSD4", "lowercaseId": "enjusd4", "symbol": "ENJ/USD4", "base": "ENJ", "quote": "USD4", "settle": null, "baseId": "ENJ", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 119667.26490618}}, "created": null, "info": {"symbol": "ENJUSD4", "status": "BREAK", "baseAsset": "ENJ", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "119667.26490618", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ONTUSD4", "lowercaseId": "ontusd4", "symbol": "ONT/USD4", "base": "ONT", "quote": "USD4", "settle": null, "baseId": "ONT", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 159092.19483322}}, "created": null, "info": {"symbol": "ONTUSD4", "status": "BREAK", "baseAsset": "ONT", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "159092.19483322", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ONTUSDT", "lowercaseId": "ontusdt", "symbol": "ONT/USDT", "base": "ONT", "quote": "USDT", "settle": null, "baseId": "ONT", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 19892.45833333}}, "created": null, "info": {"symbol": "ONTUSDT", "status": "TRADING", "baseAsset": "ONT", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "900000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "19892.45833333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ZILUSD4", "lowercaseId": "zilusd4", "symbol": "ZIL/USD4", "base": "ZIL", "quote": "USD4", "settle": null, "baseId": "ZIL", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 4995249.66417651}}, "created": null, "info": {"symbol": "ZILUSD4", "status": "BREAK", "baseAsset": "ZIL", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "4995249.66417651", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ZILBUSD", "lowercaseId": "zilbusd", "symbol": "ZIL/BUSD", "base": "ZIL", "quote": "BUSD", "settle": null, "baseId": "ZIL", "quoteId": "BUSD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 1668321.55357887}}, "created": null, "info": {"symbol": "ZILBUSD", "status": "BREAK", "baseAsset": "ZIL", "baseAssetPrecision": "8", "quoteAsset": "BUSD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1668321.55357887", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "VETUSD4", "lowercaseId": "vetusd4", "symbol": "VET/USD4", "base": "VET", "quote": "USD4", "settle": null, "baseId": "VET", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 90000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 29269120.59173037}}, "created": null, "info": {"symbol": "VETUSD4", "status": "BREAK", "baseAsset": "VET", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "90000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "29269120.59173037", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BNBBUSD", "lowercaseId": "bnbbusd", "symbol": "BNB/BUSD", "base": "BNB", "quote": "BUSD", "settle": null, "baseId": "BNB", "quoteId": "BUSD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.1, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 900000.0}, "price": {"min": 0.1, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 28.32102479}}, "created": null, "info": {"symbol": "BNBBUSD", "status": "BREAK", "baseAsset": "BNB", "baseAssetPrecision": "8", "quoteAsset": "BUSD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.10000000", "maxPrice": "100000.00000000", "tickSize": "0.10000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "900000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "28.32102479", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "XRPBUSD", "lowercaseId": "xrpbusd", "symbol": "XRP/BUSD", "base": "XRP", "quote": "BUSD", "settle": null, "baseId": "XRP", "quoteId": "BUSD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 100000.0}, "price": {"min": 0.0001, "max": 100.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 17086683.0}}, "created": null, "info": {"symbol": "XRPBUSD", "status": "BREAK", "baseAsset": "XRP", "baseAssetPrecision": "8", "quoteAsset": "BUSD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "100.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "100000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "17086683.00000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ETHBUSD", "lowercaseId": "ethbusd", "symbol": "ETH/BUSD", "base": "ETH", "quote": "BUSD", "settle": null, "baseId": "ETH", "quoteId": "BUSD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.0001, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.0001, "max": 90000.0}, "price": {"min": 0.01, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 0.549}}, "created": null, "info": {"symbol": "ETHBUSD", "status": "BREAK", "baseAsset": "ETH", "baseAssetPrecision": "8", "quoteAsset": "BUSD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "100000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00010000", "maxQty": "90000.00000000", "stepSize": "0.00010000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "0.54900000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ALGOBUSD", "lowercaseId": "algobusd", "symbol": "ALGO/BUSD", "base": "ALGO", "quote": "BUSD", "settle": null, "baseId": "ALGO", "quoteId": "BUSD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 105297.92601807}}, "created": null, "info": {"symbol": "ALGOBUSD", "status": "BREAK", "baseAsset": "ALGO", "baseAssetPrecision": "8", "quoteAsset": "BUSD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "900000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "105297.92601807", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "XTZUSD4", "lowercaseId": "xtzusd4", "symbol": "XTZ/USD4", "base": "XTZ", "quote": "USD4", "settle": null, "baseId": "XTZ", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 115671.22192361}}, "created": null, "info": {"symbol": "XTZUSD4", "status": "BREAK", "baseAsset": "XTZ", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "115671.22192361", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "XTZBUSD", "lowercaseId": "xtzbusd", "symbol": "XTZ/BUSD", "base": "XTZ", "quote": "BUSD", "settle": null, "baseId": "XTZ", "quoteId": "BUSD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 29074.57905837}}, "created": null, "info": {"symbol": "XTZBUSD", "status": "BREAK", "baseAsset": "XTZ", "baseAssetPrecision": "8", "quoteAsset": "BUSD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "29074.57905837", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "HBARUSD4", "lowercaseId": "hbarusd4", "symbol": "HBAR/USD4", "base": "HBAR", "quote": "USD4", "settle": null, "baseId": "HBAR", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 4686047.38798611}}, "created": null, "info": {"symbol": "HBARUSD4", "status": "BREAK", "baseAsset": "HBAR", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "4686047.38798611", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "HBARBUSD", "lowercaseId": "hbarbusd", "symbol": "HBAR/BUSD", "base": "HBAR", "quote": "BUSD", "settle": null, "baseId": "HBAR", "quoteId": "BUSD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 651459.22512161}}, "created": null, "info": {"symbol": "HBARBUSD", "status": "BREAK", "baseAsset": "HBAR", "baseAssetPrecision": "8", "quoteAsset": "BUSD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "651459.22512161", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "OMGUSD4", "lowercaseId": "omgusd4", "symbol": "OMG/USD4", "base": "OMG", "quote": "USD4", "settle": null, "baseId": "OMG", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 69571.07908617}}, "created": null, "info": {"symbol": "OMGUSD4", "status": "BREAK", "baseAsset": "OMG", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "69571.07908617", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "OMGBUSD", "lowercaseId": "omgbusd", "symbol": "OMG/BUSD", "base": "OMG", "quote": "BUSD", "settle": null, "baseId": "OMG", "quoteId": "BUSD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 28559.18976372}}, "created": null, "info": {"symbol": "OMGBUSD", "status": "BREAK", "baseAsset": "OMG", "baseAssetPrecision": "8", "quoteAsset": "BUSD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "28559.18976372", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "MATICUSD4", "lowercaseId": "maticusd4", "symbol": "MATIC/USD4", "base": "MATIC", "quote": "USD4", "settle": null, "baseId": "MATIC", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 550216.36824183}}, "created": null, "info": {"symbol": "MATICUSD4", "status": "BREAK", "baseAsset": "MATIC", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "550216.36824183", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "MATICBUSD", "lowercaseId": "maticbusd", "symbol": "MATIC/BUSD", "base": "MATIC", "quote": "BUSD", "settle": null, "baseId": "MATIC", "quoteId": "BUSD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 238255.33964559}}, "created": null, "info": {"symbol": "MATICBUSD", "status": "BREAK", "baseAsset": "MATIC", "baseAssetPrecision": "8", "quoteAsset": "BUSD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "238255.33964559", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "XTZBTC", "lowercaseId": "xtzbtc", "symbol": "XTZ/BTC", "base": "XTZ", "quote": "BTC", "settle": null, "baseId": "XTZ", "quoteId": "BTC", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-08, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 90000000.0}, "price": {"min": 1e-08, "max": 1000.0}, "cost": {"min": 0.0001, "max": null}, "market": {"min": 0.0, "max": 46129.285625}}, "created": null, "info": {"symbol": "XTZBTC", "status": "BREAK", "baseAsset": "XTZ", "baseAssetPrecision": "8", "quoteAsset": "BTC", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000001", "maxPrice": "1000.00000000", "tickSize": "0.00000001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "90000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "0.00010000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "46129.28562500", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ADABTC", "lowercaseId": "adabtc", "symbol": "ADA/BTC", "base": "ADA", "quote": "BTC", "settle": null, "baseId": "ADA", "quoteId": "BTC", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-08, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 90000000.0}, "price": {"min": 1e-08, "max": 1000.0}, "cost": {"min": 0.0001, "max": null}, "market": {"min": 0.0, "max": 26806.03083333}}, "created": null, "info": {"symbol": "ADABTC", "status": "TRADING", "baseAsset": "ADA", "baseAssetPrecision": "8", "quoteAsset": "BTC", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000001", "maxPrice": "1000.00000000", "tickSize": "0.00000001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "90000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "0.00010000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "26806.03083333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "REPBUSD", "lowercaseId": "repbusd", "symbol": "REP/BUSD", "base": "REP", "quote": "BUSD", "settle": null, "baseId": "REP", "quoteId": "BUSD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 1000000.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 3263.53365497}}, "created": null, "info": {"symbol": "REPBUSD", "status": "BREAK", "baseAsset": "REP", "baseAssetPrecision": "8", "quoteAsset": "BUSD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "10000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "1000000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3263.53365497", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "REPUSD", "lowercaseId": "repusd", "symbol": "REP/USD4", "base": "REP", "quote": "USD4", "settle": null, "baseId": "REP", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.01, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 1000000.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 6165.38308681}}, "created": null, "info": {"symbol": "REPUSD", "status": "BREAK", "baseAsset": "REP", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "2", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0100", "maxPrice": "10000.0000", "tickSize": "0.0100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "1000000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "6165.38308681", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "EOSBUSD", "lowercaseId": "eosbusd", "symbol": "EOS/BUSD", "base": "EOS", "quote": "BUSD", "settle": null, "baseId": "EOS", "quoteId": "BUSD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 42510.010625}}, "created": null, "info": {"symbol": "EOSBUSD", "status": "BREAK", "baseAsset": "EOS", "baseAssetPrecision": "8", "quoteAsset": "BUSD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "42510.01062500", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "EOSUSD4", "lowercaseId": "eosusd4", "symbol": "EOS/USD4", "base": "EOS", "quote": "USD4", "settle": null, "baseId": "EOS", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 120578.75166782}}, "created": null, "info": {"symbol": "EOSUSD4", "status": "BREAK", "baseAsset": "EOS", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "120578.75166782", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "DOGEUSD4", "lowercaseId": "dogeusd4", "symbol": "DOGE/USD4", "base": "DOGE", "quote": "USD4", "settle": null, "baseId": "DOGE", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 90000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 12487357.58200139}}, "created": null, "info": {"symbol": "DOGEUSD4", "status": "BREAK", "baseAsset": "DOGE", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "90000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "12487357.58200139", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "KNCUSD4", "lowercaseId": "kncusd4", "symbol": "KNC/USD4", "base": "KNC", "quote": "USD4", "settle": null, "baseId": "KNC", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 900000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 141590.56748471}}, "created": null, "info": {"symbol": "KNCUSD4", "status": "BREAK", "baseAsset": "KNC", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "10000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "900000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "141590.56748471", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "KNCUSDT", "lowercaseId": "kncusdt", "symbol": "KNC/USDT", "base": "KNC", "quote": "USDT", "settle": null, "baseId": "KNC", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9222449.0}, "price": {"min": 0.0001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 13087.81083333}}, "created": null, "info": {"symbol": "KNCUSDT", "status": "TRADING", "baseAsset": "KNC", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "10000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9222449.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "13087.81083333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "VTHOUSDT", "lowercaseId": "vthousdt", "symbol": "VTHO/USDT", "base": "VTHO", "quote": "USDT", "settle": null, "baseId": "VTHO", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-06, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 1e-06, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 4467889.1625}}, "created": null, "info": {"symbol": "VTHOUSDT", "status": "TRADING", "baseAsset": "VTHO", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000100", "maxPrice": "1000.00000000", "tickSize": "0.00000100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "4467889.16250000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "VTHOUSD4", "lowercaseId": "vthousd4", "symbol": "VTHO/USD4", "base": "VTHO", "quote": "USD4", "settle": null, "baseId": "VTHO", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 92141578.0}}, "created": null, "info": {"symbol": "VTHOUSD4", "status": "BREAK", "baseAsset": "VTHO", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "92141578.00000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "USDCUSD4", "lowercaseId": "usdcusd4", "symbol": "USDC/USD4", "base": "USDC", "quote": "USD4", "settle": null, "baseId": "USDC", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 2388543.6271647}}, "created": null, "info": {"symbol": "USDCUSD4", "status": "BREAK", "baseAsset": "USDC", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "2388543.62716470", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "COMPUSDT", "lowercaseId": "compusdt", "symbol": "COMP/USDT", "base": "COMP", "quote": "USDT", "settle": null, "baseId": "COMP", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 90000.0}, "price": {"min": 0.01, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 275.50141004}}, "created": null, "info": {"symbol": "COMPUSDT", "status": "TRADING", "baseAsset": "COMP", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "100000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "90000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "275.50141004", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "COMPUSD4", "lowercaseId": "compusd4", "symbol": "COMP/USD4", "base": "COMP", "quote": "USD4", "settle": null, "baseId": "COMP", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1e-05, "price": 0.01, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1e-05, "max": 900000.0}, "price": {"min": 0.01, "max": 100000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 1125.52000532}}, "created": null, "info": {"symbol": "COMPUSD4", "status": "BREAK", "baseAsset": "COMP", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "2", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0100", "maxPrice": "100000.0000", "tickSize": "0.0100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00001000", "maxQty": "900000.00000000", "stepSize": "0.00001000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1125.52000532", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "MANAUSD4", "lowercaseId": "manausd4", "symbol": "MANA/USD4", "base": "MANA", "quote": "USD4", "settle": null, "baseId": "MANA", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 209209.003287}}, "created": null, "info": {"symbol": "MANAUSD4", "status": "BREAK", "baseAsset": "MANA", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "209209.00328700", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "HNTUSD4", "lowercaseId": "hntusd4", "symbol": "HNT/USD4", "base": "HNT", "quote": "USD4", "settle": null, "baseId": "HNT", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 16697.3326875}}, "created": null, "info": {"symbol": "HNTUSD4", "status": "BREAK", "baseAsset": "HNT", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "16697.33268750", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "HNTUSDT", "lowercaseId": "hntusdt", "symbol": "HNT/USDT", "base": "HNT", "quote": "USDT", "settle": null, "baseId": "HNT", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 12565.23059028}}, "created": null, "info": {"symbol": "HNTUSDT", "status": "BREAK", "baseAsset": "HNT", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "10000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "12565.23059028", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "MKRUSD4", "lowercaseId": "mkrusd4", "symbol": "MKR/USD4", "base": "MKR", "quote": "USD4", "settle": null, "baseId": "MKR", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1e-05, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1e-05, "max": 900000.0}, "price": {"min": 0.0001, "max": 100000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 125.26531854}}, "created": null, "info": {"symbol": "MKRUSD4", "status": "BREAK", "baseAsset": "MKR", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "100000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00001000", "maxQty": "900000.00000000", "stepSize": "0.00001000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "125.26531854", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "MKRUSDT", "lowercaseId": "mkrusdt", "symbol": "MKR/USDT", "base": "MKR", "quote": "USDT", "settle": null, "baseId": "MKR", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.0001, "price": 1.0, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.0001, "max": 90000.0}, "price": {"min": 1.0, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 3.41541458}}, "created": null, "info": {"symbol": "MKRUSDT", "status": "TRADING", "baseAsset": "MKR", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "1.00000000", "maxPrice": "100000.00000000", "tickSize": "1.00000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00010000", "maxQty": "90000.00000000", "stepSize": "0.00010000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3.41541458", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "DAIUSD4", "lowercaseId": "daiusd4", "symbol": "DAI/USD4", "base": "DAI", "quote": "USD4", "settle": null, "baseId": "DAI", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 343079.79620833}}, "created": null, "info": {"symbol": "DAIUSD4", "status": "BREAK", "baseAsset": "DAI", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "343079.79620833", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ONEUSDT", "lowercaseId": "oneusdt", "symbol": "ONE/USDT", "base": "ONE", "quote": "USDT", "settle": null, "baseId": "ONE", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 1987419.33625}}, "created": null, "info": {"symbol": "ONEUSDT", "status": "TRADING", "baseAsset": "ONE", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1987419.33625000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ONEUSD4", "lowercaseId": "oneusd4", "symbol": "ONE/USD4", "base": "ONE", "quote": "USD4", "settle": null, "baseId": "ONE", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 9086742.86125781}}, "created": null, "info": {"symbol": "ONEUSD4", "status": "BREAK", "baseAsset": "ONE", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "9086742.86125781", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BANDUSDT", "lowercaseId": "bandusdt", "symbol": "BAND/USDT", "base": "BAND", "quote": "USDT", "settle": null, "baseId": "BAND", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 90000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 3432.76875}}, "created": null, "info": {"symbol": "BANDUSDT", "status": "TRADING", "baseAsset": "BAND", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "90000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3432.76875000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BANDUSD4", "lowercaseId": "bandusd4", "symbol": "BAND/USD4", "base": "BAND", "quote": "USD4", "settle": null, "baseId": "BAND", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 23792.02808333}}, "created": null, "info": {"symbol": "BANDUSD4", "status": "BREAK", "baseAsset": "BAND", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "23792.02808333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "STORJUSDT", "lowercaseId": "storjusdt", "symbol": "STORJ/USDT", "base": "STORJ", "quote": "USDT", "settle": null, "baseId": "STORJ", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 4599.05}}, "created": null, "info": {"symbol": "STORJUSDT", "status": "TRADING", "baseAsset": "STORJ", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "900000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "4599.05000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "STORJUSD4", "lowercaseId": "storjusd4", "symbol": "STORJ/USD4", "base": "STORJ", "quote": "USD4", "settle": null, "baseId": "STORJ", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 185856.13733148}}, "created": null, "info": {"symbol": "STORJUSD4", "status": "BREAK", "baseAsset": "STORJ", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "185856.13733148", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BUSDUSDT", "lowercaseId": "busdusdt", "symbol": "BUSD/USDT", "base": "BUSD", "quote": "USDT", "settle": null, "baseId": "BUSD", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 15000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 1804138.21485059}}, "created": null, "info": {"symbol": "BUSDUSDT", "status": "BREAK", "baseAsset": "BUSD", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "15000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1804138.21485059", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "UNIUSD4", "lowercaseId": "uniusd4", "symbol": "UNI/USD4", "base": "UNI", "quote": "USD4", "settle": null, "baseId": "UNI", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 21188.32385337}}, "created": null, "info": {"symbol": "UNIUSD4", "status": "BREAK", "baseAsset": "UNI", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "21188.32385337", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "UNIUSDT", "lowercaseId": "uniusdt", "symbol": "UNI/USDT", "base": "UNI", "quote": "USDT", "settle": null, "baseId": "UNI", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 3464.06145833}}, "created": null, "info": {"symbol": "UNIUSDT", "status": "TRADING", "baseAsset": "UNI", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3464.06145833", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SOLUSD4", "lowercaseId": "solusd4", "symbol": "SOL/USD4", "base": "SOL", "quote": "USD4", "settle": null, "baseId": "SOL", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 5053.15706393}}, "created": null, "info": {"symbol": "SOLUSD4", "status": "BREAK", "baseAsset": "SOL", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "5053.15706393", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SOLUSDT", "lowercaseId": "solusdt", "symbol": "SOL/USDT", "base": "SOL", "quote": "USDT", "settle": null, "baseId": "SOL", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 90000.0}, "price": {"min": 0.01, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 3109.94353974}}, "created": null, "info": {"symbol": "SOLUSDT", "status": "TRADING", "baseAsset": "SOL", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "1000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "90000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3109.94353974", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LINKBTC", "lowercaseId": "linkbtc", "symbol": "LINK/BTC", "base": "LINK", "quote": "BTC", "settle": null, "baseId": "LINK", "quoteId": "BTC", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 1e-07, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000000.0}, "price": {"min": 1e-07, "max": 1000.0}, "cost": {"min": 0.0001, "max": null}, "market": {"min": 0.0, "max": 1044.81366666}}, "created": null, "info": {"symbol": "LINKBTC", "status": "TRADING", "baseAsset": "LINK", "baseAssetPrecision": "8", "quoteAsset": "BTC", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000010", "maxPrice": "1000.00000000", "tickSize": "0.00000010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "0.00010000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1044.81366666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "VETBTC", "lowercaseId": "vetbtc", "symbol": "VET/BTC", "base": "VET", "quote": "BTC", "settle": null, "baseId": "VET", "quoteId": "BTC", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-08, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 90000000.0}, "price": {"min": 1e-08, "max": 1000.0}, "cost": {"min": 0.0001, "max": null}, "market": {"min": 0.0, "max": 5885613.40896456}}, "created": null, "info": {"symbol": "VETBTC", "status": "BREAK", "baseAsset": "VET", "baseAssetPrecision": "8", "quoteAsset": "BTC", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000001", "maxPrice": "1000.00000000", "tickSize": "0.00000001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "90000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "0.00010000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "5885613.40896456", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "UNIBTC", "lowercaseId": "unibtc", "symbol": "UNI/BTC", "base": "UNI", "quote": "BTC", "settle": null, "baseId": "UNI", "quoteId": "BTC", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 1e-07, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000000.0}, "price": {"min": 1e-07, "max": 1000.0}, "cost": {"min": 0.0001, "max": null}, "market": {"min": 0.0, "max": 14749.96076389}}, "created": null, "info": {"symbol": "UNIBTC", "status": "BREAK", "baseAsset": "UNI", "baseAssetPrecision": "8", "quoteAsset": "BTC", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000010", "maxPrice": "1000.00000000", "tickSize": "0.00000010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "0.00010000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "14749.96076389", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "EGLDUSDT", "lowercaseId": "egldusdt", "symbol": "EGLD/USDT", "base": "EGLD", "quote": "USDT", "settle": null, "baseId": "EGLD", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.01, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 103.99895833}}, "created": null, "info": {"symbol": "EGLDUSDT", "status": "TRADING", "baseAsset": "EGLD", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "100000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "103.99895833", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "EGLDUSD4", "lowercaseId": "egldusd4", "symbol": "EGLD/USD4", "base": "EGLD", "quote": "USD4", "settle": null, "baseId": "EGLD", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 900000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 3151.13448646}}, "created": null, "info": {"symbol": "EGLDUSD4", "status": "BREAK", "baseAsset": "EGLD", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "10000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "900000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3151.13448646", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "PAXGUSDT", "lowercaseId": "paxgusdt", "symbol": "PAXG/USDT", "base": "PAXG", "quote": "USDT", "settle": null, "baseId": "PAXG", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.0001, "price": 1.0, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.0001, "max": 9000.0}, "price": {"min": 1.0, "max": 1000000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 3.00475125}}, "created": null, "info": {"symbol": "PAXGUSDT", "status": "TRADING", "baseAsset": "PAXG", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "1.00000000", "maxPrice": "1000000.00000000", "tickSize": "1.00000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00010000", "maxQty": "9000.00000000", "stepSize": "0.00010000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3.00475125", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "PAXGUSD4", "lowercaseId": "paxgusd4", "symbol": "PAXG/USD4", "base": "PAXG", "quote": "USD4", "settle": null, "baseId": "PAXG", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1e-06, "price": 0.01, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1e-06, "max": 9000.0}, "price": {"min": 0.01, "max": 1000000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 64.75339461}}, "created": null, "info": {"symbol": "PAXGUSD4", "status": "BREAK", "baseAsset": "PAXG", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "2", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0100", "maxPrice": "1000000.0000", "tickSize": "0.0100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00000100", "maxQty": "9000.00000000", "stepSize": "0.00000100"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "64.75339461", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "OXTUSDT", "lowercaseId": "oxtusdt", "symbol": "OXT/USDT", "base": "OXT", "quote": "USDT", "settle": null, "baseId": "OXT", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 50956.53333333}}, "created": null, "info": {"symbol": "OXTUSDT", "status": "TRADING", "baseAsset": "OXT", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "900000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "50956.53333333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "OXTUSD4", "lowercaseId": "oxtusd4", "symbol": "OXT/USD4", "base": "OXT", "quote": "USD4", "settle": null, "baseId": "OXT", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 682136.82107366}}, "created": null, "info": {"symbol": "OXTUSD4", "status": "BREAK", "baseAsset": "OXT", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "682136.82107366", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ZENUSDT", "lowercaseId": "zenusdt", "symbol": "ZEN/USDT", "base": "ZEN", "quote": "USDT", "settle": null, "baseId": "ZEN", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.01, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 674.69354166}}, "created": null, "info": {"symbol": "ZENUSDT", "status": "TRADING", "baseAsset": "ZEN", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "100000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "674.69354166", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ZENUSD4", "lowercaseId": "zenusd4", "symbol": "ZEN/USD4", "base": "ZEN", "quote": "USD4", "settle": null, "baseId": "ZEN", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 900000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 3922.88797292}}, "created": null, "info": {"symbol": "ZENUSD4", "status": "BREAK", "baseAsset": "ZEN", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "10000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "900000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3922.88797292", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BTCUSDC", "lowercaseId": "btcusdc", "symbol": "BTC/USDC", "base": "BTC", "quote": "USDC", "settle": null, "baseId": "BTC", "quoteId": "USDC", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1e-05, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1e-05, "max": 9000.0}, "price": {"min": 0.01, "max": 1000000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 5.91441425}}, "created": null, "info": {"symbol": "BTCUSDC", "status": "TRADING", "baseAsset": "BTC", "baseAssetPrecision": "8", "quoteAsset": "USDC", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "1000000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00001000", "maxQty": "9000.00000000", "stepSize": "0.00001000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "5.91441425", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ONEBUSD", "lowercaseId": "onebusd", "symbol": "ONE/BUSD", "base": "ONE", "quote": "BUSD", "settle": null, "baseId": "ONE", "quoteId": "BUSD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 786723.0552467}}, "created": null, "info": {"symbol": "ONEBUSD", "status": "BREAK", "baseAsset": "ONE", "baseAssetPrecision": "8", "quoteAsset": "BUSD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "786723.05524670", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "FILUSDT", "lowercaseId": "filusdt", "symbol": "FIL/USDT", "base": "FIL", "quote": "USDT", "settle": null, "baseId": "FIL", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 9222449.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 2615.58379166}}, "created": null, "info": {"symbol": "FILUSDT", "status": "TRADING", "baseAsset": "FIL", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "9222449.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "2615.58379166", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "FILUSD4", "lowercaseId": "filusd4", "symbol": "FIL/USD4", "base": "FIL", "quote": "USD4", "settle": null, "baseId": "FIL", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.0001, "price": 0.01, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.0001, "max": 9222449.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 6149.96227773}}, "created": null, "info": {"symbol": "FILUSD4", "status": "BREAK", "baseAsset": "FIL", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "2", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0100", "maxPrice": "10000.0000", "tickSize": "0.0100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00010000", "maxQty": "9222449.00000000", "stepSize": "0.00010000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "6149.96227773", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "AAVEUSDT", "lowercaseId": "aaveusdt", "symbol": "AAVE/USDT", "base": "AAVE", "quote": "USDT", "settle": null, "baseId": "AAVE", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 900000.0}, "price": {"min": 0.01, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 159.85402083}}, "created": null, "info": {"symbol": "AAVEUSDT", "status": "TRADING", "baseAsset": "AAVE", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "100000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "900000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "159.85402083", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "AAVEUSD4", "lowercaseId": "aaveusd4", "symbol": "AAVE/USD4", "base": "AAVE", "quote": "USD4", "settle": null, "baseId": "AAVE", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.0001, "price": 0.01, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.0001, "max": 900000.0}, "price": {"min": 0.01, "max": 100000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 1584.77607542}}, "created": null, "info": {"symbol": "AAVEUSD4", "status": "BREAK", "baseAsset": "AAVE", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "2", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0100", "maxPrice": "100000.0000", "tickSize": "0.0100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00010000", "maxQty": "900000.00000000", "stepSize": "0.00010000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1584.77607542", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "GRTUSDT", "lowercaseId": "grtusdt", "symbol": "GRT/USDT", "base": "GRT", "quote": "USDT", "settle": null, "baseId": "GRT", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 19780.10833333}}, "created": null, "info": {"symbol": "GRTUSDT", "status": "TRADING", "baseAsset": "GRT", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "900000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "19780.10833333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "GRTUSD4", "lowercaseId": "grtusd4", "symbol": "GRT/USD4", "base": "GRT", "quote": "USD4", "settle": null, "baseId": "GRT", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 638136.67414871}}, "created": null, "info": {"symbol": "GRTUSD4", "status": "BREAK", "baseAsset": "GRT", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "638136.67414871", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SUSHIUSD4", "lowercaseId": "sushiusd4", "symbol": "SUSHI/USD4", "base": "SUSHI", "quote": "USD4", "settle": null, "baseId": "SUSHI", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 90000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 122718.05450208}}, "created": null, "info": {"symbol": "SUSHIUSD4", "status": "BREAK", "baseAsset": "SUSHI", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "10000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "90000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "122718.05450208", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ANKRUSD4", "lowercaseId": "ankrusd4", "symbol": "ANKR/USD4", "base": "ANKR", "quote": "USD4", "settle": null, "baseId": "ANKR", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 4075695.89444058}}, "created": null, "info": {"symbol": "ANKRUSD4", "status": "BREAK", "baseAsset": "ANKR", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "4075695.89444058", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "AMPUSD", "lowercaseId": "ampusd", "symbol": "AMP/USD4", "base": "AMP", "quote": "USD4", "settle": null, "baseId": "AMP", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 9222449.0}}, "created": null, "info": {"symbol": "AMPUSD", "status": "BREAK", "baseAsset": "AMP", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "9222449.00000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SHIBUSDT", "lowercaseId": "shibusdt", "symbol": "SHIB/USDT", "base": "SHIB", "quote": "USDT", "settle": null, "baseId": "SHIB", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-08, "cost": null, "base": 0.01, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 46116860414.0}, "price": {"min": 1e-08, "max": 1.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 11701990467.71}}, "created": null, "info": {"symbol": "SHIBUSDT", "status": "TRADING", "baseAsset": "SHIB", "baseAssetPrecision": "2", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "2", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000001", "maxPrice": "1.00000000", "tickSize": "0.00000001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00", "maxQty": "46116860414.00", "stepSize": "1.00"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00", "maxQty": "11701990467.71", "stepSize": "0.00"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SHIBBUSD", "lowercaseId": "shibbusd", "symbol": "SHIB/BUSD", "base": "SHIB", "quote": "BUSD", "settle": null, "baseId": "SHIB", "quoteId": "BUSD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-08, "cost": null, "base": 0.01, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 46116860414.0}, "price": {"min": 1e-08, "max": 1.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 46116860414.0}}, "created": null, "info": {"symbol": "SHIBBUSD", "status": "BREAK", "baseAsset": "SHIB", "baseAssetPrecision": "2", "quoteAsset": "BUSD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "2", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000001", "maxPrice": "1.00000000", "tickSize": "0.00000001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00", "maxQty": "46116860414.00", "stepSize": "1.00"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00", "maxQty": "46116860414.00", "stepSize": "0.00"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "CRVUSDT", "lowercaseId": "crvusdt", "symbol": "CRV/USDT", "base": "CRV", "quote": "USDT", "settle": null, "baseId": "CRV", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 922327.0}, "price": {"min": 0.0001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 31140.70875}}, "created": null, "info": {"symbol": "CRVUSDT", "status": "TRADING", "baseAsset": "CRV", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "10000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "922327.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "31140.70875000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "CRVUSD4", "lowercaseId": "crvusd4", "symbol": "CRV/USD4", "base": "CRV", "quote": "USD4", "settle": null, "baseId": "CRV", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 922327.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 191361.34013889}}, "created": null, "info": {"symbol": "CRVUSD4", "status": "BREAK", "baseAsset": "CRV", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "10000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "922327.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "191361.34013889", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "AXSUSDT", "lowercaseId": "axsusdt", "symbol": "AXS/USDT", "base": "AXS", "quote": "USDT", "settle": null, "baseId": "AXS", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 466.74429166}}, "created": null, "info": {"symbol": "AXSUSDT", "status": "TRADING", "baseAsset": "AXS", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "466.74429166", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "AXSUSD4", "lowercaseId": "axsusd4", "symbol": "AXS/USD4", "base": "AXS", "quote": "USD4", "settle": null, "baseId": "AXS", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.01, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 3434.86926736}}, "created": null, "info": {"symbol": "AXSUSD4", "status": "BREAK", "baseAsset": "AXS", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "2", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0100", "maxPrice": "1000.0000", "tickSize": "0.0100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3434.86926736", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SOLBTC", "lowercaseId": "solbtc", "symbol": "SOL/BTC", "base": "SOL", "quote": "BTC", "settle": null, "baseId": "SOL", "quoteId": "BTC", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 1e-07, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000000.0}, "price": {"min": 1e-07, "max": 1000.0}, "cost": {"min": 0.0001, "max": null}, "market": {"min": 0.0, "max": 1259.33558333}}, "created": null, "info": {"symbol": "SOLBTC", "status": "TRADING", "baseAsset": "SOL", "baseAssetPrecision": "8", "quoteAsset": "BTC", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000010", "maxPrice": "1000.00000000", "tickSize": "0.00000010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "0.00010000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1259.33558333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "AVAXUSDT", "lowercaseId": "avaxusdt", "symbol": "AVAX/USDT", "base": "AVAX", "quote": "USDT", "settle": null, "baseId": "AVAX", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 1507.81333333}}, "created": null, "info": {"symbol": "AVAXUSDT", "status": "TRADING", "baseAsset": "AVAX", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "10000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1507.81333333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "AVAXUSD4", "lowercaseId": "avaxusd4", "symbol": "AVAX/USD4", "base": "AVAX", "quote": "USD4", "settle": null, "baseId": "AVAX", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 10692.24099375}}, "created": null, "info": {"symbol": "AVAXUSD4", "status": "BREAK", "baseAsset": "AVAX", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "2", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0100", "maxPrice": "10000.0000", "tickSize": "0.0100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "10692.24099375", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "CTSIUSDT", "lowercaseId": "ctsiusdt", "symbol": "CTSI/USDT", "base": "CTSI", "quote": "USDT", "settle": null, "baseId": "CTSI", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 9453.29166666}}, "created": null, "info": {"symbol": "CTSIUSDT", "status": "TRADING", "baseAsset": "CTSI", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "9453.29166666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "CTSIUSD4", "lowercaseId": "ctsiusd4", "symbol": "CTSI/USD4", "base": "CTSI", "quote": "USD4", "settle": null, "baseId": "CTSI", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 297335.32592078}}, "created": null, "info": {"symbol": "CTSIUSD4", "status": "BREAK", "baseAsset": "CTSI", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "297335.32592078", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "DOTUSDT", "lowercaseId": "dotusdt", "symbol": "DOT/USDT", "base": "DOT", "quote": "USDT", "settle": null, "baseId": "DOT", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 9662.940875}}, "created": null, "info": {"symbol": "DOTUSDT", "status": "TRADING", "baseAsset": "DOT", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "9662.94087500", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "DOTUSD4", "lowercaseId": "dotusd4", "symbol": "DOT/USD4", "base": "DOT", "quote": "USD4", "settle": null, "baseId": "DOT", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 40710.69671181}}, "created": null, "info": {"symbol": "DOTUSD4", "status": "BREAK", "baseAsset": "DOT", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "2", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0100", "maxPrice": "10000.0000", "tickSize": "0.0100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "40710.69671181", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "YFIUSDT", "lowercaseId": "yfiusdt", "symbol": "YFI/USDT", "base": "YFI", "quote": "USDT", "settle": null, "baseId": "YFI", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1e-05, "price": 1.0, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1e-05, "max": 9000.0}, "price": {"min": 1.0, "max": 1000000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 0.54802466}}, "created": null, "info": {"symbol": "YFIUSDT", "status": "TRADING", "baseAsset": "YFI", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "1.00000000", "maxPrice": "1000000.00000000", "tickSize": "1.00000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00001000", "maxQty": "9000.00000000", "stepSize": "0.00001000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "0.54802466", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "YFIUSD4", "lowercaseId": "yfiusd4", "symbol": "YFI/USD4", "base": "YFI", "quote": "USD4", "settle": null, "baseId": "YFI", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1e-05, "price": 0.01, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1e-05, "max": 9000.0}, "price": {"min": 0.01, "max": 1000000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 13.13710975}}, "created": null, "info": {"symbol": "YFIUSD4", "status": "BREAK", "baseAsset": "YFI", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "2", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0100", "maxPrice": "1000000.0000", "tickSize": "0.0100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00001000", "maxQty": "9000.00000000", "stepSize": "0.00001000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "13.13710975", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "1INCHUSDT", "lowercaseId": "1inchusdt", "symbol": "1INCH/USDT", "base": "1INCH", "quote": "USDT", "settle": null, "baseId": "1INCH", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 13603.28625}}, "created": null, "info": {"symbol": "1INCHUSDT", "status": "TRADING", "baseAsset": "1INCH", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "13603.28625000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "1INCHUSD4", "lowercaseId": "1inchusd4", "symbol": "1INCH/USD4", "base": "1INCH", "quote": "USD4", "settle": null, "baseId": "1INCH", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 67278.2319319}}, "created": null, "info": {"symbol": "1INCHUSD4", "status": "BREAK", "baseAsset": "1INCH", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "1000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "67278.23193190", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "FTMUSDT", "lowercaseId": "ftmusdt", "symbol": "FTM/USDT", "base": "FTM", "quote": "USDT", "settle": null, "baseId": "FTM", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 66056.5875}}, "created": null, "info": {"symbol": "FTMUSDT", "status": "TRADING", "baseAsset": "FTM", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "66056.58750000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "FTMUSD4", "lowercaseId": "ftmusd4", "symbol": "FTM/USD4", "base": "FTM", "quote": "USD4", "settle": null, "baseId": "FTM", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 365811.70048645}}, "created": null, "info": {"symbol": "FTMUSD4", "status": "BREAK", "baseAsset": "FTM", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "365811.70048645", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "USDCUSDT", "lowercaseId": "usdcusdt", "symbol": "USDC/USDT", "base": "USDC", "quote": "USDT", "settle": null, "baseId": "USDC", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 137394.10416666}}, "created": null, "info": {"symbol": "USDCUSDT", "status": "TRADING", "baseAsset": "USDC", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "1.2", "multiplierDown": "0.8", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "137394.10416666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.05", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.95", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ETHUSDC", "lowercaseId": "ethusdc", "symbol": "ETH/USDC", "base": "ETH", "quote": "USDC", "settle": null, "baseId": "ETH", "quoteId": "USDC", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.0001, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.0001, "max": 90000.0}, "price": {"min": 0.01, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 10.86018875}}, "created": null, "info": {"symbol": "ETHUSDC", "status": "TRADING", "baseAsset": "ETH", "baseAssetPrecision": "8", "quoteAsset": "USDC", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "100000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00010000", "maxQty": "90000.00000000", "stepSize": "0.00010000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "10.86018875", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "USDCBUSD", "lowercaseId": "usdcbusd", "symbol": "USDC/BUSD", "base": "USDC", "quote": "BUSD", "settle": null, "baseId": "USDC", "quoteId": "BUSD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 913751.12184851}}, "created": null, "info": {"symbol": "USDCBUSD", "status": "BREAK", "baseAsset": "USDC", "baseAssetPrecision": "8", "quoteAsset": "BUSD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "1.2", "multiplierDown": "0.8", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "9000000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "913751.12184851", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "MATICUSDT", "lowercaseId": "maticusdt", "symbol": "MATIC/USDT", "base": "MATIC", "quote": "USDT", "settle": null, "baseId": "MATIC", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 59731.01625}}, "created": null, "info": {"symbol": "MATICUSDT", "status": "TRADING", "baseAsset": "MATIC", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "59731.01625000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "MANAUSDT", "lowercaseId": "manausdt", "symbol": "MANA/USDT", "base": "MANA", "quote": "USDT", "settle": null, "baseId": "MANA", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 43211.75933609}}, "created": null, "info": {"symbol": "MANAUSDT", "status": "TRADING", "baseAsset": "MANA", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "900000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "43211.75933609", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "MANABUSD", "lowercaseId": "manabusd", "symbol": "MANA/BUSD", "base": "MANA", "quote": "BUSD", "settle": null, "baseId": "MANA", "quoteId": "BUSD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 219905.74930555}}, "created": null, "info": {"symbol": "MANABUSD", "status": "BREAK", "baseAsset": "MANA", "baseAssetPrecision": "8", "quoteAsset": "BUSD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "900000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "219905.74930555", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ALGOUSDT", "lowercaseId": "algousdt", "symbol": "ALGO/USDT", "base": "ALGO", "quote": "USDT", "settle": null, "baseId": "ALGO", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 922327.0}, "price": {"min": 0.0001, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 87081.6625}}, "created": null, "info": {"symbol": "ALGOUSDT", "status": "TRADING", "baseAsset": "ALGO", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "100000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "922327.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "87081.66250000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ADABUSD", "lowercaseId": "adabusd", "symbol": "ADA/BUSD", "base": "ADA", "quote": "BUSD", "settle": null, "baseId": "ADA", "quoteId": "BUSD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 1141186.88813888}}, "created": null, "info": {"symbol": "ADABUSD", "status": "BREAK", "baseAsset": "ADA", "baseAssetPrecision": "8", "quoteAsset": "BUSD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1141186.88813888", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SOLBUSD", "lowercaseId": "solbusd", "symbol": "SOL/BUSD", "base": "SOL", "quote": "BUSD", "settle": null, "baseId": "SOL", "quoteId": "BUSD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 14342.77578472}}, "created": null, "info": {"symbol": "SOLBUSD", "status": "BREAK", "baseAsset": "SOL", "baseAssetPrecision": "8", "quoteAsset": "BUSD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "10000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "14342.77578472", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LINKUSDT", "lowercaseId": "linkusdt", "symbol": "LINK/USDT", "base": "LINK", "quote": "USDT", "settle": null, "baseId": "LINK", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 1953.42141666}}, "created": null, "info": {"symbol": "LINKUSDT", "status": "TRADING", "baseAsset": "LINK", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "10000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1953.42141666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "EOSUSDT", "lowercaseId": "eosusdt", "symbol": "EOS/USDT", "base": "EOS", "quote": "USDT", "settle": null, "baseId": "EOS", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 23480.77666666}}, "created": null, "info": {"symbol": "EOSUSDT", "status": "TRADING", "baseAsset": "EOS", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "23480.77666666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ZECUSDT", "lowercaseId": "zecusdt", "symbol": "ZEC/USDT", "base": "ZEC", "quote": "USDT", "settle": null, "baseId": "ZEC", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 90000.0}, "price": {"min": 0.01, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 122.79282083}}, "created": null, "info": {"symbol": "ZECUSDT", "status": "TRADING", "baseAsset": "ZEC", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "100000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "90000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "122.79282083", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ENJUSDT", "lowercaseId": "enjusdt", "symbol": "ENJ/USDT", "base": "ENJ", "quote": "USDT", "settle": null, "baseId": "ENJ", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 41165.06833333}}, "created": null, "info": {"symbol": "ENJUSDT", "status": "TRADING", "baseAsset": "ENJ", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "41165.06833333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "NEARUSDT", "lowercaseId": "nearusdt", "symbol": "NEAR/USDT", "base": "NEAR", "quote": "USDT", "settle": null, "baseId": "NEAR", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 8163.41506276}}, "created": null, "info": {"symbol": "NEARUSDT", "status": "TRADING", "baseAsset": "NEAR", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "8163.41506276", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "NEARBUSD", "lowercaseId": "nearbusd", "symbol": "NEAR/BUSD", "base": "NEAR", "quote": "BUSD", "settle": null, "baseId": "NEAR", "quoteId": "BUSD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 45798.40781944}}, "created": null, "info": {"symbol": "NEARBUSD", "status": "BREAK", "baseAsset": "NEAR", "baseAssetPrecision": "8", "quoteAsset": "BUSD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "45798.40781944", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "NEARUSD4", "lowercaseId": "nearusd4", "symbol": "NEAR/USD4", "base": "NEAR", "quote": "USD4", "settle": null, "baseId": "NEAR", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 219157.5346875}}, "created": null, "info": {"symbol": "NEARUSD4", "status": "BREAK", "baseAsset": "NEAR", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "1000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "219157.53468750", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "OMGUSDT", "lowercaseId": "omgusdt", "symbol": "OMG/USDT", "base": "OMG", "quote": "USDT", "settle": null, "baseId": "OMG", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 13095.65}}, "created": null, "info": {"symbol": "OMGUSDT", "status": "BREAK", "baseAsset": "OMG", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "13095.65000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SUSHIUSDT", "lowercaseId": "sushiusdt", "symbol": "SUSHI/USDT", "base": "SUSHI", "quote": "USDT", "settle": null, "baseId": "SUSHI", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 90000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 3841.22041666}}, "created": null, "info": {"symbol": "SUSHIUSDT", "status": "TRADING", "baseAsset": "SUSHI", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "90000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3841.22041666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LRCUSDT", "lowercaseId": "lrcusdt", "symbol": "LRC/USDT", "base": "LRC", "quote": "USDT", "settle": null, "baseId": "LRC", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 14089.56666666}}, "created": null, "info": {"symbol": "LRCUSDT", "status": "TRADING", "baseAsset": "LRC", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "14089.56666666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LRCUSD4", "lowercaseId": "lrcusd4", "symbol": "LRC/USD4", "base": "LRC", "quote": "USD4", "settle": null, "baseId": "LRC", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 569490.2625}}, "created": null, "info": {"symbol": "LRCUSD4", "status": "BREAK", "baseAsset": "LRC", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "569490.26250000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LRCBTC", "lowercaseId": "lrcbtc", "symbol": "LRC/BTC", "base": "LRC", "quote": "BTC", "settle": null, "baseId": "LRC", "quoteId": "BTC", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-08, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 90000000.0}, "price": {"min": 1e-08, "max": 1000.0}, "cost": {"min": 0.0001, "max": null}, "market": {"min": 0.0, "max": 244582.64652777}}, "created": null, "info": {"symbol": "LRCBTC", "status": "BREAK", "baseAsset": "LRC", "baseAssetPrecision": "8", "quoteAsset": "BTC", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000001", "maxPrice": "1000.00000000", "tickSize": "0.00000001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "90000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "0.00010000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "244582.64652777", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "KSHIBUSD4", "lowercaseId": "kshibusd4", "symbol": "KSHIB/USD4", "base": "KSHIB", "quote": "USD4", "settle": null, "baseId": "KSHIB", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 0.0001, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 46116860.0}, "price": {"min": 0.0001, "max": 3.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 46116860.0}}, "created": null, "info": {"symbol": "KSHIBUSD4", "status": "BREAK", "baseAsset": "KSHIB", "baseAssetPrecision": "4", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "4", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "3.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.0000", "maxQty": "46116860.0000", "stepSize": "1.0000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.0000", "maxQty": "46116860.0000", "stepSize": "0.0000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LPTUSDT", "lowercaseId": "lptusdt", "symbol": "LPT/USDT", "base": "LPT", "quote": "USDT", "settle": null, "baseId": "LPT", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 9222440.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 460.64366666}}, "created": null, "info": {"symbol": "LPTUSDT", "status": "TRADING", "baseAsset": "LPT", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "9222440.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "460.64366666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LPTBUSD", "lowercaseId": "lptbusd", "symbol": "LPT/BUSD", "base": "LPT", "quote": "BUSD", "settle": null, "baseId": "LPT", "quoteId": "BUSD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 9222440.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 6853.50953472}}, "created": null, "info": {"symbol": "LPTBUSD", "status": "BREAK", "baseAsset": "LPT", "baseAssetPrecision": "8", "quoteAsset": "BUSD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "10000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "9222440.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "6853.50953472", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LPTUSD4", "lowercaseId": "lptusd4", "symbol": "LPT/USD4", "base": "LPT", "quote": "USD4", "settle": null, "baseId": "LPT", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 9222440.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 6853.50953472}}, "created": null, "info": {"symbol": "LPTUSD4", "status": "BREAK", "baseAsset": "LPT", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "2", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0100", "maxPrice": "10000.0000", "tickSize": "0.0100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "9222440.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "6853.50953472", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "POLYUSDT", "lowercaseId": "polyusdt", "symbol": "POLY/USDT", "base": "POLY", "quote": "USDT", "settle": null, "baseId": "POLY", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 913205152.0}, "price": {"min": 0.0001, "max": 100.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 290408.03145833}}, "created": null, "info": {"symbol": "POLYUSDT", "status": "BREAK", "baseAsset": "POLY", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "100.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "913205152.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "290408.03145833", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "POLYBUSD", "lowercaseId": "polybusd", "symbol": "POLY/BUSD", "base": "POLY", "quote": "BUSD", "settle": null, "baseId": "POLY", "quoteId": "BUSD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 913205152.0}, "price": {"min": 0.0001, "max": 100.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 290408.03145833}}, "created": null, "info": {"symbol": "POLYBUSD", "status": "BREAK", "baseAsset": "POLY", "baseAssetPrecision": "8", "quoteAsset": "BUSD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "100.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "913205152.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "290408.03145833", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "POLYUSD", "lowercaseId": "polyusd", "symbol": "POLY/USD4", "base": "POLY", "quote": "USD4", "settle": null, "baseId": "POLY", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 913205152.0}, "price": {"min": 0.0001, "max": 100.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 290408.03145833}}, "created": null, "info": {"symbol": "POLYUSD", "status": "BREAK", "baseAsset": "POLY", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "100.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "913205152.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "290408.03145833", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "POLYBTC", "lowercaseId": "polybtc", "symbol": "POLY/BTC", "base": "POLY", "quote": "BTC", "settle": null, "baseId": "POLY", "quoteId": "BTC", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-08, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 90000000.0}, "price": {"min": 1e-08, "max": 1000.0}, "cost": {"min": 0.0001, "max": null}, "market": {"min": 0.0, "max": 135640.92222222}}, "created": null, "info": {"symbol": "POLYBTC", "status": "BREAK", "baseAsset": "POLY", "baseAssetPrecision": "8", "quoteAsset": "BTC", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000001", "maxPrice": "1000.00000000", "tickSize": "0.00000001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "90000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "0.00010000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "135640.92222222", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "MATICBTC", "lowercaseId": "maticbtc", "symbol": "MATIC/BTC", "base": "MATIC", "quote": "BTC", "settle": null, "baseId": "MATIC", "quoteId": "BTC", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-08, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 90000000.0}, "price": {"min": 1e-08, "max": 1000.0}, "cost": {"min": 0.0001, "max": null}, "market": {"min": 0.0, "max": 14357.96291666}}, "created": null, "info": {"symbol": "MATICBTC", "status": "TRADING", "baseAsset": "MATIC", "baseAssetPrecision": "8", "quoteAsset": "BTC", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000001", "maxPrice": "1000.00000000", "tickSize": "0.00000001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "90000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "0.00010000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "14357.96291666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "DOTBTC", "lowercaseId": "dotbtc", "symbol": "DOT/BTC", "base": "DOT", "quote": "BTC", "settle": null, "baseId": "DOT", "quoteId": "BTC", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 1e-07, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000000.0}, "price": {"min": 1e-07, "max": 1000.0}, "cost": {"min": 0.0001, "max": null}, "market": {"min": 0.0, "max": 47223.22929861}}, "created": null, "info": {"symbol": "DOTBTC", "status": "BREAK", "baseAsset": "DOT", "baseAssetPrecision": "8", "quoteAsset": "BTC", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000010", "maxPrice": "1000.00000000", "tickSize": "0.00000010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "0.00010000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "47223.22929861", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "NMRUSDT", "lowercaseId": "nmrusdt", "symbol": "NMR/USDT", "base": "NMR", "quote": "USDT", "settle": null, "baseId": "NMR", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 89.11916666}}, "created": null, "info": {"symbol": "NMRUSDT", "status": "TRADING", "baseAsset": "NMR", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "10000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "89.11916666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "NMRUSD4", "lowercaseId": "nmrusd4", "symbol": "NMR/USD4", "base": "NMR", "quote": "USD4", "settle": null, "baseId": "NMR", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 4802.3121721}}, "created": null, "info": {"symbol": "NMRUSD4", "status": "BREAK", "baseAsset": "NMR", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "2", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0100", "maxPrice": "10000.0000", "tickSize": "0.0100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "4802.31217210", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SLPUSDT", "lowercaseId": "slpusdt", "symbol": "SLP/USDT", "base": "SLP", "quote": "USDT", "settle": null, "baseId": "SLP", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-06, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 90000000.0}, "price": {"min": 1e-06, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 636919.9875}}, "created": null, "info": {"symbol": "SLPUSDT", "status": "TRADING", "baseAsset": "SLP", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000100", "maxPrice": "1000.00000000", "tickSize": "0.00000100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "90000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "636919.98750000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SLPUSD4", "lowercaseId": "slpusd4", "symbol": "SLP/USD4", "base": "SLP", "quote": "USD4", "settle": null, "baseId": "SLP", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 90000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 92141578.0}}, "created": null, "info": {"symbol": "SLPUSD4", "status": "BREAK", "baseAsset": "SLP", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "90000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "92141578.00000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ANTUSDT", "lowercaseId": "antusdt", "symbol": "ANT/USDT", "base": "ANT", "quote": "USDT", "settle": null, "baseId": "ANT", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 191.03916666}}, "created": null, "info": {"symbol": "ANTUSDT", "status": "BREAK", "baseAsset": "ANT", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "191.03916666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ANTUSD4", "lowercaseId": "antusd4", "symbol": "ANT/USD4", "base": "ANT", "quote": "USD4", "settle": null, "baseId": "ANT", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 90347.52763888}}, "created": null, "info": {"symbol": "ANTUSD4", "status": "BREAK", "baseAsset": "ANT", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "1000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "90347.52763888", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "XNOUSD4", "lowercaseId": "xnousd4", "symbol": "XNO/USD4", "base": "XNO", "quote": "USD4", "settle": null, "baseId": "XNO", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 36240.93046527}}, "created": null, "info": {"symbol": "XNOUSD4", "status": "BREAK", "baseAsset": "XNO", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "1000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "36240.93046527", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "CHZUSDT", "lowercaseId": "chzusdt", "symbol": "CHZ/USDT", "base": "CHZ", "quote": "USDT", "settle": null, "baseId": "CHZ", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 22343.27615062}}, "created": null, "info": {"symbol": "CHZUSDT", "status": "TRADING", "baseAsset": "CHZ", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "22343.27615062", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "CHZUSD4", "lowercaseId": "chzusd4", "symbol": "CHZ/USD4", "base": "CHZ", "quote": "USD4", "settle": null, "baseId": "CHZ", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 2767789.21666666}}, "created": null, "info": {"symbol": "CHZUSD4", "status": "BREAK", "baseAsset": "CHZ", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "2767789.21666666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "OGNUSDT", "lowercaseId": "ognusdt", "symbol": "OGN/USDT", "base": "OGN", "quote": "USDT", "settle": null, "baseId": "OGN", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 14554.37916666}}, "created": null, "info": {"symbol": "OGNUSDT", "status": "TRADING", "baseAsset": "OGN", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "900000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "14554.37916666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "OGNUSD4", "lowercaseId": "ognusd4", "symbol": "OGN/USD4", "base": "OGN", "quote": "USD4", "settle": null, "baseId": "OGN", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 569534.45726198}}, "created": null, "info": {"symbol": "OGNUSD4", "status": "BREAK", "baseAsset": "OGN", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "900000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "569534.45726198", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "GALAUSDT", "lowercaseId": "galausdt", "symbol": "GALA/USDT", "base": "GALA", "quote": "USDT", "settle": null, "baseId": "GALA", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 1132111.10833333}}, "created": null, "info": {"symbol": "GALAUSDT", "status": "TRADING", "baseAsset": "GALA", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1132111.10833333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "GALAUSD4", "lowercaseId": "galausd4", "symbol": "GALA/USD4", "base": "GALA", "quote": "USD4", "settle": null, "baseId": "GALA", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 10519306.87074357}}, "created": null, "info": {"symbol": "GALAUSD4", "status": "BREAK", "baseAsset": "GALA", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "10519306.87074357", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "TLMUSDT", "lowercaseId": "tlmusdt", "symbol": "TLM/USDT", "base": "TLM", "quote": "USDT", "settle": null, "baseId": "TLM", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 900000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 30668.90833333}}, "created": null, "info": {"symbol": "TLMUSDT", "status": "TRADING", "baseAsset": "TLM", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "900000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "30668.90833333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "TLMUSD4", "lowercaseId": "tlmusd4", "symbol": "TLM/USD4", "base": "TLM", "quote": "USD4", "settle": null, "baseId": "TLM", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 4720997.96455872}}, "created": null, "info": {"symbol": "TLMUSD4", "status": "BREAK", "baseAsset": "TLM", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "900000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "4720997.96455872", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SNXUSDT", "lowercaseId": "snxusdt", "symbol": "SNX/USDT", "base": "SNX", "quote": "USDT", "settle": null, "baseId": "SNX", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 90000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 866.76333333}}, "created": null, "info": {"symbol": "SNXUSDT", "status": "TRADING", "baseAsset": "SNX", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "90000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "866.76333333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SNXUSD4", "lowercaseId": "snxusd4", "symbol": "SNX/USD4", "base": "SNX", "quote": "USD4", "settle": null, "baseId": "SNX", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 90000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 39955.67692842}}, "created": null, "info": {"symbol": "SNXUSD4", "status": "BREAK", "baseAsset": "SNX", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "10000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "90000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "39955.67692842", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "AUDIOUSDT", "lowercaseId": "audiousdt", "symbol": "AUDIO/USDT", "base": "AUDIO", "quote": "USDT", "settle": null, "baseId": "AUDIO", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 20796.29375}}, "created": null, "info": {"symbol": "AUDIOUSDT", "status": "TRADING", "baseAsset": "AUDIO", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "20796.29375000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "AUDIOUSD4", "lowercaseId": "audiousd4", "symbol": "AUDIO/USD4", "base": "AUDIO", "quote": "USD4", "settle": null, "baseId": "AUDIO", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 241109.39353717}}, "created": null, "info": {"symbol": "AUDIOUSD4", "status": "BREAK", "baseAsset": "AUDIO", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "1000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "241109.39353717", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ENSUSDT", "lowercaseId": "ensusdt", "symbol": "ENS/USDT", "base": "ENS", "quote": "USDT", "settle": null, "baseId": "ENS", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 9222449.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 560.32016666}}, "created": null, "info": {"symbol": "ENSUSDT", "status": "TRADING", "baseAsset": "ENS", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "10000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "9222449.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "560.32016666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "MANABTC", "lowercaseId": "manabtc", "symbol": "MANA/BTC", "base": "MANA", "quote": "BTC", "settle": null, "baseId": "MANA", "quoteId": "BTC", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-08, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 90000000.0}, "price": {"min": 1e-08, "max": 1000.0}, "cost": {"min": 0.0001, "max": null}, "market": {"min": 0.0, "max": 149011.94788047}}, "created": null, "info": {"symbol": "MANABTC", "status": "BREAK", "baseAsset": "MANA", "baseAssetPrecision": "8", "quoteAsset": "BTC", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000001", "maxPrice": "1000.00000000", "tickSize": "0.00000001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "90000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "0.00010000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "149011.94788047", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ATOMBTC", "lowercaseId": "atombtc", "symbol": "ATOM/BTC", "base": "ATOM", "quote": "BTC", "settle": null, "baseId": "ATOM", "quoteId": "BTC", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 1e-07, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000000.0}, "price": {"min": 1e-07, "max": 1000.0}, "cost": {"min": 0.0001, "max": null}, "market": {"min": 0.0, "max": 27114.50637751}}, "created": null, "info": {"symbol": "ATOMBTC", "status": "BREAK", "baseAsset": "ATOM", "baseAssetPrecision": "8", "quoteAsset": "BTC", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000010", "maxPrice": "1000.00000000", "tickSize": "0.00000010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "0.00010000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "27114.50637751", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "AVAXBTC", "lowercaseId": "avaxbtc", "symbol": "AVAX/BTC", "base": "AVAX", "quote": "BTC", "settle": null, "baseId": "AVAX", "quoteId": "BTC", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 1e-07, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000000.0}, "price": {"min": 1e-07, "max": 1000.0}, "cost": {"min": 0.0001, "max": null}, "market": {"min": 0.0, "max": 505.93258333}}, "created": null, "info": {"symbol": "AVAXBTC", "status": "TRADING", "baseAsset": "AVAX", "baseAssetPrecision": "8", "quoteAsset": "BTC", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000010", "maxPrice": "1000.00000000", "tickSize": "0.00000010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "0.00010000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "505.93258333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "WBTCBTC", "lowercaseId": "wbtcbtc", "symbol": "WBTC/BTC", "base": "WBTC", "quote": "BTC", "settle": null, "baseId": "WBTC", "quoteId": "BTC", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1e-05, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1e-05, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 0.0001, "max": null}, "market": {"min": 0.0, "max": 0.02873}}, "created": null, "info": {"symbol": "WBTCBTC", "status": "TRADING", "baseAsset": "WBTC", "baseAssetPrecision": "8", "quoteAsset": "BTC", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00001000", "maxQty": "92141578.00000000", "stepSize": "0.00001000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "0.00010000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "0.02873000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "REQUSDT", "lowercaseId": "requsdt", "symbol": "REQ/USDT", "base": "REQ", "quote": "USDT", "settle": null, "baseId": "REQ", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 16609.92083333}}, "created": null, "info": {"symbol": "REQUSDT", "status": "TRADING", "baseAsset": "REQ", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "16609.92083333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "REQUSD4", "lowercaseId": "requsd4", "symbol": "REQ/USD4", "base": "REQ", "quote": "USD4", "settle": null, "baseId": "REQ", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 933555.52605976}}, "created": null, "info": {"symbol": "REQUSD4", "status": "BREAK", "baseAsset": "REQ", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "933555.52605976", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "APEUSDT", "lowercaseId": "apeusdt", "symbol": "APE/USDT", "base": "APE", "quote": "USDT", "settle": null, "baseId": "APE", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 2940.49254166}}, "created": null, "info": {"symbol": "APEUSDT", "status": "TRADING", "baseAsset": "APE", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "2940.49254166", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "APEUSD4", "lowercaseId": "apeusd4", "symbol": "APE/USD4", "base": "APE", "quote": "USD4", "settle": null, "baseId": "APE", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 92141578.0}}, "created": null, "info": {"symbol": "APEUSD4", "status": "BREAK", "baseAsset": "APE", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "92141578.00000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "FLUXUSDT", "lowercaseId": "fluxusdt", "symbol": "FLUX/USDT", "base": "FLUX", "quote": "USDT", "settle": null, "baseId": "FLUX", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 7366.94179166}}, "created": null, "info": {"symbol": "FLUXUSDT", "status": "TRADING", "baseAsset": "FLUX", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "7366.94179166", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "FLUXUSD4", "lowercaseId": "fluxusd4", "symbol": "FLUX/USD4", "base": "FLUX", "quote": "USD4", "settle": null, "baseId": "FLUX", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 175409.69921473}}, "created": null, "info": {"symbol": "FLUXUSD4", "status": "BREAK", "baseAsset": "FLUX", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "1000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "175409.69921473", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "TRXBTC", "lowercaseId": "trxbtc", "symbol": "TRX/BTC", "base": "TRX", "quote": "BTC", "settle": null, "baseId": "TRX", "quoteId": "BTC", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-08, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 90000000.0}, "price": {"min": 1e-08, "max": 1000.0}, "cost": {"min": 0.0001, "max": null}, "market": {"min": 0.0, "max": 26990722.39749826}}, "created": null, "info": {"symbol": "TRXBTC", "status": "BREAK", "baseAsset": "TRX", "baseAssetPrecision": "8", "quoteAsset": "BTC", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000001", "maxPrice": "1000.00000000", "tickSize": "0.00000001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "90000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "0.00010000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "26990722.39749826", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "TRXBUSD", "lowercaseId": "trxbusd", "symbol": "TRX/BUSD", "base": "TRX", "quote": "BUSD", "settle": null, "baseId": "TRX", "quoteId": "BUSD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 3197272.47527778}}, "created": null, "info": {"symbol": "TRXBUSD", "status": "BREAK", "baseAsset": "TRX", "baseAssetPrecision": "8", "quoteAsset": "BUSD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3197272.47527778", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "TRXUSDT", "lowercaseId": "trxusdt", "symbol": "TRX/USDT", "base": "TRX", "quote": "USDT", "settle": null, "baseId": "TRX", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 20995303.38346073}}, "created": null, "info": {"symbol": "TRXUSDT", "status": "BREAK", "baseAsset": "TRX", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "20995303.38346073", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "TRXUSD4", "lowercaseId": "trxusd4", "symbol": "TRX/USD4", "base": "TRX", "quote": "USD4", "settle": null, "baseId": "TRX", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 20995303.38346073}}, "created": null, "info": {"symbol": "TRXUSD4", "status": "BREAK", "baseAsset": "TRX", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "20995303.38346073", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "COTIUSDT", "lowercaseId": "cotiusdt", "symbol": "COTI/USDT", "base": "COTI", "quote": "USDT", "settle": null, "baseId": "COTI", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 15693.79668049}}, "created": null, "info": {"symbol": "COTIUSDT", "status": "TRADING", "baseAsset": "COTI", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "15693.79668049", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "COTIUSD4", "lowercaseId": "cotiusd4", "symbol": "COTI/USD4", "base": "COTI", "quote": "USD4", "settle": null, "baseId": "COTI", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 1820701.29186935}}, "created": null, "info": {"symbol": "COTIUSD4", "status": "BREAK", "baseAsset": "COTI", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1820701.29186935", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "VOXELUSDT", "lowercaseId": "voxelusdt", "symbol": "VOXEL/USDT", "base": "VOXEL", "quote": "USDT", "settle": null, "baseId": "VOXEL", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 14241.68}}, "created": null, "info": {"symbol": "VOXELUSDT", "status": "TRADING", "baseAsset": "VOXEL", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "14241.68000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "VOXELUSD4", "lowercaseId": "voxelusd4", "symbol": "VOXEL/USD4", "base": "VOXEL", "quote": "USD4", "settle": null, "baseId": "VOXEL", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 351360.30125086}}, "created": null, "info": {"symbol": "VOXELUSD4", "status": "BREAK", "baseAsset": "VOXEL", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "351360.30125086", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "RLCUSDT", "lowercaseId": "rlcusdt", "symbol": "RLC/USDT", "base": "RLC", "quote": "USDT", "settle": null, "baseId": "RLC", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 570.32875}}, "created": null, "info": {"symbol": "RLCUSDT", "status": "TRADING", "baseAsset": "RLC", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "570.32875000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "RLCUSD4", "lowercaseId": "rlcusd4", "symbol": "RLC/USD4", "base": "RLC", "quote": "USD4", "settle": null, "baseId": "RLC", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 53872.75618055}}, "created": null, "info": {"symbol": "RLCUSD4", "status": "BREAK", "baseAsset": "RLC", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "1000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "53872.75618055", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "USTUSDT", "lowercaseId": "ustusdt", "symbol": "UST/USDT", "base": "UST", "quote": "USDT", "settle": null, "baseId": "UST", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-06, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 1e-06, "max": 1.3}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 3338230.24600416}}, "created": null, "info": {"symbol": "USTUSDT", "status": "BREAK", "baseAsset": "UST", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000100", "maxPrice": "1.30000000", "tickSize": "0.00000100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "1.2", "multiplierDown": "0.8", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3338230.24600416", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "USTUSD", "lowercaseId": "ustusd", "symbol": "UST/USD4", "base": "UST", "quote": "USD4", "settle": null, "baseId": "UST", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1.3}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 3338230.24600416}}, "created": null, "info": {"symbol": "USTUSD", "status": "BREAK", "baseAsset": "UST", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1.3000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "1.2", "multiplierDown": "0.8", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3338230.24600416", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BICOUSDT", "lowercaseId": "bicousdt", "symbol": "BICO/USDT", "base": "BICO", "quote": "USDT", "settle": null, "baseId": "BICO", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 3862.680625}}, "created": null, "info": {"symbol": "BICOUSDT", "status": "TRADING", "baseAsset": "BICO", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3862.68062500", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BICOUSD4", "lowercaseId": "bicousd4", "symbol": "BICO/USD4", "base": "BICO", "quote": "USD4", "settle": null, "baseId": "BICO", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 146093.6861779}}, "created": null, "info": {"symbol": "BICOUSD4", "status": "BREAK", "baseAsset": "BICO", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "1000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "146093.68617790", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "API3USDT", "lowercaseId": "api3usdt", "symbol": "API3/USDT", "base": "API3", "quote": "USDT", "settle": null, "baseId": "API3", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 1314.47275}}, "created": null, "info": {"symbol": "API3USDT", "status": "TRADING", "baseAsset": "API3", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1314.47275000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "API3USD4", "lowercaseId": "api3usd4", "symbol": "API3/USD4", "base": "API3", "quote": "USD4", "settle": null, "baseId": "API3", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 94155.98463194}}, "created": null, "info": {"symbol": "API3USD4", "status": "BREAK", "baseAsset": "API3", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "1000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "94155.98463194", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ENSUSD4", "lowercaseId": "ensusd4", "symbol": "ENS/USD4", "base": "ENS", "quote": "USD4", "settle": null, "baseId": "ENS", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 9222449.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 33626.31374565}}, "created": null, "info": {"symbol": "ENSUSD4", "status": "BREAK", "baseAsset": "ENS", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "2", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0100", "maxPrice": "10000.0000", "tickSize": "0.0100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "9222449.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "33626.31374565", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BTCUST", "lowercaseId": "btcust", "symbol": "BTC/UST", "base": "BTC", "quote": "UST", "settle": null, "baseId": "BTC", "quoteId": "UST", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1e-05, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1e-05, "max": 9000.0}, "price": {"min": 0.01, "max": 1000000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 9000.0}}, "created": null, "info": {"symbol": "BTCUST", "status": "BREAK", "baseAsset": "BTC", "baseAssetPrecision": "8", "quoteAsset": "UST", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "1000000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00001000", "maxQty": "9000.00000000", "stepSize": "0.00001000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "9000.00000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BNTUSDT", "lowercaseId": "bntusdt", "symbol": "BNT/USDT", "base": "BNT", "quote": "USDT", "settle": null, "baseId": "BNT", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 4666.04916666}}, "created": null, "info": {"symbol": "BNTUSDT", "status": "TRADING", "baseAsset": "BNT", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "4666.04916666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BNTUSD4", "lowercaseId": "bntusd4", "symbol": "BNT/USD4", "base": "BNT", "quote": "USD4", "settle": null, "baseId": "BNT", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 34504.70986796}}, "created": null, "info": {"symbol": "BNTUSD4", "status": "BREAK", "baseAsset": "BNT", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "1000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "34504.70986796", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "IMXUSDT", "lowercaseId": "imxusdt", "symbol": "IMX/USDT", "base": "IMX", "quote": "USDT", "settle": null, "baseId": "IMX", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 3820.9705}}, "created": null, "info": {"symbol": "IMXUSDT", "status": "TRADING", "baseAsset": "IMX", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3820.97050000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "IMXUSD4", "lowercaseId": "imxusd4", "symbol": "IMX/USD4", "base": "IMX", "quote": "USD4", "settle": null, "baseId": "IMX", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 273678.31160416}}, "created": null, "info": {"symbol": "IMXUSD4", "status": "BREAK", "baseAsset": "IMX", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "1000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "273678.31160416", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SPELLUSDT", "lowercaseId": "spellusdt", "symbol": "SPELL/USDT", "base": "SPELL", "quote": "USDT", "settle": null, "baseId": "SPELL", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-07, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 1e-07, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 63442841.67060458}}, "created": null, "info": {"symbol": "SPELLUSDT", "status": "BREAK", "baseAsset": "SPELL", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000010", "maxPrice": "1000.00000000", "tickSize": "0.00000010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "63442841.67060458", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SPELLUSD4", "lowercaseId": "spellusd4", "symbol": "SPELL/USD4", "base": "SPELL", "quote": "USD4", "settle": null, "baseId": "SPELL", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 63442841.67060458}}, "created": null, "info": {"symbol": "SPELLUSD4", "status": "BREAK", "baseAsset": "SPELL", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "63442841.67060458", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "JASMYUSDT", "lowercaseId": "jasmyusdt", "symbol": "JASMY/USDT", "base": "JASMY", "quote": "USDT", "settle": null, "baseId": "JASMY", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-06, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 1e-06, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 92141578.0}}, "created": null, "info": {"symbol": "JASMYUSDT", "status": "BREAK", "baseAsset": "JASMY", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000100", "maxPrice": "1000.00000000", "tickSize": "0.00000100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "92141578.00000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "JASMYUSD4", "lowercaseId": "jasmyusd4", "symbol": "JASMY/USD4", "base": "JASMY", "quote": "USD4", "settle": null, "baseId": "JASMY", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 92141578.0}}, "created": null, "info": {"symbol": "JASMYUSD4", "status": "BREAK", "baseAsset": "JASMY", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "92141578.00000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "FLOWUSDT", "lowercaseId": "flowusdt", "symbol": "FLOW/USDT", "base": "FLOW", "quote": "USDT", "settle": null, "baseId": "FLOW", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 922327.0}, "price": {"min": 0.001, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 6096.85070539}}, "created": null, "info": {"symbol": "FLOWUSDT", "status": "TRADING", "baseAsset": "FLOW", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "100000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "922327.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "6096.85070539", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "FLOWUSD4", "lowercaseId": "flowusd4", "symbol": "FLOW/USD4", "base": "FLOW", "quote": "USD4", "settle": null, "baseId": "FLOW", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 922327.0}, "price": {"min": 0.01, "max": 100000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 94183.15727966}}, "created": null, "info": {"symbol": "FLOWUSD4", "status": "BREAK", "baseAsset": "FLOW", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "2", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0100", "maxPrice": "100000.0000", "tickSize": "0.0100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "922327.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "94183.15727966", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "GTCUSDT", "lowercaseId": "gtcusdt", "symbol": "GTC/USDT", "base": "GTC", "quote": "USDT", "settle": null, "baseId": "GTC", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9222449.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 2294.88125}}, "created": null, "info": {"symbol": "GTCUSDT", "status": "TRADING", "baseAsset": "GTC", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9222449.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "2294.88125000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "GTCUSD4", "lowercaseId": "gtcusd4", "symbol": "GTC/USD4", "base": "GTC", "quote": "USD4", "settle": null, "baseId": "GTC", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9222449.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 23724.14420541}}, "created": null, "info": {"symbol": "GTCUSD4", "status": "BREAK", "baseAsset": "GTC", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "10000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9222449.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "23724.14420541", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "THETAUSDT", "lowercaseId": "thetausdt", "symbol": "THETA/USDT", "base": "THETA", "quote": "USDT", "settle": null, "baseId": "THETA", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 90000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 11108.63791666}}, "created": null, "info": {"symbol": "THETAUSDT", "status": "TRADING", "baseAsset": "THETA", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "90000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "11108.63791666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "THETAUSD4", "lowercaseId": "thetausd4", "symbol": "THETA/USD4", "base": "THETA", "quote": "USD4", "settle": null, "baseId": "THETA", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 90000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 158385.32960389}}, "created": null, "info": {"symbol": "THETAUSD4", "status": "BREAK", "baseAsset": "THETA", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "10000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "90000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "158385.32960389", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "TFUELUSDT", "lowercaseId": "tfuelusdt", "symbol": "TFUEL/USDT", "base": "TFUEL", "quote": "USDT", "settle": null, "baseId": "TFUEL", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 38784.30833333}}, "created": null, "info": {"symbol": "TFUELUSDT", "status": "TRADING", "baseAsset": "TFUEL", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "38784.30833333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "TFUELUSD4", "lowercaseId": "tfuelusd4", "symbol": "TFUEL/USD4", "base": "TFUEL", "quote": "USD4", "settle": null, "baseId": "TFUEL", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 827041.82861111}}, "created": null, "info": {"symbol": "TFUELUSD4", "status": "BREAK", "baseAsset": "TFUEL", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "827041.82861111", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "OCEANUSDT", "lowercaseId": "oceanusdt", "symbol": "OCEAN/USDT", "base": "OCEAN", "quote": "USDT", "settle": null, "baseId": "OCEAN", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 2376.0}}, "created": null, "info": {"symbol": "OCEANUSDT", "status": "TRADING", "baseAsset": "OCEAN", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "900000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "2376.00000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "OCEANUSD4", "lowercaseId": "oceanusd4", "symbol": "OCEAN/USD4", "base": "OCEAN", "quote": "USD4", "settle": null, "baseId": "OCEAN", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 145997.2870229}}, "created": null, "info": {"symbol": "OCEANUSD4", "status": "BREAK", "baseAsset": "OCEAN", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "900000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "145997.28702290", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LAZIOUSDT", "lowercaseId": "laziousdt", "symbol": "LAZIO/USDT", "base": "LAZIO", "quote": "USDT", "settle": null, "baseId": "LAZIO", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 249.06}}, "created": null, "info": {"symbol": "LAZIOUSDT", "status": "TRADING", "baseAsset": "LAZIO", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "249.06000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LAZIOUSD4", "lowercaseId": "laziousd4", "symbol": "LAZIO/USD4", "base": "LAZIO", "quote": "USD4", "settle": null, "baseId": "LAZIO", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 30354.65209027}}, "created": null, "info": {"symbol": "LAZIOUSD4", "status": "BREAK", "baseAsset": "LAZIO", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "30354.65209027", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SANTOSUSDT", "lowercaseId": "santosusdt", "symbol": "SANTOS/USDT", "base": "SANTOS", "quote": "USDT", "settle": null, "baseId": "SANTOS", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 436.45766666}}, "created": null, "info": {"symbol": "SANTOSUSDT", "status": "TRADING", "baseAsset": "SANTOS", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "436.45766666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SANTOSUSD4", "lowercaseId": "santosusd4", "symbol": "SANTOS/USD4", "base": "SANTOS", "quote": "USD4", "settle": null, "baseId": "SANTOS", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 26075.60693268}}, "created": null, "info": {"symbol": "SANTOSUSD4", "status": "BREAK", "baseAsset": "SANTOS", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "1000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "26075.60693268", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ALPINEUSDT", "lowercaseId": "alpineusdt", "symbol": "ALPINE/USDT", "base": "ALPINE", "quote": "USDT", "settle": null, "baseId": "ALPINE", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 453.88}}, "created": null, "info": {"symbol": "ALPINEUSDT", "status": "TRADING", "baseAsset": "ALPINE", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "453.88000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ALPINEUSD4", "lowercaseId": "alpineusd4", "symbol": "ALPINE/USD4", "base": "ALPINE", "quote": "USD4", "settle": null, "baseId": "ALPINE", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 19539.58060458}}, "created": null, "info": {"symbol": "ALPINEUSD4", "status": "BREAK", "baseAsset": "ALPINE", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "19539.58060458", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "PORTOUSDT", "lowercaseId": "portousdt", "symbol": "PORTO/USDT", "base": "PORTO", "quote": "USDT", "settle": null, "baseId": "PORTO", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 9.32}}, "created": null, "info": {"symbol": "PORTOUSDT", "status": "TRADING", "baseAsset": "PORTO", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "9.32000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "PORTOUSD4", "lowercaseId": "portousd4", "symbol": "PORTO/USD4", "base": "PORTO", "quote": "USD4", "settle": null, "baseId": "PORTO", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 29530.82736805}}, "created": null, "info": {"symbol": "PORTOUSD4", "status": "BREAK", "baseAsset": "PORTO", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "29530.82736805", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "RENUSDT", "lowercaseId": "renusdt", "symbol": "REN/USDT", "base": "REN", "quote": "USDT", "settle": null, "baseId": "REN", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 46996.03333333}}, "created": null, "info": {"symbol": "RENUSDT", "status": "TRADING", "baseAsset": "REN", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "46996.03333333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "RENUSD4", "lowercaseId": "renusd4", "symbol": "REN/USD4", "base": "REN", "quote": "USD4", "settle": null, "baseId": "REN", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 567579.47108066}}, "created": null, "info": {"symbol": "RENUSD4", "status": "BREAK", "baseAsset": "REN", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "567579.47108066", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "CELRUSDT", "lowercaseId": "celrusdt", "symbol": "CELR/USDT", "base": "CELR", "quote": "USDT", "settle": null, "baseId": "CELR", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9222449.0}, "price": {"min": 1e-05, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 123393.38833333}}, "created": null, "info": {"symbol": "CELRUSDT", "status": "TRADING", "baseAsset": "CELR", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "10000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9222449.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "123393.38833333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "CELRUSD4", "lowercaseId": "celrusd4", "symbol": "CELR/USD4", "base": "CELR", "quote": "USD4", "settle": null, "baseId": "CELR", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9222449.0}, "price": {"min": 0.0001, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 7822685.38698578}}, "created": null, "info": {"symbol": "CELRUSD4", "status": "BREAK", "baseAsset": "CELR", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "10000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9222449.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "7822685.38698578", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SKLUSDT", "lowercaseId": "sklusdt", "symbol": "SKL/USDT", "base": "SKL", "quote": "USDT", "settle": null, "baseId": "SKL", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 93202.49792531}}, "created": null, "info": {"symbol": "SKLUSDT", "status": "TRADING", "baseAsset": "SKL", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "93202.49792531", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SKLUSD4", "lowercaseId": "sklusd4", "symbol": "SKL/USD4", "base": "SKL", "quote": "USD4", "settle": null, "baseId": "SKL", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 1759560.96742957}}, "created": null, "info": {"symbol": "SKLUSD4", "status": "BREAK", "baseAsset": "SKL", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1759560.96742957", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "VITEUSDT", "lowercaseId": "viteusdt", "symbol": "VITE/USDT", "base": "VITE", "quote": "USDT", "settle": null, "baseId": "VITE", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 1058291.87916666}}, "created": null, "info": {"symbol": "VITEUSDT", "status": "TRADING", "baseAsset": "VITE", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1058291.87916666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "VITEUSD4", "lowercaseId": "viteusd4", "symbol": "VITE/USD4", "base": "VITE", "quote": "USD4", "settle": null, "baseId": "VITE", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 2860785.40650263}}, "created": null, "info": {"symbol": "VITEUSD4", "status": "BREAK", "baseAsset": "VITE", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "2860785.40650263", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "WAXPUSDT", "lowercaseId": "waxpusdt", "symbol": "WAXP/USDT", "base": "WAXP", "quote": "USDT", "settle": null, "baseId": "WAXP", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 45295.95}}, "created": null, "info": {"symbol": "WAXPUSDT", "status": "TRADING", "baseAsset": "WAXP", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "45295.95000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "WAXPUSD4", "lowercaseId": "waxpusd4", "symbol": "WAXP/USD4", "base": "WAXP", "quote": "USD4", "settle": null, "baseId": "WAXP", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 551927.59930555}}, "created": null, "info": {"symbol": "WAXPUSD4", "status": "BREAK", "baseAsset": "WAXP", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "551927.59930555", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LTOUSDT", "lowercaseId": "ltousdt", "symbol": "LTO/USDT", "base": "LTO", "quote": "USDT", "settle": null, "baseId": "LTO", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 7494.72083333}}, "created": null, "info": {"symbol": "LTOUSDT", "status": "TRADING", "baseAsset": "LTO", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "7494.72083333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LTOUSD4", "lowercaseId": "ltousd4", "symbol": "LTO/USD4", "base": "LTO", "quote": "USD4", "settle": null, "baseId": "LTO", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 1042269.79360667}}, "created": null, "info": {"symbol": "LTOUSD4", "status": "BREAK", "baseAsset": "LTO", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1042269.79360667", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "FETUSDT", "lowercaseId": "fetusdt", "symbol": "FET/USDT", "base": "FET", "quote": "USDT", "settle": null, "baseId": "FET", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 34011.29875}}, "created": null, "info": {"symbol": "FETUSDT", "status": "TRADING", "baseAsset": "FET", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "34011.29875000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "FETUSD4", "lowercaseId": "fetusd4", "symbol": "FET/USD4", "base": "FET", "quote": "USD4", "settle": null, "baseId": "FET", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 679752.96942321}}, "created": null, "info": {"symbol": "FETUSD4", "status": "BREAK", "baseAsset": "FET", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "679752.96942321", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BONDUSDT", "lowercaseId": "bondusdt", "symbol": "BOND/USDT", "base": "BOND", "quote": "USDT", "settle": null, "baseId": "BOND", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 9222449.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 516.34578512}}, "created": null, "info": {"symbol": "BONDUSDT", "status": "BREAK", "baseAsset": "BOND", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "9222449.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "516.34578512", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BONDUSD4", "lowercaseId": "bondusd4", "symbol": "BOND/USD4", "base": "BOND", "quote": "USD4", "settle": null, "baseId": "BOND", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 9222449.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 16758.01062107}}, "created": null, "info": {"symbol": "BONDUSD4", "status": "BREAK", "baseAsset": "BOND", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "2", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0100", "maxPrice": "10000.0000", "tickSize": "0.0100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "9222449.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "16758.01062107", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LOKAUSDT", "lowercaseId": "lokausdt", "symbol": "LOKA/USDT", "base": "LOKA", "quote": "USDT", "settle": null, "baseId": "LOKA", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 3588.92125}}, "created": null, "info": {"symbol": "LOKAUSDT", "status": "TRADING", "baseAsset": "LOKA", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3588.92125000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LOKAUSD4", "lowercaseId": "lokausd4", "symbol": "LOKA/USD4", "base": "LOKA", "quote": "USD4", "settle": null, "baseId": "LOKA", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 136861.61279554}}, "created": null, "info": {"symbol": "LOKAUSD4", "status": "BREAK", "baseAsset": "LOKA", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "136861.61279554", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ICPUSDT", "lowercaseId": "icpusdt", "symbol": "ICP/USDT", "base": "ICP", "quote": "USDT", "settle": null, "baseId": "ICP", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 922327.0}, "price": {"min": 0.001, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 2126.0145}}, "created": null, "info": {"symbol": "ICPUSDT", "status": "TRADING", "baseAsset": "ICP", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "100000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "922327.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "2126.01450000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ICPUSD4", "lowercaseId": "icpusd4", "symbol": "ICP/USD4", "base": "ICP", "quote": "USD4", "settle": null, "baseId": "ICP", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 922327.0}, "price": {"min": 0.01, "max": 100000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 86715.77427777}}, "created": null, "info": {"symbol": "ICPUSD4", "status": "BREAK", "baseAsset": "ICP", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "2", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0100", "maxPrice": "100000.0000", "tickSize": "0.0100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "922327.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "86715.77427777", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "TUSDT", "lowercaseId": "tusdt", "symbol": "T/USDT", "base": "T", "quote": "USDT", "settle": null, "baseId": "T", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 34823.19875}}, "created": null, "info": {"symbol": "TUSDT", "status": "TRADING", "baseAsset": "T", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "34823.19875000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "TUSD4", "lowercaseId": "tusd4", "symbol": "T/USD4", "base": "T", "quote": "USD4", "settle": null, "baseId": "T", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 3671409.47795549}}, "created": null, "info": {"symbol": "TUSD4", "status": "BREAK", "baseAsset": "T", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3671409.47795549", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "OPUSDT", "lowercaseId": "opusdt", "symbol": "OP/USDT", "base": "OP", "quote": "USDT", "settle": null, "baseId": "OP", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 17040.64508333}}, "created": null, "info": {"symbol": "OPUSDT", "status": "TRADING", "baseAsset": "OP", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "17040.64508333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "OPUSD4", "lowercaseId": "opusd4", "symbol": "OP/USD4", "base": "OP", "quote": "USD4", "settle": null, "baseId": "OP", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 1154139.26001388}}, "created": null, "info": {"symbol": "OPUSD4", "status": "BREAK", "baseAsset": "OP", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "1000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1154139.26001388", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ROSEUSDT", "lowercaseId": "roseusdt", "symbol": "ROSE/USDT", "base": "ROSE", "quote": "USDT", "settle": null, "baseId": "ROSE", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 65095.02625}}, "created": null, "info": {"symbol": "ROSEUSDT", "status": "TRADING", "baseAsset": "ROSE", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "65095.02625000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ROSEUSD4", "lowercaseId": "roseusd4", "symbol": "ROSE/USD4", "base": "ROSE", "quote": "USD4", "settle": null, "baseId": "ROSE", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 6038736.71125781}}, "created": null, "info": {"symbol": "ROSEUSD4", "status": "BREAK", "baseAsset": "ROSE", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "6038736.71125781", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "CELOUSDT", "lowercaseId": "celousdt", "symbol": "CELO/USDT", "base": "CELO", "quote": "USDT", "settle": null, "baseId": "CELO", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 7276.11375}}, "created": null, "info": {"symbol": "CELOUSDT", "status": "TRADING", "baseAsset": "CELO", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "7276.11375000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "CELOUSD4", "lowercaseId": "celousd4", "symbol": "CELO/USD4", "base": "CELO", "quote": "USD4", "settle": null, "baseId": "CELO", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 191076.78429464}}, "created": null, "info": {"symbol": "CELOUSD4", "status": "BREAK", "baseAsset": "CELO", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "1000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "191076.78429464", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "KDAUSDT", "lowercaseId": "kdausdt", "symbol": "KDA/USDT", "base": "KDA", "quote": "USDT", "settle": null, "baseId": "KDA", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 6872.32770833}}, "created": null, "info": {"symbol": "KDAUSDT", "status": "TRADING", "baseAsset": "KDA", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "6872.32770833", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "KDAUSD4", "lowercaseId": "kdausd4", "symbol": "KDA/USD4", "base": "KDA", "quote": "USD4", "settle": null, "baseId": "KDA", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 130664.17344537}}, "created": null, "info": {"symbol": "KDAUSD4", "status": "BREAK", "baseAsset": "KDA", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "1000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "130664.17344537", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "KSMUSDT", "lowercaseId": "ksmusdt", "symbol": "KSM/USDT", "base": "KSM", "quote": "USDT", "settle": null, "baseId": "KSM", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 900000.0}, "price": {"min": 0.01, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 164.28202916}}, "created": null, "info": {"symbol": "KSMUSDT", "status": "TRADING", "baseAsset": "KSM", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "100000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "900000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "164.28202916", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "KSMUSD4", "lowercaseId": "ksmusd4", "symbol": "KSM/USD4", "base": "KSM", "quote": "USD4", "settle": null, "baseId": "KSM", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.1, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 900000.0}, "price": {"min": 0.1, "max": 100000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 3294.32087482}}, "created": null, "info": {"symbol": "KSMUSD4", "status": "BREAK", "baseAsset": "KSM", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "1", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.1000", "maxPrice": "100000.0000", "tickSize": "0.1000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "900000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3294.32087482", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ACHUSDT", "lowercaseId": "achusdt", "symbol": "ACH/USDT", "base": "ACH", "quote": "USDT", "settle": null, "baseId": "ACH", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 119890.49166666}}, "created": null, "info": {"symbol": "ACHUSDT", "status": "TRADING", "baseAsset": "ACH", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "119890.49166666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ACHUSD4", "lowercaseId": "achusd4", "symbol": "ACH/USD4", "base": "ACH", "quote": "USD4", "settle": null, "baseId": "ACH", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 9285110.73706293}}, "created": null, "info": {"symbol": "ACHUSD4", "status": "BREAK", "baseAsset": "ACH", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "9285110.73706293", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "DARUSDT", "lowercaseId": "darusdt", "symbol": "DAR/USDT", "base": "DAR", "quote": "USDT", "settle": null, "baseId": "DAR", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 27584.32916666}}, "created": null, "info": {"symbol": "DARUSDT", "status": "TRADING", "baseAsset": "DAR", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "27584.32916666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "DARUSD4", "lowercaseId": "darusd4", "symbol": "DAR/USD4", "base": "DAR", "quote": "USD4", "settle": null, "baseId": "DAR", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 593200.79861111}}, "created": null, "info": {"symbol": "DARUSD4", "status": "BREAK", "baseAsset": "DAR", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "593200.79861111", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "RNDRUSDT", "lowercaseId": "rndrusdt", "symbol": "RNDR/USDT", "base": "RNDR", "quote": "USDT", "settle": null, "baseId": "RNDR", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 3863.83438016}}, "created": null, "info": {"symbol": "RNDRUSDT", "status": "BREAK", "baseAsset": "RNDR", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3863.83438016", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "RNDRUSD4", "lowercaseId": "rndrusd4", "symbol": "RNDR/USD4", "base": "RNDR", "quote": "USD4", "settle": null, "baseId": "RNDR", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 161822.84313671}}, "created": null, "info": {"symbol": "RNDRUSD4", "status": "BREAK", "baseAsset": "RNDR", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "1000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "161822.84313671", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SYSUSDT", "lowercaseId": "sysusdt", "symbol": "SYS/USDT", "base": "SYS", "quote": "USDT", "settle": null, "baseId": "SYS", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 36683.5125}}, "created": null, "info": {"symbol": "SYSUSDT", "status": "TRADING", "baseAsset": "SYS", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "36683.51250000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SYSUSD4", "lowercaseId": "sysusd4", "symbol": "SYS/USD4", "base": "SYS", "quote": "USD4", "settle": null, "baseId": "SYS", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 966695.58905058}}, "created": null, "info": {"symbol": "SYSUSD4", "status": "BREAK", "baseAsset": "SYS", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "966695.58905058", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "RADUSDT", "lowercaseId": "radusdt", "symbol": "RAD/USDT", "base": "RAD", "quote": "USDT", "settle": null, "baseId": "RAD", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 870.59833333}}, "created": null, "info": {"symbol": "RADUSDT", "status": "TRADING", "baseAsset": "RAD", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "870.59833333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "RADUSD4", "lowercaseId": "radusd4", "symbol": "RAD/USD4", "base": "RAD", "quote": "USD4", "settle": null, "baseId": "RAD", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 47546.77801664}}, "created": null, "info": {"symbol": "RADUSD4", "status": "BREAK", "baseAsset": "RAD", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "1000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "47546.77801664", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ILVUSDT", "lowercaseId": "ilvusdt", "symbol": "ILV/USDT", "base": "ILV", "quote": "USDT", "settle": null, "baseId": "ILV", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 922327.0}, "price": {"min": 0.01, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 18.82657916}}, "created": null, "info": {"symbol": "ILVUSDT", "status": "TRADING", "baseAsset": "ILV", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "100000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "922327.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "18.82657916", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ILVUSD4", "lowercaseId": "ilvusd4", "symbol": "ILV/USD4", "base": "ILV", "quote": "USD4", "settle": null, "baseId": "ILV", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.1, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 922327.0}, "price": {"min": 0.1, "max": 100000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 916.93937734}}, "created": null, "info": {"symbol": "ILVUSD4", "status": "BREAK", "baseAsset": "ILV", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "1", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.1000", "maxPrice": "100000.0000", "tickSize": "0.1000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "922327.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "916.93937734", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LDOUSDT", "lowercaseId": "ldousdt", "symbol": "LDO/USDT", "base": "LDO", "quote": "USDT", "settle": null, "baseId": "LDO", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 4001.340875}}, "created": null, "info": {"symbol": "LDOUSDT", "status": "TRADING", "baseAsset": "LDO", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "4001.34087500", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LDOUSD4", "lowercaseId": "ldousd4", "symbol": "LDO/USD4", "base": "LDO", "quote": "USD4", "settle": null, "baseId": "LDO", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 316622.60368932}}, "created": null, "info": {"symbol": "LDOUSD4", "status": "BREAK", "baseAsset": "LDO", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "1000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "316622.60368932", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "RAREUSDT", "lowercaseId": "rareusdt", "symbol": "RARE/USDT", "base": "RARE", "quote": "USDT", "settle": null, "baseId": "RARE", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 9962.99541666}}, "created": null, "info": {"symbol": "RAREUSDT", "status": "TRADING", "baseAsset": "RARE", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "9962.99541666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "RAREUSD4", "lowercaseId": "rareusd4", "symbol": "RARE/USD4", "base": "RARE", "quote": "USD4", "settle": null, "baseId": "RARE", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 390621.67142857}}, "created": null, "info": {"symbol": "RAREUSD4", "status": "BREAK", "baseAsset": "RARE", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "1000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "390621.67142857", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LSKUSDT", "lowercaseId": "lskusdt", "symbol": "LSK/USDT", "base": "LSK", "quote": "USDT", "settle": null, "baseId": "LSK", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 1032.40666666}}, "created": null, "info": {"symbol": "LSKUSDT", "status": "TRADING", "baseAsset": "LSK", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1032.40666666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LSKUSD4", "lowercaseId": "lskusd4", "symbol": "LSK/USD4", "base": "LSK", "quote": "USD4", "settle": null, "baseId": "LSK", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 62355.45987525}}, "created": null, "info": {"symbol": "LSKUSD4", "status": "BREAK", "baseAsset": "LSK", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "1000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "62355.45987525", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "DGBUSDT", "lowercaseId": "dgbusdt", "symbol": "DGB/USDT", "base": "DGB", "quote": "USDT", "settle": null, "baseId": "DGB", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 412765.27833333}}, "created": null, "info": {"symbol": "DGBUSDT", "status": "TRADING", "baseAsset": "DGB", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "412765.27833333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "DGBUSD4", "lowercaseId": "dgbusd4", "symbol": "DGB/USD4", "base": "DGB", "quote": "USD4", "settle": null, "baseId": "DGB", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 6176501.64299583}}, "created": null, "info": {"symbol": "DGBUSD4", "status": "BREAK", "baseAsset": "DGB", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "6176501.64299583", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "REEFUSDT", "lowercaseId": "reefusdt", "symbol": "REEF/USDT", "base": "REEF", "quote": "USDT", "settle": null, "baseId": "REEF", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-06, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 1e-06, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 2844299.1286307}}, "created": null, "info": {"symbol": "REEFUSDT", "status": "TRADING", "baseAsset": "REEF", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000100", "maxPrice": "1000.00000000", "tickSize": "0.00000100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "2844299.12863070", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "REEFUSD4", "lowercaseId": "reefusd4", "symbol": "REEF/USD4", "base": "REEF", "quote": "USD4", "settle": null, "baseId": "REEF", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 19979356.49271339}}, "created": null, "info": {"symbol": "REEFUSD4", "status": "BREAK", "baseAsset": "REEF", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "19979356.49271339", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SRMUSDT", "lowercaseId": "srmusdt", "symbol": "SRM/USDT", "base": "SRM", "quote": "USDT", "settle": null, "baseId": "SRM", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 161222.85177206}}, "created": null, "info": {"symbol": "SRMUSDT", "status": "BREAK", "baseAsset": "SRM", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "161222.85177206", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SRMUSD", "lowercaseId": "srmusd", "symbol": "SRM/USD4", "base": "SRM", "quote": "USD4", "settle": null, "baseId": "SRM", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 161222.85177206}}, "created": null, "info": {"symbol": "SRMUSD", "status": "BREAK", "baseAsset": "SRM", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "161222.85177206", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ALICEUSDT", "lowercaseId": "aliceusdt", "symbol": "ALICE/USDT", "base": "ALICE", "quote": "USDT", "settle": null, "baseId": "ALICE", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 2387.807}}, "created": null, "info": {"symbol": "ALICEUSDT", "status": "TRADING", "baseAsset": "ALICE", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "2387.80700000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ALICEUSD4", "lowercaseId": "aliceusd4", "symbol": "ALICE/USD4", "base": "ALICE", "quote": "USD4", "settle": null, "baseId": "ALICE", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 90594.63785962}}, "created": null, "info": {"symbol": "ALICEUSD4", "status": "BREAK", "baseAsset": "ALICE", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "1000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "90594.63785962", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "FORTHUSDT", "lowercaseId": "forthusdt", "symbol": "FORTH/USDT", "base": "FORTH", "quote": "USDT", "settle": null, "baseId": "FORTH", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 330.608625}}, "created": null, "info": {"symbol": "FORTHUSDT", "status": "TRADING", "baseAsset": "FORTH", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "330.60862500", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "FORTHUSD4", "lowercaseId": "forthusd4", "symbol": "FORTH/USD4", "base": "FORTH", "quote": "USD4", "settle": null, "baseId": "FORTH", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 42037.71566666}}, "created": null, "info": {"symbol": "FORTHUSD4", "status": "BREAK", "baseAsset": "FORTH", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "2", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0100", "maxPrice": "10000.0000", "tickSize": "0.0100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "42037.71566666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ASTRUSDT", "lowercaseId": "astrusdt", "symbol": "ASTR/USDT", "base": "ASTR", "quote": "USDT", "settle": null, "baseId": "ASTR", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 93965.6}}, "created": null, "info": {"symbol": "ASTRUSDT", "status": "TRADING", "baseAsset": "ASTR", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "93965.60000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ASTRUSD4", "lowercaseId": "astrusd4", "symbol": "ASTR/USD4", "base": "ASTR", "quote": "USD4", "settle": null, "baseId": "ASTR", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 2998944.25447605}}, "created": null, "info": {"symbol": "ASTRUSD4", "status": "BREAK", "baseAsset": "ASTR", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "2998944.25447605", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BTRSTUSDT", "lowercaseId": "btrstusdt", "symbol": "BTRST/USDT", "base": "BTRST", "quote": "USDT", "settle": null, "baseId": "BTRST", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9222449.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 1995.06458333}}, "created": null, "info": {"symbol": "BTRSTUSDT", "status": "TRADING", "baseAsset": "BTRST", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9222449.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1995.06458333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BTRSTUSD4", "lowercaseId": "btrstusd4", "symbol": "BTRST/USD4", "base": "BTRST", "quote": "USD4", "settle": null, "baseId": "BTRST", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9222449.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 63083.19006944}}, "created": null, "info": {"symbol": "BTRSTUSD4", "status": "BREAK", "baseAsset": "BTRST", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "10000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9222449.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "63083.19006944", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "GALUSDT", "lowercaseId": "galusdt", "symbol": "GAL/USDT", "base": "GAL", "quote": "USDT", "settle": null, "baseId": "GAL", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 243.68947933}}, "created": null, "info": {"symbol": "GALUSDT", "status": "BREAK", "baseAsset": "GAL", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "92141578.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "243.68947933", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "GALUSD4", "lowercaseId": "galusd4", "symbol": "GAL/USD4", "base": "GAL", "quote": "USD4", "settle": null, "baseId": "GAL", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 154650.16084403}}, "created": null, "info": {"symbol": "GALUSD4", "status": "BREAK", "baseAsset": "GAL", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "1000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "92141578.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "154650.16084403", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SANDUSDT", "lowercaseId": "sandusdt", "symbol": "SAND/USDT", "base": "SAND", "quote": "USDT", "settle": null, "baseId": "SAND", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 22985.975}}, "created": null, "info": {"symbol": "SANDUSDT", "status": "TRADING", "baseAsset": "SAND", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "22985.97500000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SANDUSD4", "lowercaseId": "sandusd4", "symbol": "SAND/USD4", "base": "SAND", "quote": "USD4", "settle": null, "baseId": "SAND", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 928454.74068965}}, "created": null, "info": {"symbol": "SANDUSD4", "status": "BREAK", "baseAsset": "SAND", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "928454.74068965", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BALUSDT", "lowercaseId": "balusdt", "symbol": "BAL/USDT", "base": "BAL", "quote": "USDT", "settle": null, "baseId": "BAL", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 630.69945833}}, "created": null, "info": {"symbol": "BALUSDT", "status": "TRADING", "baseAsset": "BAL", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "630.69945833", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BALUSD4", "lowercaseId": "balusd4", "symbol": "BAL/USD4", "base": "BAL", "quote": "USD4", "settle": null, "baseId": "BAL", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 18684.8525625}}, "created": null, "info": {"symbol": "BALUSD4", "status": "BREAK", "baseAsset": "BAL", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "10000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "18684.85256250", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "POLYXUSD4", "lowercaseId": "polyxusd4", "symbol": "POLYX/USD4", "base": "POLYX", "quote": "USD4", "settle": null, "baseId": "POLYX", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 92141578.0}}, "created": null, "info": {"symbol": "POLYXUSD4", "status": "BREAK", "baseAsset": "POLYX", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "92141578.00000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "GLMUSDT", "lowercaseId": "glmusdt", "symbol": "GLM/USDT", "base": "GLM", "quote": "USDT", "settle": null, "baseId": "GLM", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 4516.71208333}}, "created": null, "info": {"symbol": "GLMUSDT", "status": "TRADING", "baseAsset": "GLM", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "4516.71208333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "GLMUSD4", "lowercaseId": "glmusd4", "symbol": "GLM/USD4", "base": "GLM", "quote": "USD4", "settle": null, "baseId": "GLM", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 106586.10466898}}, "created": null, "info": {"symbol": "GLMUSD4", "status": "BREAK", "baseAsset": "GLM", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "106586.10466898", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "CLVUSDT", "lowercaseId": "clvusdt", "symbol": "CLV/USDT", "base": "CLV", "quote": "USDT", "settle": null, "baseId": "CLV", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 74351.05833333}}, "created": null, "info": {"symbol": "CLVUSDT", "status": "TRADING", "baseAsset": "CLV", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "74351.05833333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "CLVUSD4", "lowercaseId": "clvusd4", "symbol": "CLV/USD4", "base": "CLV", "quote": "USD4", "settle": null, "baseId": "CLV", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 1341583.34923398}}, "created": null, "info": {"symbol": "CLVUSD4", "status": "BREAK", "baseAsset": "CLV", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1341583.34923398", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "TUSDUSDT", "lowercaseId": "tusdusdt", "symbol": "TUSD/USDT", "base": "TUSD", "quote": "USDT", "settle": null, "baseId": "TUSD", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 718.0}}, "created": null, "info": {"symbol": "TUSDUSDT", "status": "BREAK", "baseAsset": "TUSD", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "1.2", "multiplierDown": "0.8", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "900000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "718.00000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.05", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.95", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "TUSDUSD4", "lowercaseId": "tusdusd4", "symbol": "TUSD/USD4", "base": "TUSD", "quote": "USD4", "settle": null, "baseId": "TUSD", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 1301033.56392265}}, "created": null, "info": {"symbol": "TUSDUSD4", "status": "BREAK", "baseAsset": "TUSD", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "1.2", "multiplierDown": "0.8", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "900000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1301033.56392265", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "QNTUSDT", "lowercaseId": "qntusdt", "symbol": "QNT/USDT", "base": "QNT", "quote": "USDT", "settle": null, "baseId": "QNT", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.1, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 922327.0}, "price": {"min": 0.1, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 36.8777125}}, "created": null, "info": {"symbol": "QNTUSDT", "status": "TRADING", "baseAsset": "QNT", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.10000000", "maxPrice": "100000.00000000", "tickSize": "0.10000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "922327.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "36.87771250", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "QNTUSD4", "lowercaseId": "qntusd4", "symbol": "QNT/USD4", "base": "QNT", "quote": "USD4", "settle": null, "baseId": "QNT", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.1, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 922327.0}, "price": {"min": 0.1, "max": 100000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 6389.58374712}}, "created": null, "info": {"symbol": "QNTUSD4", "status": "BREAK", "baseAsset": "QNT", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "1", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.1000", "maxPrice": "100000.0000", "tickSize": "0.1000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "922327.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "6389.58374712", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "STGUSDT", "lowercaseId": "stgusdt", "symbol": "STG/USDT", "base": "STG", "quote": "USDT", "settle": null, "baseId": "STG", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 913205152.0}, "price": {"min": 0.0001, "max": 100.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 6727.77125}}, "created": null, "info": {"symbol": "STGUSDT", "status": "TRADING", "baseAsset": "STG", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "100.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "913205152.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "6727.77125000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "STGUSD4", "lowercaseId": "stgusd4", "symbol": "STG/USD4", "base": "STG", "quote": "USD4", "settle": null, "baseId": "STG", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 913205152.0}, "price": {"min": 0.0001, "max": 100.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 260651.43672883}}, "created": null, "info": {"symbol": "STGUSD4", "status": "BREAK", "baseAsset": "STG", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "100.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "913205152.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "260651.43672883", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "AXLUSDT", "lowercaseId": "axlusdt", "symbol": "AXL/USDT", "base": "AXL", "quote": "USDT", "settle": null, "baseId": "AXL", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 922327.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 9408.51666666}}, "created": null, "info": {"symbol": "AXLUSDT", "status": "TRADING", "baseAsset": "AXL", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "922327.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "9408.51666666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "AXLUSD4", "lowercaseId": "axlusd4", "symbol": "AXL/USD4", "base": "AXL", "quote": "USD4", "settle": null, "baseId": "AXL", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 922327.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 511719.87214644}}, "created": null, "info": {"symbol": "AXLUSD4", "status": "BREAK", "baseAsset": "AXL", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "10000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "922327.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "511719.87214644", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "KAVAUSDT", "lowercaseId": "kavausdt", "symbol": "KAVA/USDT", "base": "KAVA", "quote": "USDT", "settle": null, "baseId": "KAVA", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 7059.96833333}}, "created": null, "info": {"symbol": "KAVAUSDT", "status": "TRADING", "baseAsset": "KAVA", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "7059.96833333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "KAVAUSD4", "lowercaseId": "kavausd4", "symbol": "KAVA/USD4", "base": "KAVA", "quote": "USD4", "settle": null, "baseId": "KAVA", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 103605.40152883}}, "created": null, "info": {"symbol": "KAVAUSD4", "status": "BREAK", "baseAsset": "KAVA", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "1000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "103605.40152883", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "APTUSDT", "lowercaseId": "aptusdt", "symbol": "APT/USDT", "base": "APT", "quote": "USDT", "settle": null, "baseId": "APT", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.01, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 1882.77820833}}, "created": null, "info": {"symbol": "APTUSDT", "status": "TRADING", "baseAsset": "APT", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "1000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1882.77820833", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "APTUSD4", "lowercaseId": "aptusd4", "symbol": "APT/USD4", "base": "APT", "quote": "USD4", "settle": null, "baseId": "APT", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 131395.7994371}}, "created": null, "info": {"symbol": "APTUSD4", "status": "BREAK", "baseAsset": "APT", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "131395.79943710", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "MASKUSDT", "lowercaseId": "maskusdt", "symbol": "MASK/USDT", "base": "MASK", "quote": "USDT", "settle": null, "baseId": "MASK", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9222449.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 966.70458333}}, "created": null, "info": {"symbol": "MASKUSDT", "status": "TRADING", "baseAsset": "MASK", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9222449.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "966.70458333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "MASKUSD4", "lowercaseId": "maskusd4", "symbol": "MASK/USD4", "base": "MASK", "quote": "USD4", "settle": null, "baseId": "MASK", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9222449.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 146077.9675}}, "created": null, "info": {"symbol": "MASKUSD4", "status": "BREAK", "baseAsset": "MASK", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "10000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9222449.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "146077.96750000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BOSONUSDT", "lowercaseId": "bosonusdt", "symbol": "BOSON/USDT", "base": "BOSON", "quote": "USDT", "settle": null, "baseId": "BOSON", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 15562.59958333}}, "created": null, "info": {"symbol": "BOSONUSDT", "status": "TRADING", "baseAsset": "BOSON", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "15562.59958333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BOSONUSD4", "lowercaseId": "bosonusd4", "symbol": "BOSON/USD4", "base": "BOSON", "quote": "USD4", "settle": null, "baseId": "BOSON", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 96324.3375}}, "created": null, "info": {"symbol": "BOSONUSD4", "status": "BREAK", "baseAsset": "BOSON", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "96324.33750000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "PONDUSDT", "lowercaseId": "pondusdt", "symbol": "POND/USDT", "base": "POND", "quote": "USDT", "settle": null, "baseId": "POND", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9222449.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 141447.725}}, "created": null, "info": {"symbol": "PONDUSDT", "status": "TRADING", "baseAsset": "POND", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9222449.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "141447.72500000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "PONDUSD4", "lowercaseId": "pondusd4", "symbol": "POND/USD4", "base": "POND", "quote": "USD4", "settle": null, "baseId": "POND", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 9222449.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 9222449.0}}, "created": null, "info": {"symbol": "PONDUSD4", "status": "BREAK", "baseAsset": "POND", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "9222449.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "9222449.00000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SOLUSDC", "lowercaseId": "solusdc", "symbol": "SOL/USDC", "base": "SOL", "quote": "USDC", "settle": null, "baseId": "SOL", "quoteId": "USDC", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 90000.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 457.091625}}, "created": null, "info": {"symbol": "SOLUSDC", "status": "TRADING", "baseAsset": "SOL", "baseAssetPrecision": "8", "quoteAsset": "USDC", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "10000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "90000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "457.09162500", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ADAUSDC", "lowercaseId": "adausdc", "symbol": "ADA/USDC", "base": "ADA", "quote": "USDC", "settle": null, "baseId": "ADA", "quoteId": "USDC", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 18554.86625}}, "created": null, "info": {"symbol": "ADAUSDC", "status": "TRADING", "baseAsset": "ADA", "baseAssetPrecision": "8", "quoteAsset": "USDC", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "18554.86625000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "MXCUSDT", "lowercaseId": "mxcusdt", "symbol": "MXC/USDT", "base": "MXC", "quote": "USDT", "settle": null, "baseId": "MXC", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 382251.40833333}}, "created": null, "info": {"symbol": "MXCUSDT", "status": "TRADING", "baseAsset": "MXC", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "382251.40833333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "MXCUSD4", "lowercaseId": "mxcusd4", "symbol": "MXC/USD4", "base": "MXC", "quote": "USD4", "settle": null, "baseId": "MXC", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 16852320.33958333}}, "created": null, "info": {"symbol": "MXCUSD4", "status": "BREAK", "baseAsset": "MXC", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "16852320.33958333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "JAMUSDT", "lowercaseId": "jamusdt", "symbol": "JAM/USDT", "base": "JAM", "quote": "USDT", "settle": null, "baseId": "JAM", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-06, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 90000000.0}, "price": {"min": 1e-06, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 9538785.3125}}, "created": null, "info": {"symbol": "JAMUSDT", "status": "TRADING", "baseAsset": "JAM", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000100", "maxPrice": "1000.00000000", "tickSize": "0.00000100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "90000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "9538785.31250000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "JAMUSD4", "lowercaseId": "jamusd4", "symbol": "JAM/USD4", "base": "JAM", "quote": "USD4", "settle": null, "baseId": "JAM", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 90000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 67578970.50451702}}, "created": null, "info": {"symbol": "JAMUSD4", "status": "BREAK", "baseAsset": "JAM", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "90000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "67578970.50451702", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "TRACUSDT", "lowercaseId": "tracusdt", "symbol": "TRAC/USDT", "base": "TRAC", "quote": "USDT", "settle": null, "baseId": "TRAC", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 3655.15352697}}, "created": null, "info": {"symbol": "TRACUSDT", "status": "TRADING", "baseAsset": "TRAC", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3655.15352697", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "PROMUSDT", "lowercaseId": "promusdt", "symbol": "PROM/USDT", "base": "PROM", "quote": "USDT", "settle": null, "baseId": "PROM", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 97.80470833}}, "created": null, "info": {"symbol": "PROMUSDT", "status": "TRADING", "baseAsset": "PROM", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "97.80470833", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "PROMUSD4", "lowercaseId": "promusd4", "symbol": "PROM/USD4", "base": "PROM", "quote": "USD4", "settle": null, "baseId": "PROM", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 31980.2771369}}, "created": null, "info": {"symbol": "PROMUSD4", "status": "BREAK", "baseAsset": "PROM", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "2", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0100", "maxPrice": "10000.0000", "tickSize": "0.0100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "31980.27713690", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "DIAUSDT", "lowercaseId": "diausdt", "symbol": "DIA/USDT", "base": "DIA", "quote": "USDT", "settle": null, "baseId": "DIA", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 922327.0}, "price": {"min": 0.0001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 1119.23875}}, "created": null, "info": {"symbol": "DIAUSDT", "status": "TRADING", "baseAsset": "DIA", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "10000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "922327.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1119.23875000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "DIAUSD4", "lowercaseId": "diausd4", "symbol": "DIA/USD4", "base": "DIA", "quote": "USD4", "settle": null, "baseId": "DIA", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 922327.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 300013.60541666}}, "created": null, "info": {"symbol": "DIAUSD4", "status": "BREAK", "baseAsset": "DIA", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "3", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0010", "maxPrice": "10000.0000", "tickSize": "0.0010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "922327.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "300013.60541666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BTCDAI", "lowercaseId": "btcdai", "symbol": "BTC/DAI", "base": "BTC", "quote": "DAI", "settle": null, "baseId": "BTC", "quoteId": "DAI", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1e-05, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1e-05, "max": 9000.0}, "price": {"min": 0.01, "max": 1000000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 0.00094}}, "created": null, "info": {"symbol": "BTCDAI", "status": "BREAK", "baseAsset": "BTC", "baseAssetPrecision": "8", "quoteAsset": "DAI", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "1000000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00001000", "maxQty": "9000.00000000", "stepSize": "0.00001000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "0.00094000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ETHDAI", "lowercaseId": "ethdai", "symbol": "ETH/DAI", "base": "ETH", "quote": "DAI", "settle": null, "baseId": "ETH", "quoteId": "DAI", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.0001, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.0001, "max": 90000.0}, "price": {"min": 0.01, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 0.011}}, "created": null, "info": {"symbol": "ETHDAI", "status": "BREAK", "baseAsset": "ETH", "baseAssetPrecision": "8", "quoteAsset": "DAI", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "100000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00010000", "maxQty": "90000.00000000", "stepSize": "0.00010000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "0.01100000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ADAETH", "lowercaseId": "adaeth", "symbol": "ADA/ETH", "base": "ADA", "quote": "ETH", "settle": null, "baseId": "ADA", "quoteId": "ETH", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-07, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 90000000.0}, "price": {"min": 1e-07, "max": 1000.0}, "cost": {"min": 0.005, "max": null}, "market": {"min": 0.0, "max": 15932.4975}}, "created": null, "info": {"symbol": "ADAETH", "status": "TRADING", "baseAsset": "ADA", "baseAssetPrecision": "8", "quoteAsset": "ETH", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000010", "maxPrice": "1000.00000000", "tickSize": "0.00000010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "90000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "0.00500000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "15932.49750000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "DOGEBTC", "lowercaseId": "dogebtc", "symbol": "DOGE/BTC", "base": "DOGE", "quote": "BTC", "settle": null, "baseId": "DOGE", "quoteId": "BTC", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-08, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 90000000.0}, "price": {"min": 1e-08, "max": 1000.0}, "cost": {"min": 0.0001, "max": null}, "market": {"min": 0.0, "max": 251729.25}}, "created": null, "info": {"symbol": "DOGEBTC", "status": "TRADING", "baseAsset": "DOGE", "baseAssetPrecision": "8", "quoteAsset": "BTC", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000001", "maxPrice": "1000.00000000", "tickSize": "0.00000001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "90000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "0.00010000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "251729.25000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LOOMUSDT", "lowercaseId": "loomusdt", "symbol": "LOOM/USDT", "base": "LOOM", "quote": "USDT", "settle": null, "baseId": "LOOM", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 30301.00833333}}, "created": null, "info": {"symbol": "LOOMUSDT", "status": "TRADING", "baseAsset": "LOOM", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "30301.00833333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LOOMUSD4", "lowercaseId": "loomusd4", "symbol": "LOOM/USD4", "base": "LOOM", "quote": "USD4", "settle": null, "baseId": "LOOM", "quoteId": "USD4", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 0.0001}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 1060513.60833333}}, "created": null, "info": {"symbol": "LOOMUSD4", "status": "BREAK", "baseAsset": "LOOM", "baseAssetPrecision": "8", "quoteAsset": "USD4", "quotePrecision": "4", "quoteAssetPrecision": "4", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "4", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.0001", "maxPrice": "1000.0000", "tickSize": "0.0001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.0000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1060513.60833333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "STMXUSDT", "lowercaseId": "stmxusdt", "symbol": "STMX/USDT", "base": "STMX", "quote": "USDT", "settle": null, "baseId": "STMX", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-06, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 1e-06, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 232273.75}}, "created": null, "info": {"symbol": "STMXUSDT", "status": "TRADING", "baseAsset": "STMX", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000100", "maxPrice": "1000.00000000", "tickSize": "0.00000100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "232273.75000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BTCUSD", "lowercaseId": "btcusd", "symbol": "BTC/USD", "base": "BTC", "quote": "USD", "settle": null, "baseId": "BTC", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1e-05, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1e-05, "max": 9000.0}, "price": {"min": 0.01, "max": 1000000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 3.53641561}}, "created": null, "info": {"symbol": "BTCUSD", "status": "BREAK", "baseAsset": "BTC", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "1000000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00001000", "maxQty": "9000.00000000", "stepSize": "0.00001000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3.53641561", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ETHUSD", "lowercaseId": "ethusd", "symbol": "ETH/USD", "base": "ETH", "quote": "USD", "settle": null, "baseId": "ETH", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.0001, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.0001, "max": 9000.0}, "price": {"min": 0.01, "max": 1000000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 38.6527682}}, "created": null, "info": {"symbol": "ETHUSD", "status": "BREAK", "baseAsset": "ETH", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "1000000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00010000", "maxQty": "9000.00000000", "stepSize": "0.00010000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "38.65276820", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BCHUSD", "lowercaseId": "bchusd", "symbol": "BCH/USD", "base": "BCH", "quote": "USD", "settle": null, "baseId": "BCH", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.1, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 90000.0}, "price": {"min": 0.1, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 598.64156564}}, "created": null, "info": {"symbol": "BCHUSD", "status": "BREAK", "baseAsset": "BCH", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.10000000", "maxPrice": "100000.00000000", "tickSize": "0.10000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "90000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "598.64156564", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LTCUSD", "lowercaseId": "ltcusd", "symbol": "LTC/USD", "base": "LTC", "quote": "USD", "settle": null, "baseId": "LTC", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 90000.0}, "price": {"min": 0.01, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 175.50162561}}, "created": null, "info": {"symbol": "LTCUSD", "status": "BREAK", "baseAsset": "LTC", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "100000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "90000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "175.50162561", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "USDTUSD", "lowercaseId": "usdtusd", "symbol": "USDT/USD", "base": "USDT", "quote": "USD", "settle": null, "baseId": "USDT", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 15000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 234723.40833333}}, "created": null, "info": {"symbol": "USDTUSD", "status": "TRADING", "baseAsset": "USDT", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "15000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "234723.40833333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.05", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.95", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BNBUSD", "lowercaseId": "bnbusd", "symbol": "BNB/USD", "base": "BNB", "quote": "USD", "settle": null, "baseId": "BNB", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.1, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 900000.0}, "price": {"min": 0.1, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 271.09617992}}, "created": null, "info": {"symbol": "BNBUSD", "status": "BREAK", "baseAsset": "BNB", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.10000000", "maxPrice": "100000.00000000", "tickSize": "0.10000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "900000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "271.09617992", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ADAUSD", "lowercaseId": "adausd", "symbol": "ADA/USD", "base": "ADA", "quote": "USD", "settle": null, "baseId": "ADA", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 161338.56125}}, "created": null, "info": {"symbol": "ADAUSD", "status": "BREAK", "baseAsset": "ADA", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "161338.56125000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BATUSD", "lowercaseId": "batusd", "symbol": "BAT/USD", "base": "BAT", "quote": "USD", "settle": null, "baseId": "BAT", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 337996.54613542}}, "created": null, "info": {"symbol": "BATUSD", "status": "BREAK", "baseAsset": "BAT", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "900000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "337996.54613542", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ETCUSD", "lowercaseId": "etcusd", "symbol": "ETC/USD", "base": "ETC", "quote": "USD", "settle": null, "baseId": "ETC", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 7672.11976042}}, "created": null, "info": {"symbol": "ETCUSD", "status": "BREAK", "baseAsset": "ETC", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "10000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "7672.11976042", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "XLMUSD", "lowercaseId": "xlmusd", "symbol": "XLM/USD", "base": "XLM", "quote": "USD", "settle": null, "baseId": "XLM", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 1789500.0799861}}, "created": null, "info": {"symbol": "XLMUSD", "status": "BREAK", "baseAsset": "XLM", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1789500.07998610", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ZRXUSD", "lowercaseId": "zrxusd", "symbol": "ZRX/USD", "base": "ZRX", "quote": "USD", "settle": null, "baseId": "ZRX", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 231944.61956597}}, "created": null, "info": {"symbol": "ZRXUSD", "status": "BREAK", "baseAsset": "ZRX", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "900000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "231944.61956597", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LINKUSD", "lowercaseId": "linkusd", "symbol": "LINK/USD", "base": "LINK", "quote": "USD", "settle": null, "baseId": "LINK", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 39973.88115358}}, "created": null, "info": {"symbol": "LINKUSD", "status": "BREAK", "baseAsset": "LINK", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "39973.88115358", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "RVNUSD", "lowercaseId": "rvnusd", "symbol": "RVN/USD", "base": "RVN", "quote": "USD", "settle": null, "baseId": "RVN", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 3164694.50375261}}, "created": null, "info": {"symbol": "RVNUSD", "status": "BREAK", "baseAsset": "RVN", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3164694.50375261", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "DASHUSD", "lowercaseId": "dashusd", "symbol": "DASH/USD", "base": "DASH", "quote": "USD", "settle": null, "baseId": "DASH", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1e-05, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1e-05, "max": 900000.0}, "price": {"min": 0.01, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 1187.00316815}}, "created": null, "info": {"symbol": "DASHUSD", "status": "BREAK", "baseAsset": "DASH", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "100000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00001000", "maxQty": "900000.00000000", "stepSize": "0.00001000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1187.00316815", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ZECUSD", "lowercaseId": "zecusd", "symbol": "ZEC/USD", "base": "ZEC", "quote": "USD", "settle": null, "baseId": "ZEC", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.1, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 900000.0}, "price": {"min": 0.1, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 3010.12044466}}, "created": null, "info": {"symbol": "ZECUSD", "status": "BREAK", "baseAsset": "ZEC", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.10000000", "maxPrice": "100000.00000000", "tickSize": "0.10000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "900000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3010.12044466", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ALGOUSD", "lowercaseId": "algousd", "symbol": "ALGO/USD", "base": "ALGO", "quote": "USD", "settle": null, "baseId": "ALGO", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 922327.0}, "price": {"min": 0.0001, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 922327.0}}, "created": null, "info": {"symbol": "ALGOUSD", "status": "BREAK", "baseAsset": "ALGO", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "100000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "922327.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "922327.00000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "IOTAUSD", "lowercaseId": "iotausd", "symbol": "IOTA/USD", "base": "IOTA", "quote": "USD", "settle": null, "baseId": "IOTA", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 146719.39888464}}, "created": null, "info": {"symbol": "IOTAUSD", "status": "BREAK", "baseAsset": "IOTA", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "900000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "146719.39888464", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BUSDUSD", "lowercaseId": "busdusd", "symbol": "BUSD/USD", "base": "BUSD", "quote": "USD", "settle": null, "baseId": "BUSD", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 15000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 4652432.407625}}, "created": null, "info": {"symbol": "BUSDUSD", "status": "BREAK", "baseAsset": "BUSD", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "15000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "4652432.40762500", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "WAVESUSD", "lowercaseId": "wavesusd", "symbol": "WAVES/USD", "base": "WAVES", "quote": "USD", "settle": null, "baseId": "WAVES", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 13445.78859375}}, "created": null, "info": {"symbol": "WAVESUSD", "status": "BREAK", "baseAsset": "WAVES", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "13445.78859375", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ATOMUSD", "lowercaseId": "atomusd", "symbol": "ATOM/USD", "base": "ATOM", "quote": "USD", "settle": null, "baseId": "ATOM", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 23542.84988472}}, "created": null, "info": {"symbol": "ATOMUSD", "status": "BREAK", "baseAsset": "ATOM", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "23542.84988472", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "NEOUSD", "lowercaseId": "neousd", "symbol": "NEO/USD", "base": "NEO", "quote": "USD", "settle": null, "baseId": "NEO", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 900000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 6970.17727589}}, "created": null, "info": {"symbol": "NEOUSD", "status": "BREAK", "baseAsset": "NEO", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "900000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "6970.17727589", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "QTUMUSD", "lowercaseId": "qtumusd", "symbol": "QTUM/USD", "base": "QTUM", "quote": "USD", "settle": null, "baseId": "QTUM", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 90000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 27528.80557847}}, "created": null, "info": {"symbol": "QTUMUSD", "status": "BREAK", "baseAsset": "QTUM", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "90000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "27528.80557847", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ICXUSD", "lowercaseId": "icxusd", "symbol": "ICX/USD", "base": "ICX", "quote": "USD", "settle": null, "baseId": "ICX", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 209894.29939541}}, "created": null, "info": {"symbol": "ICXUSD", "status": "BREAK", "baseAsset": "ICX", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "209894.29939541", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ENJUSD", "lowercaseId": "enjusd", "symbol": "ENJ/USD", "base": "ENJ", "quote": "USD", "settle": null, "baseId": "ENJ", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 119667.26490618}}, "created": null, "info": {"symbol": "ENJUSD", "status": "BREAK", "baseAsset": "ENJ", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "119667.26490618", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ONTUSD", "lowercaseId": "ontusd", "symbol": "ONT/USD", "base": "ONT", "quote": "USD", "settle": null, "baseId": "ONT", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 159092.19483322}}, "created": null, "info": {"symbol": "ONTUSD", "status": "BREAK", "baseAsset": "ONT", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "900000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "159092.19483322", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ZILUSD", "lowercaseId": "zilusd", "symbol": "ZIL/USD", "base": "ZIL", "quote": "USD", "settle": null, "baseId": "ZIL", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 4995249.66417651}}, "created": null, "info": {"symbol": "ZILUSD", "status": "BREAK", "baseAsset": "ZIL", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "4995249.66417651", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "VETUSD", "lowercaseId": "vetusd", "symbol": "VET/USD", "base": "VET", "quote": "USD", "settle": null, "baseId": "VET", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 9000000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 2248632.00368201}}, "created": null, "info": {"symbol": "VETUSD", "status": "BREAK", "baseAsset": "VET", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "9000000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "2248632.00368201", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "XTZUSD", "lowercaseId": "xtzusd", "symbol": "XTZ/USD", "base": "XTZ", "quote": "USD", "settle": null, "baseId": "XTZ", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 115671.22192361}}, "created": null, "info": {"symbol": "XTZUSD", "status": "BREAK", "baseAsset": "XTZ", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "115671.22192361", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "HBARUSD", "lowercaseId": "hbarusd", "symbol": "HBAR/USD", "base": "HBAR", "quote": "USD", "settle": null, "baseId": "HBAR", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 4686047.38798611}}, "created": null, "info": {"symbol": "HBARUSD", "status": "BREAK", "baseAsset": "HBAR", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "4686047.38798611", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "OMGUSD", "lowercaseId": "omgusd", "symbol": "OMG/USD", "base": "OMG", "quote": "USD", "settle": null, "baseId": "OMG", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 69571.07908617}}, "created": null, "info": {"symbol": "OMGUSD", "status": "BREAK", "baseAsset": "OMG", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "69571.07908617", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "MATICUSD", "lowercaseId": "maticusd", "symbol": "MATIC/USD", "base": "MATIC", "quote": "USD", "settle": null, "baseId": "MATIC", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 82437.30209205}}, "created": null, "info": {"symbol": "MATICUSD", "status": "BREAK", "baseAsset": "MATIC", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "82437.30209205", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "EOSUSD", "lowercaseId": "eosusd", "symbol": "EOS/USD", "base": "EOS", "quote": "USD", "settle": null, "baseId": "EOS", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 120578.75166782}}, "created": null, "info": {"symbol": "EOSUSD", "status": "BREAK", "baseAsset": "EOS", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "120578.75166782", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "DOGEUSD", "lowercaseId": "dogeusd", "symbol": "DOGE/USD", "base": "DOGE", "quote": "USD", "settle": null, "baseId": "DOGE", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 12487357.58200139}}, "created": null, "info": {"symbol": "DOGEUSD", "status": "BREAK", "baseAsset": "DOGE", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "12487357.58200139", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "KNCUSD", "lowercaseId": "kncusd", "symbol": "KNC/USD", "base": "KNC", "quote": "USD", "settle": null, "baseId": "KNC", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9222449.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 141590.56748471}}, "created": null, "info": {"symbol": "KNCUSD", "status": "BREAK", "baseAsset": "KNC", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9222449.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "141590.56748471", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "VTHOUSD", "lowercaseId": "vthousd", "symbol": "VTHO/USD", "base": "VTHO", "quote": "USD", "settle": null, "baseId": "VTHO", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-06, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 1e-06, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 92141578.0}}, "created": null, "info": {"symbol": "VTHOUSD", "status": "BREAK", "baseAsset": "VTHO", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000100", "maxPrice": "1000.00000000", "tickSize": "0.00000100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "92141578.00000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "USDCUSD", "lowercaseId": "usdcusd", "symbol": "USDC/USD", "base": "USDC", "quote": "USD", "settle": null, "baseId": "USDC", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 5208.87627615}}, "created": null, "info": {"symbol": "USDCUSD", "status": "BREAK", "baseAsset": "USDC", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "5208.87627615", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "COMPUSD", "lowercaseId": "compusd", "symbol": "COMP/USD", "base": "COMP", "quote": "USD", "settle": null, "baseId": "COMP", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 90000.0}, "price": {"min": 0.01, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 1125.52000532}}, "created": null, "info": {"symbol": "COMPUSD", "status": "BREAK", "baseAsset": "COMP", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "100000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "90000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1125.52000532", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "MANAUSD", "lowercaseId": "manausd", "symbol": "MANA/USD", "base": "MANA", "quote": "USD", "settle": null, "baseId": "MANA", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 209209.003287}}, "created": null, "info": {"symbol": "MANAUSD", "status": "BREAK", "baseAsset": "MANA", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "900000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "209209.00328700", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "HNTUSD", "lowercaseId": "hntusd", "symbol": "HNT/USD", "base": "HNT", "quote": "USD", "settle": null, "baseId": "HNT", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 16697.3326875}}, "created": null, "info": {"symbol": "HNTUSD", "status": "BREAK", "baseAsset": "HNT", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "10000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "16697.33268750", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "MKRUSD", "lowercaseId": "mkrusd", "symbol": "MKR/USD", "base": "MKR", "quote": "USD", "settle": null, "baseId": "MKR", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.0001, "price": 1.0, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.0001, "max": 90000.0}, "price": {"min": 1.0, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 125.26531854}}, "created": null, "info": {"symbol": "MKRUSD", "status": "BREAK", "baseAsset": "MKR", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "1.00000000", "maxPrice": "100000.00000000", "tickSize": "1.00000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00010000", "maxQty": "90000.00000000", "stepSize": "0.00010000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "125.26531854", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "DAIUSD", "lowercaseId": "daiusd", "symbol": "DAI/USD", "base": "DAI", "quote": "USD", "settle": null, "baseId": "DAI", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 343079.79620833}}, "created": null, "info": {"symbol": "DAIUSD", "status": "BREAK", "baseAsset": "DAI", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "343079.79620833", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ONEUSD", "lowercaseId": "oneusd", "symbol": "ONE/USD", "base": "ONE", "quote": "USD", "settle": null, "baseId": "ONE", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 9086742.86125781}}, "created": null, "info": {"symbol": "ONEUSD", "status": "BREAK", "baseAsset": "ONE", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "9086742.86125781", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BANDUSD", "lowercaseId": "bandusd", "symbol": "BAND/USD", "base": "BAND", "quote": "USD", "settle": null, "baseId": "BAND", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 90000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 23792.02808333}}, "created": null, "info": {"symbol": "BANDUSD", "status": "BREAK", "baseAsset": "BAND", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "90000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "23792.02808333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "STORJUSD", "lowercaseId": "storjusd", "symbol": "STORJ/USD", "base": "STORJ", "quote": "USD", "settle": null, "baseId": "STORJ", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9222448.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 185856.13733148}}, "created": null, "info": {"symbol": "STORJUSD", "status": "BREAK", "baseAsset": "STORJ", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9222448.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "185856.13733148", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "UNIUSD", "lowercaseId": "uniusd", "symbol": "UNI/USD", "base": "UNI", "quote": "USD", "settle": null, "baseId": "UNI", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 21188.32385337}}, "created": null, "info": {"symbol": "UNIUSD", "status": "BREAK", "baseAsset": "UNI", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "21188.32385337", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SOLUSD", "lowercaseId": "solusd", "symbol": "SOL/USD", "base": "SOL", "quote": "USD", "settle": null, "baseId": "SOL", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 6299.06732218}}, "created": null, "info": {"symbol": "SOLUSD", "status": "BREAK", "baseAsset": "SOL", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "10000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "6299.06732218", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "EGLDUSD", "lowercaseId": "egldusd", "symbol": "EGLD/USD", "base": "EGLD", "quote": "USD", "settle": null, "baseId": "EGLD", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.01, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 3151.13448646}}, "created": null, "info": {"symbol": "EGLDUSD", "status": "BREAK", "baseAsset": "EGLD", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "100000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3151.13448646", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "PAXGUSD", "lowercaseId": "paxgusd", "symbol": "PAXG/USD", "base": "PAXG", "quote": "USD", "settle": null, "baseId": "PAXG", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.0001, "price": 1.0, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.0001, "max": 9000.0}, "price": {"min": 1.0, "max": 1000000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 64.75339461}}, "created": null, "info": {"symbol": "PAXGUSD", "status": "BREAK", "baseAsset": "PAXG", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "1.00000000", "maxPrice": "1000000.00000000", "tickSize": "1.00000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00010000", "maxQty": "9000.00000000", "stepSize": "0.00010000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "64.75339461", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "OXTUSD", "lowercaseId": "oxtusd", "symbol": "OXT/USD", "base": "OXT", "quote": "USD", "settle": null, "baseId": "OXT", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 922327.0}, "price": {"min": 0.0001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 682136.82107366}}, "created": null, "info": {"symbol": "OXTUSD", "status": "BREAK", "baseAsset": "OXT", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "10000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "922327.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "682136.82107366", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ZENUSD", "lowercaseId": "zenusd", "symbol": "ZEN/USD", "base": "ZEN", "quote": "USD", "settle": null, "baseId": "ZEN", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 3922.88797292}}, "created": null, "info": {"symbol": "ZENUSD", "status": "BREAK", "baseAsset": "ZEN", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "10000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3922.88797292", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "FILUSD", "lowercaseId": "filusd", "symbol": "FIL/USD", "base": "FIL", "quote": "USD", "settle": null, "baseId": "FIL", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 9222449.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 6149.96227773}}, "created": null, "info": {"symbol": "FILUSD", "status": "BREAK", "baseAsset": "FIL", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "9222449.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "6149.96227773", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "AAVEUSD", "lowercaseId": "aaveusd", "symbol": "AAVE/USD", "base": "AAVE", "quote": "USD", "settle": null, "baseId": "AAVE", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.1, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 900000.0}, "price": {"min": 0.1, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 1584.77607542}}, "created": null, "info": {"symbol": "AAVEUSD", "status": "BREAK", "baseAsset": "AAVE", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.10000000", "maxPrice": "100000.00000000", "tickSize": "0.10000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "900000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1584.77607542", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "GRTUSD", "lowercaseId": "grtusd", "symbol": "GRT/USD", "base": "GRT", "quote": "USD", "settle": null, "baseId": "GRT", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 638136.67414871}}, "created": null, "info": {"symbol": "GRTUSD", "status": "BREAK", "baseAsset": "GRT", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "638136.67414871", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SUSHIUSD", "lowercaseId": "sushiusd", "symbol": "SUSHI/USD", "base": "SUSHI", "quote": "USD", "settle": null, "baseId": "SUSHI", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 90000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 122718.05450208}}, "created": null, "info": {"symbol": "SUSHIUSD", "status": "BREAK", "baseAsset": "SUSHI", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "90000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "122718.05450208", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ANKRUSD", "lowercaseId": "ankrusd", "symbol": "ANKR/USD", "base": "ANKR", "quote": "USD", "settle": null, "baseId": "ANKR", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 4075695.89444058}}, "created": null, "info": {"symbol": "ANKRUSD", "status": "BREAK", "baseAsset": "ANKR", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "4075695.89444058", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "CRVUSD", "lowercaseId": "crvusd", "symbol": "CRV/USD", "base": "CRV", "quote": "USD", "settle": null, "baseId": "CRV", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 922327.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 191361.34013889}}, "created": null, "info": {"symbol": "CRVUSD", "status": "BREAK", "baseAsset": "CRV", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "922327.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "191361.34013889", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "AXSUSD", "lowercaseId": "axsusd", "symbol": "AXS/USD", "base": "AXS", "quote": "USD", "settle": null, "baseId": "AXS", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.01, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 3434.86926736}}, "created": null, "info": {"symbol": "AXSUSD", "status": "BREAK", "baseAsset": "AXS", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "1000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3434.86926736", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "AVAXUSD", "lowercaseId": "avaxusd", "symbol": "AVAX/USD", "base": "AVAX", "quote": "USD", "settle": null, "baseId": "AVAX", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 10692.24099375}}, "created": null, "info": {"symbol": "AVAXUSD", "status": "BREAK", "baseAsset": "AVAX", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "10000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "10692.24099375", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "CTSIUSD", "lowercaseId": "ctsiusd", "symbol": "CTSI/USD", "base": "CTSI", "quote": "USD", "settle": null, "baseId": "CTSI", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 297335.32592078}}, "created": null, "info": {"symbol": "CTSIUSD", "status": "BREAK", "baseAsset": "CTSI", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "297335.32592078", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "DOTUSD", "lowercaseId": "dotusd", "symbol": "DOT/USD", "base": "DOT", "quote": "USD", "settle": null, "baseId": "DOT", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 40710.69671181}}, "created": null, "info": {"symbol": "DOTUSD", "status": "BREAK", "baseAsset": "DOT", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "10000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "40710.69671181", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "YFIUSD", "lowercaseId": "yfiusd", "symbol": "YFI/USD", "base": "YFI", "quote": "USD", "settle": null, "baseId": "YFI", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1e-05, "price": 1.0, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1e-05, "max": 9000.0}, "price": {"min": 1.0, "max": 1000000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 13.13710975}}, "created": null, "info": {"symbol": "YFIUSD", "status": "BREAK", "baseAsset": "YFI", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "1.00000000", "maxPrice": "1000000.00000000", "tickSize": "1.00000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00001000", "maxQty": "9000.00000000", "stepSize": "0.00001000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "13.13710975", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "1INCHUSD", "lowercaseId": "1inchusd", "symbol": "1INCH/USD", "base": "1INCH", "quote": "USD", "settle": null, "baseId": "1INCH", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 67278.2319319}}, "created": null, "info": {"symbol": "1INCHUSD", "status": "BREAK", "baseAsset": "1INCH", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "67278.23193190", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "FTMUSD", "lowercaseId": "ftmusd", "symbol": "FTM/USD", "base": "FTM", "quote": "USD", "settle": null, "baseId": "FTM", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 365811.70048645}}, "created": null, "info": {"symbol": "FTMUSD", "status": "BREAK", "baseAsset": "FTM", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "365811.70048645", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "NEARUSD", "lowercaseId": "nearusd", "symbol": "NEAR/USD", "base": "NEAR", "quote": "USD", "settle": null, "baseId": "NEAR", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 219157.5346875}}, "created": null, "info": {"symbol": "NEARUSD", "status": "BREAK", "baseAsset": "NEAR", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "219157.53468750", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LRCUSD", "lowercaseId": "lrcusd", "symbol": "LRC/USD", "base": "LRC", "quote": "USD", "settle": null, "baseId": "LRC", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 569490.2625}}, "created": null, "info": {"symbol": "LRCUSD", "status": "BREAK", "baseAsset": "LRC", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "569490.26250000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "KSHIBUSD", "lowercaseId": "kshibusd", "symbol": "KSHIB/USD", "base": "KSHIB", "quote": "USD", "settle": null, "baseId": "KSHIB", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 0.0001, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 46116860.0}, "price": {"min": 0.0001, "max": 3.0}, "cost": {"min": 10.0, "max": null}, "market": {"min": 0.0, "max": 46116860.0}}, "created": null, "info": {"symbol": "KSHIBUSD", "status": "BREAK", "baseAsset": "KSHIB", "baseAssetPrecision": "4", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "4", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "3.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.0000", "maxQty": "46116860.0000", "stepSize": "1.0000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "10.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.0000", "maxQty": "46116860.0000", "stepSize": "0.0000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LPTUSD", "lowercaseId": "lptusd", "symbol": "LPT/USD", "base": "LPT", "quote": "USD", "settle": null, "baseId": "LPT", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 9222440.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 6853.50953472}}, "created": null, "info": {"symbol": "LPTUSD", "status": "BREAK", "baseAsset": "LPT", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "10000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "9222440.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "6853.50953472", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "NMRUSD", "lowercaseId": "nmrusd", "symbol": "NMR/USD", "base": "NMR", "quote": "USD", "settle": null, "baseId": "NMR", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 4802.3121721}}, "created": null, "info": {"symbol": "NMRUSD", "status": "BREAK", "baseAsset": "NMR", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "10000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "4802.31217210", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SLPUSD", "lowercaseId": "slpusd", "symbol": "SLP/USD", "base": "SLP", "quote": "USD", "settle": null, "baseId": "SLP", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-06, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 90000000.0}, "price": {"min": 1e-06, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 92141578.0}}, "created": null, "info": {"symbol": "SLPUSD", "status": "BREAK", "baseAsset": "SLP", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000100", "maxPrice": "1000.00000000", "tickSize": "0.00000100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "90000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "92141578.00000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ANTUSD", "lowercaseId": "antusd", "symbol": "ANT/USD", "base": "ANT", "quote": "USD", "settle": null, "baseId": "ANT", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 90347.52763888}}, "created": null, "info": {"symbol": "ANTUSD", "status": "BREAK", "baseAsset": "ANT", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "90347.52763888", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "XNOUSD", "lowercaseId": "xnousd", "symbol": "XNO/USD", "base": "XNO", "quote": "USD", "settle": null, "baseId": "XNO", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 36240.93046527}}, "created": null, "info": {"symbol": "XNOUSD", "status": "BREAK", "baseAsset": "XNO", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "36240.93046527", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "CHZUSD", "lowercaseId": "chzusd", "symbol": "CHZ/USD", "base": "CHZ", "quote": "USD", "settle": null, "baseId": "CHZ", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 2767789.21666666}}, "created": null, "info": {"symbol": "CHZUSD", "status": "BREAK", "baseAsset": "CHZ", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "2767789.21666666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "OGNUSD", "lowercaseId": "ognusd", "symbol": "OGN/USD", "base": "OGN", "quote": "USD", "settle": null, "baseId": "OGN", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 569534.45726198}}, "created": null, "info": {"symbol": "OGNUSD", "status": "BREAK", "baseAsset": "OGN", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "569534.45726198", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "GALAUSD", "lowercaseId": "galausd", "symbol": "GALA/USD", "base": "GALA", "quote": "USD", "settle": null, "baseId": "GALA", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 10519306.87074357}}, "created": null, "info": {"symbol": "GALAUSD", "status": "BREAK", "baseAsset": "GALA", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "10519306.87074357", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "TLMUSD", "lowercaseId": "tlmusd", "symbol": "TLM/USD", "base": "TLM", "quote": "USD", "settle": null, "baseId": "TLM", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 4720997.96455872}}, "created": null, "info": {"symbol": "TLMUSD", "status": "BREAK", "baseAsset": "TLM", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "900000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "4720997.96455872", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SNXUSD", "lowercaseId": "snxusd", "symbol": "SNX/USD", "base": "SNX", "quote": "USD", "settle": null, "baseId": "SNX", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 90000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 39955.67692842}}, "created": null, "info": {"symbol": "SNXUSD", "status": "BREAK", "baseAsset": "SNX", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "90000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "39955.67692842", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "AUDIOUSD", "lowercaseId": "audiousd", "symbol": "AUDIO/USD", "base": "AUDIO", "quote": "USD", "settle": null, "baseId": "AUDIO", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 241109.39353717}}, "created": null, "info": {"symbol": "AUDIOUSD", "status": "BREAK", "baseAsset": "AUDIO", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "241109.39353717", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "REQUSD", "lowercaseId": "requsd", "symbol": "REQ/USD", "base": "REQ", "quote": "USD", "settle": null, "baseId": "REQ", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 933555.52605976}}, "created": null, "info": {"symbol": "REQUSD", "status": "BREAK", "baseAsset": "REQ", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "933555.52605976", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "APEUSD", "lowercaseId": "apeusd", "symbol": "APE/USD", "base": "APE", "quote": "USD", "settle": null, "baseId": "APE", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 92141578.0}}, "created": null, "info": {"symbol": "APEUSD", "status": "BREAK", "baseAsset": "APE", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "92141578.00000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "FLUXUSD", "lowercaseId": "fluxusd", "symbol": "FLUX/USD", "base": "FLUX", "quote": "USD", "settle": null, "baseId": "FLUX", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 175409.69921473}}, "created": null, "info": {"symbol": "FLUXUSD", "status": "BREAK", "baseAsset": "FLUX", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "175409.69921473", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "TRXUSD", "lowercaseId": "trxusd", "symbol": "TRX/USD", "base": "TRX", "quote": "USD", "settle": null, "baseId": "TRX", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 20995303.38346073}}, "created": null, "info": {"symbol": "TRXUSD", "status": "BREAK", "baseAsset": "TRX", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "20995303.38346073", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "COTIUSD", "lowercaseId": "cotiusd", "symbol": "COTI/USD", "base": "COTI", "quote": "USD", "settle": null, "baseId": "COTI", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 1820701.29186935}}, "created": null, "info": {"symbol": "COTIUSD", "status": "BREAK", "baseAsset": "COTI", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1820701.29186935", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "VOXELUSD", "lowercaseId": "voxelusd", "symbol": "VOXEL/USD", "base": "VOXEL", "quote": "USD", "settle": null, "baseId": "VOXEL", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 351360.30125086}}, "created": null, "info": {"symbol": "VOXELUSD", "status": "BREAK", "baseAsset": "VOXEL", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "351360.30125086", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "RLCUSD", "lowercaseId": "rlcusd", "symbol": "RLC/USD", "base": "RLC", "quote": "USD", "settle": null, "baseId": "RLC", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9222449.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 53872.75618055}}, "created": null, "info": {"symbol": "RLCUSD", "status": "BREAK", "baseAsset": "RLC", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9222449.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "53872.75618055", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BICOUSD", "lowercaseId": "bicousd", "symbol": "BICO/USD", "base": "BICO", "quote": "USD", "settle": null, "baseId": "BICO", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 146093.6861779}}, "created": null, "info": {"symbol": "BICOUSD", "status": "BREAK", "baseAsset": "BICO", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "146093.68617790", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "API3USD", "lowercaseId": "api3usd", "symbol": "API3/USD", "base": "API3", "quote": "USD", "settle": null, "baseId": "API3", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 94155.98463194}}, "created": null, "info": {"symbol": "API3USD", "status": "BREAK", "baseAsset": "API3", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "94155.98463194", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ENSUSD", "lowercaseId": "ensusd", "symbol": "ENS/USD", "base": "ENS", "quote": "USD", "settle": null, "baseId": "ENS", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 9222449.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 33626.31374565}}, "created": null, "info": {"symbol": "ENSUSD", "status": "BREAK", "baseAsset": "ENS", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "10000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "9222449.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "33626.31374565", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BNTUSD", "lowercaseId": "bntusd", "symbol": "BNT/USD", "base": "BNT", "quote": "USD", "settle": null, "baseId": "BNT", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 34504.70986796}}, "created": null, "info": {"symbol": "BNTUSD", "status": "BREAK", "baseAsset": "BNT", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "34504.70986796", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "IMXUSD", "lowercaseId": "imxusd", "symbol": "IMX/USD", "base": "IMX", "quote": "USD", "settle": null, "baseId": "IMX", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 273678.31160416}}, "created": null, "info": {"symbol": "IMXUSD", "status": "BREAK", "baseAsset": "IMX", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "273678.31160416", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SPELLUSD", "lowercaseId": "spellusd", "symbol": "SPELL/USD", "base": "SPELL", "quote": "USD", "settle": null, "baseId": "SPELL", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-07, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 1e-07, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 63442841.67060458}}, "created": null, "info": {"symbol": "SPELLUSD", "status": "BREAK", "baseAsset": "SPELL", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000010", "maxPrice": "1000.00000000", "tickSize": "0.00000010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "63442841.67060458", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "JASMYUSD", "lowercaseId": "jasmyusd", "symbol": "JASMY/USD", "base": "JASMY", "quote": "USD", "settle": null, "baseId": "JASMY", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-06, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 1e-06, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 92141578.0}}, "created": null, "info": {"symbol": "JASMYUSD", "status": "BREAK", "baseAsset": "JASMY", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000100", "maxPrice": "1000.00000000", "tickSize": "0.00000100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "92141578.00000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "FLOWUSD", "lowercaseId": "flowusd", "symbol": "FLOW/USD", "base": "FLOW", "quote": "USD", "settle": null, "baseId": "FLOW", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 922327.0}, "price": {"min": 0.001, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 94183.15727966}}, "created": null, "info": {"symbol": "FLOWUSD", "status": "BREAK", "baseAsset": "FLOW", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "100000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "922327.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "94183.15727966", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "GTCUSD", "lowercaseId": "gtcusd", "symbol": "GTC/USD", "base": "GTC", "quote": "USD", "settle": null, "baseId": "GTC", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9222449.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 23724.14420541}}, "created": null, "info": {"symbol": "GTCUSD", "status": "BREAK", "baseAsset": "GTC", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9222449.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "23724.14420541", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "THETAUSD", "lowercaseId": "thetausd", "symbol": "THETA/USD", "base": "THETA", "quote": "USD", "settle": null, "baseId": "THETA", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 90000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 158385.32960389}}, "created": null, "info": {"symbol": "THETAUSD", "status": "BREAK", "baseAsset": "THETA", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "90000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "158385.32960389", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "TFUELUSD", "lowercaseId": "tfuelusd", "symbol": "TFUEL/USD", "base": "TFUEL", "quote": "USD", "settle": null, "baseId": "TFUEL", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 827041.82861111}}, "created": null, "info": {"symbol": "TFUELUSD", "status": "BREAK", "baseAsset": "TFUEL", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "827041.82861111", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "OCEANUSD", "lowercaseId": "oceanusd", "symbol": "OCEAN/USD", "base": "OCEAN", "quote": "USD", "settle": null, "baseId": "OCEAN", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 145997.2870229}}, "created": null, "info": {"symbol": "OCEANUSD", "status": "BREAK", "baseAsset": "OCEAN", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "900000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "145997.28702290", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LAZIOUSD", "lowercaseId": "laziousd", "symbol": "LAZIO/USD", "base": "LAZIO", "quote": "USD", "settle": null, "baseId": "LAZIO", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 922327.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 30354.65209027}}, "created": null, "info": {"symbol": "LAZIOUSD", "status": "BREAK", "baseAsset": "LAZIO", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "922327.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "30354.65209027", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SANTOSUSD", "lowercaseId": "santosusd", "symbol": "SANTOS/USD", "base": "SANTOS", "quote": "USD", "settle": null, "baseId": "SANTOS", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 26075.60693268}}, "created": null, "info": {"symbol": "SANTOSUSD", "status": "BREAK", "baseAsset": "SANTOS", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "26075.60693268", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ALPINEUSD", "lowercaseId": "alpineusd", "symbol": "ALPINE/USD", "base": "ALPINE", "quote": "USD", "settle": null, "baseId": "ALPINE", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 19539.58060458}}, "created": null, "info": {"symbol": "ALPINEUSD", "status": "BREAK", "baseAsset": "ALPINE", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "19539.58060458", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "PORTOUSD", "lowercaseId": "portousd", "symbol": "PORTO/USD", "base": "PORTO", "quote": "USD", "settle": null, "baseId": "PORTO", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 29530.82736805}}, "created": null, "info": {"symbol": "PORTOUSD", "status": "BREAK", "baseAsset": "PORTO", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "29530.82736805", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "RENUSD", "lowercaseId": "renusd", "symbol": "REN/USD", "base": "REN", "quote": "USD", "settle": null, "baseId": "REN", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-06, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 1e-06, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 567579.47108066}}, "created": null, "info": {"symbol": "RENUSD", "status": "BREAK", "baseAsset": "REN", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000100", "maxPrice": "1000.00000000", "tickSize": "0.00000100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "567579.47108066", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "CELRUSD", "lowercaseId": "celrusd", "symbol": "CELR/USD", "base": "CELR", "quote": "USD", "settle": null, "baseId": "CELR", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9222449.0}, "price": {"min": 1e-05, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 7822685.38698578}}, "created": null, "info": {"symbol": "CELRUSD", "status": "BREAK", "baseAsset": "CELR", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "10000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9222449.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "7822685.38698578", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SKLUSD", "lowercaseId": "sklusd", "symbol": "SKL/USD", "base": "SKL", "quote": "USD", "settle": null, "baseId": "SKL", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 1759560.96742957}}, "created": null, "info": {"symbol": "SKLUSD", "status": "BREAK", "baseAsset": "SKL", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1759560.96742957", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "VITEUSD", "lowercaseId": "viteusd", "symbol": "VITE/USD", "base": "VITE", "quote": "USD", "settle": null, "baseId": "VITE", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 2860785.40650263}}, "created": null, "info": {"symbol": "VITEUSD", "status": "BREAK", "baseAsset": "VITE", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "2860785.40650263", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "WAXPUSD", "lowercaseId": "waxpusd", "symbol": "WAXP/USD", "base": "WAXP", "quote": "USD", "settle": null, "baseId": "WAXP", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 551927.59930555}}, "created": null, "info": {"symbol": "WAXPUSD", "status": "BREAK", "baseAsset": "WAXP", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "551927.59930555", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LTOUSD", "lowercaseId": "ltousd", "symbol": "LTO/USD", "base": "LTO", "quote": "USD", "settle": null, "baseId": "LTO", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 1042269.79360667}}, "created": null, "info": {"symbol": "LTOUSD", "status": "BREAK", "baseAsset": "LTO", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1042269.79360667", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "FETUSD", "lowercaseId": "fetusd", "symbol": "FET/USD", "base": "FET", "quote": "USD", "settle": null, "baseId": "FET", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 679752.96942321}}, "created": null, "info": {"symbol": "FETUSD", "status": "BREAK", "baseAsset": "FET", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "679752.96942321", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BONDUSD", "lowercaseId": "bondusd", "symbol": "BOND/USD", "base": "BOND", "quote": "USD", "settle": null, "baseId": "BOND", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 9222449.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 16758.01062107}}, "created": null, "info": {"symbol": "BONDUSD", "status": "BREAK", "baseAsset": "BOND", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "9222449.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "16758.01062107", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LOKAUSD", "lowercaseId": "lokausd", "symbol": "LOKA/USD", "base": "LOKA", "quote": "USD", "settle": null, "baseId": "LOKA", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 136861.61279554}}, "created": null, "info": {"symbol": "LOKAUSD", "status": "BREAK", "baseAsset": "LOKA", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "136861.61279554", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ICPUSD", "lowercaseId": "icpusd", "symbol": "ICP/USD", "base": "ICP", "quote": "USD", "settle": null, "baseId": "ICP", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 922327.0}, "price": {"min": 0.001, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 86715.77427777}}, "created": null, "info": {"symbol": "ICPUSD", "status": "BREAK", "baseAsset": "ICP", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "100000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "922327.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "86715.77427777", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "TUSD", "lowercaseId": "tusd", "symbol": "T/USD", "base": "T", "quote": "USD", "settle": null, "baseId": "T", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 3671409.47795549}}, "created": null, "info": {"symbol": "TUSD", "status": "BREAK", "baseAsset": "T", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3671409.47795549", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "OPUSD", "lowercaseId": "opusd", "symbol": "OP/USD", "base": "OP", "quote": "USD", "settle": null, "baseId": "OP", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 1154139.26001388}}, "created": null, "info": {"symbol": "OPUSD", "status": "BREAK", "baseAsset": "OP", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1154139.26001388", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ROSEUSD", "lowercaseId": "roseusd", "symbol": "ROSE/USD", "base": "ROSE", "quote": "USD", "settle": null, "baseId": "ROSE", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 6038736.71125781}}, "created": null, "info": {"symbol": "ROSEUSD", "status": "BREAK", "baseAsset": "ROSE", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "6038736.71125781", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "CELOUSD", "lowercaseId": "celousd", "symbol": "CELO/USD", "base": "CELO", "quote": "USD", "settle": null, "baseId": "CELO", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 191076.78429464}}, "created": null, "info": {"symbol": "CELOUSD", "status": "BREAK", "baseAsset": "CELO", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "191076.78429464", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "KDAUSD", "lowercaseId": "kdausd", "symbol": "KDA/USD", "base": "KDA", "quote": "USD", "settle": null, "baseId": "KDA", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 130664.17344537}}, "created": null, "info": {"symbol": "KDAUSD", "status": "BREAK", "baseAsset": "KDA", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "130664.17344537", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "KSMUSD", "lowercaseId": "ksmusd", "symbol": "KSM/USD", "base": "KSM", "quote": "USD", "settle": null, "baseId": "KSM", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 900000.0}, "price": {"min": 0.01, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 3294.32087482}}, "created": null, "info": {"symbol": "KSMUSD", "status": "BREAK", "baseAsset": "KSM", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "100000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "900000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "3294.32087482", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ACHUSD", "lowercaseId": "achusd", "symbol": "ACH/USD", "base": "ACH", "quote": "USD", "settle": null, "baseId": "ACH", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 9285110.73706293}}, "created": null, "info": {"symbol": "ACHUSD", "status": "BREAK", "baseAsset": "ACH", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "9285110.73706293", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "DARUSD", "lowercaseId": "darusd", "symbol": "DAR/USD", "base": "DAR", "quote": "USD", "settle": null, "baseId": "DAR", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 593200.79861111}}, "created": null, "info": {"symbol": "DARUSD", "status": "BREAK", "baseAsset": "DAR", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "593200.79861111", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "RNDRUSD", "lowercaseId": "rndrusd", "symbol": "RNDR/USD", "base": "RNDR", "quote": "USD", "settle": null, "baseId": "RNDR", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 161822.84313671}}, "created": null, "info": {"symbol": "RNDRUSD", "status": "BREAK", "baseAsset": "RNDR", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "161822.84313671", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SYSUSD", "lowercaseId": "sysusd", "symbol": "SYS/USD", "base": "SYS", "quote": "USD", "settle": null, "baseId": "SYS", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 966695.58905058}}, "created": null, "info": {"symbol": "SYSUSD", "status": "BREAK", "baseAsset": "SYS", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "966695.58905058", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "RADUSD", "lowercaseId": "radusd", "symbol": "RAD/USD", "base": "RAD", "quote": "USD", "settle": null, "baseId": "RAD", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 47546.77801664}}, "created": null, "info": {"symbol": "RADUSD", "status": "BREAK", "baseAsset": "RAD", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "47546.77801664", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ILVUSD", "lowercaseId": "ilvusd", "symbol": "ILV/USD", "base": "ILV", "quote": "USD", "settle": null, "baseId": "ILV", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 922327.0}, "price": {"min": 0.01, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 916.93937734}}, "created": null, "info": {"symbol": "ILVUSD", "status": "BREAK", "baseAsset": "ILV", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "100000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "922327.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "916.93937734", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LDOUSD", "lowercaseId": "ldousd", "symbol": "LDO/USD", "base": "LDO", "quote": "USD", "settle": null, "baseId": "LDO", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 316622.60368932}}, "created": null, "info": {"symbol": "LDOUSD", "status": "BREAK", "baseAsset": "LDO", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "316622.60368932", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "RAREUSD", "lowercaseId": "rareusd", "symbol": "RARE/USD", "base": "RARE", "quote": "USD", "settle": null, "baseId": "RARE", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 390621.67142857}}, "created": null, "info": {"symbol": "RAREUSD", "status": "BREAK", "baseAsset": "RARE", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "390621.67142857", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LSKUSD", "lowercaseId": "lskusd", "symbol": "LSK/USD", "base": "LSK", "quote": "USD", "settle": null, "baseId": "LSK", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 62355.45987525}}, "created": null, "info": {"symbol": "LSKUSD", "status": "BREAK", "baseAsset": "LSK", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "62355.45987525", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "DGBUSD", "lowercaseId": "dgbusd", "symbol": "DGB/USD", "base": "DGB", "quote": "USD", "settle": null, "baseId": "DGB", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 6176501.64299583}}, "created": null, "info": {"symbol": "DGBUSD", "status": "BREAK", "baseAsset": "DGB", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "6176501.64299583", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "REEFUSD", "lowercaseId": "reefusd", "symbol": "REEF/USD", "base": "REEF", "quote": "USD", "settle": null, "baseId": "REEF", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-06, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 1e-06, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 19979356.49271339}}, "created": null, "info": {"symbol": "REEFUSD", "status": "BREAK", "baseAsset": "REEF", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000100", "maxPrice": "1000.00000000", "tickSize": "0.00000100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "19979356.49271339", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ALICEUSD", "lowercaseId": "aliceusd", "symbol": "ALICE/USD", "base": "ALICE", "quote": "USD", "settle": null, "baseId": "ALICE", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 90594.63785962}}, "created": null, "info": {"symbol": "ALICEUSD", "status": "BREAK", "baseAsset": "ALICE", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "90594.63785962", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "FORTHUSD", "lowercaseId": "forthusd", "symbol": "FORTH/USD", "base": "FORTH", "quote": "USD", "settle": null, "baseId": "FORTH", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 42037.71566666}}, "created": null, "info": {"symbol": "FORTHUSD", "status": "BREAK", "baseAsset": "FORTH", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "42037.71566666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ASTRUSD", "lowercaseId": "astrusd", "symbol": "ASTR/USD", "base": "ASTR", "quote": "USD", "settle": null, "baseId": "ASTR", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 2998944.25447605}}, "created": null, "info": {"symbol": "ASTRUSD", "status": "BREAK", "baseAsset": "ASTR", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "2998944.25447605", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BTRSTUSD", "lowercaseId": "btrstusd", "symbol": "BTRST/USD", "base": "BTRST", "quote": "USD", "settle": null, "baseId": "BTRST", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9222449.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 63083.19006944}}, "created": null, "info": {"symbol": "BTRSTUSD", "status": "BREAK", "baseAsset": "BTRST", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9222449.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "63083.19006944", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "GALUSD", "lowercaseId": "galusd", "symbol": "GAL/USD", "base": "GAL", "quote": "USD", "settle": null, "baseId": "GAL", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 154650.16084403}}, "created": null, "info": {"symbol": "GALUSD", "status": "BREAK", "baseAsset": "GAL", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "92141578.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "154650.16084403", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SANDUSD", "lowercaseId": "sandusd", "symbol": "SAND/USD", "base": "SAND", "quote": "USD", "settle": null, "baseId": "SAND", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 928454.74068965}}, "created": null, "info": {"symbol": "SANDUSD", "status": "BREAK", "baseAsset": "SAND", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "928454.74068965", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BALUSD", "lowercaseId": "balusd", "symbol": "BAL/USD", "base": "BAL", "quote": "USD", "settle": null, "baseId": "BAL", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 18684.8525625}}, "created": null, "info": {"symbol": "BALUSD", "status": "BREAK", "baseAsset": "BAL", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "18684.85256250", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "POLYXUSD", "lowercaseId": "polyxusd", "symbol": "POLYX/USD", "base": "POLYX", "quote": "USD", "settle": null, "baseId": "POLYX", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 92141578.0}}, "created": null, "info": {"symbol": "POLYXUSD", "status": "BREAK", "baseAsset": "POLYX", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "92141578.00000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "GLMUSD", "lowercaseId": "glmusd", "symbol": "GLM/USD", "base": "GLM", "quote": "USD", "settle": null, "baseId": "GLM", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 106586.10466898}}, "created": null, "info": {"symbol": "GLMUSD", "status": "BREAK", "baseAsset": "GLM", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "106586.10466898", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "CLVUSD", "lowercaseId": "clvusd", "symbol": "CLV/USD", "base": "CLV", "quote": "USD", "settle": null, "baseId": "CLV", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 1341583.34923398}}, "created": null, "info": {"symbol": "CLVUSD", "status": "BREAK", "baseAsset": "CLV", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1341583.34923398", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "TUSDUSD", "lowercaseId": "tusdusd", "symbol": "TUSD/USD", "base": "TUSD", "quote": "USD", "settle": null, "baseId": "TUSD", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 1301033.56392265}}, "created": null, "info": {"symbol": "TUSDUSD", "status": "BREAK", "baseAsset": "TUSD", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "1.2", "multiplierDown": "0.8", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1301033.56392265", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "QNTUSD", "lowercaseId": "qntusd", "symbol": "QNT/USD", "base": "QNT", "quote": "USD", "settle": null, "baseId": "QNT", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.1, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 922327.0}, "price": {"min": 0.1, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 6389.58374712}}, "created": null, "info": {"symbol": "QNTUSD", "status": "BREAK", "baseAsset": "QNT", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.10000000", "maxPrice": "100000.00000000", "tickSize": "0.10000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "922327.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "6389.58374712", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "STGUSD", "lowercaseId": "stgusd", "symbol": "STG/USD", "base": "STG", "quote": "USD", "settle": null, "baseId": "STG", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 913205152.0}, "price": {"min": 0.0001, "max": 100.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 260651.43672883}}, "created": null, "info": {"symbol": "STGUSD", "status": "BREAK", "baseAsset": "STG", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "100.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "913205152.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "260651.43672883", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "AXLUSD", "lowercaseId": "axlusd", "symbol": "AXL/USD", "base": "AXL", "quote": "USD", "settle": null, "baseId": "AXL", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 922327.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 511719.87214644}}, "created": null, "info": {"symbol": "AXLUSD", "status": "BREAK", "baseAsset": "AXL", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "922327.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "511719.87214644", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "KAVAUSD", "lowercaseId": "kavausd", "symbol": "KAVA/USD", "base": "KAVA", "quote": "USD", "settle": null, "baseId": "KAVA", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9222449.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 103605.40152883}}, "created": null, "info": {"symbol": "KAVAUSD", "status": "BREAK", "baseAsset": "KAVA", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9222449.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "103605.40152883", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "APTUSD", "lowercaseId": "aptusd", "symbol": "APT/USD", "base": "APT", "quote": "USD", "settle": null, "baseId": "APT", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 131395.7994371}}, "created": null, "info": {"symbol": "APTUSD", "status": "BREAK", "baseAsset": "APT", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "131395.79943710", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "MASKUSD", "lowercaseId": "maskusd", "symbol": "MASK/USD", "base": "MASK", "quote": "USD", "settle": null, "baseId": "MASK", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9222449.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 146077.9675}}, "created": null, "info": {"symbol": "MASKUSD", "status": "BREAK", "baseAsset": "MASK", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9222449.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "146077.96750000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BOSONUSD", "lowercaseId": "bosonusd", "symbol": "BOSON/USD", "base": "BOSON", "quote": "USD", "settle": null, "baseId": "BOSON", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 96324.3375}}, "created": null, "info": {"symbol": "BOSONUSD", "status": "BREAK", "baseAsset": "BOSON", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "96324.33750000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "PONDUSD", "lowercaseId": "pondusd", "symbol": "POND/USD", "base": "POND", "quote": "USD", "settle": null, "baseId": "POND", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 9222449.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 9222449.0}}, "created": null, "info": {"symbol": "PONDUSD", "status": "BREAK", "baseAsset": "POND", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "9222449.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "9222449.00000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "MXCUSD", "lowercaseId": "mxcusd", "symbol": "MXC/USD", "base": "MXC", "quote": "USD", "settle": null, "baseId": "MXC", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9222449.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 9222449.0}}, "created": null, "info": {"symbol": "MXCUSD", "status": "BREAK", "baseAsset": "MXC", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9222449.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "9222449.00000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "JAMUSD", "lowercaseId": "jamusd", "symbol": "JAM/USD", "base": "JAM", "quote": "USD", "settle": null, "baseId": "JAM", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-06, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 90000000.0}, "price": {"min": 1e-06, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 67578970.50451702}}, "created": null, "info": {"symbol": "JAMUSD", "status": "BREAK", "baseAsset": "JAM", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000100", "maxPrice": "1000.00000000", "tickSize": "0.00000100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "90000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "67578970.50451702", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "PROMUSD", "lowercaseId": "promusd", "symbol": "PROM/USD", "base": "PROM", "quote": "USD", "settle": null, "baseId": "PROM", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.01, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 31980.2771369}}, "created": null, "info": {"symbol": "PROMUSD", "status": "BREAK", "baseAsset": "PROM", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "10000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "31980.27713690", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "DIAUSD", "lowercaseId": "diausd", "symbol": "DIA/USD", "base": "DIA", "quote": "USD", "settle": null, "baseId": "DIA", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 922327.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 300013.60541666}}, "created": null, "info": {"symbol": "DIAUSD", "status": "BREAK", "baseAsset": "DIA", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "922327.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "300013.60541666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "LOOMUSD", "lowercaseId": "loomusd", "symbol": "LOOM/USD", "base": "LOOM", "quote": "USD", "settle": null, "baseId": "LOOM", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 1060513.60833333}}, "created": null, "info": {"symbol": "LOOMUSD", "status": "BREAK", "baseAsset": "LOOM", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1060513.60833333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "STMXUSD", "lowercaseId": "stmxusd", "symbol": "STMX/USD", "base": "STMX", "quote": "USD", "settle": null, "baseId": "STMX", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-06, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 1e-06, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 18312721.08819444}}, "created": null, "info": {"symbol": "STMXUSD", "status": "BREAK", "baseAsset": "STMX", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000100", "maxPrice": "1000.00000000", "tickSize": "0.00000100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "18312721.08819444", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SHIBUSD", "lowercaseId": "shibusd", "symbol": "SHIB/USD", "base": "SHIB", "quote": "USD", "settle": null, "baseId": "SHIB", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-08, "cost": null, "base": 0.01, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 46116860414.0}, "price": {"min": 1e-08, "max": 1.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 46116860414.0}}, "created": null, "info": {"symbol": "SHIBUSD", "status": "BREAK", "baseAsset": "SHIB", "baseAssetPrecision": "2", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "2", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000001", "maxPrice": "1.00000000", "tickSize": "0.00000001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00", "maxQty": "46116860414.00", "stepSize": "1.00"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00", "maxQty": "46116860414.00", "stepSize": "0.00"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "TRACUSD", "lowercaseId": "tracusd", "symbol": "TRAC/USD", "base": "TRAC", "quote": "USD", "settle": null, "baseId": "TRAC", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 4394577.71805555}}, "created": null, "info": {"symbol": "TRACUSD", "status": "BREAK", "baseAsset": "TRAC", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "4394577.71805555", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "POLYXUSDT", "lowercaseId": "polyxusdt", "symbol": "POLYX/USDT", "base": "POLYX", "quote": "USDT", "settle": null, "baseId": "POLYX", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 60027.04458333}}, "created": null, "info": {"symbol": "POLYXUSDT", "status": "TRADING", "baseAsset": "POLYX", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "60027.04458333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "IOSTUSDT", "lowercaseId": "iostusdt", "symbol": "IOST/USDT", "base": "IOST", "quote": "USDT", "settle": null, "baseId": "IOST", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 358952.27083333}}, "created": null, "info": {"symbol": "IOSTUSDT", "status": "TRADING", "baseAsset": "IOST", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "358952.27083333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "IOSTUSD", "lowercaseId": "iostusd", "symbol": "IOST/USD", "base": "IOST", "quote": "USD", "settle": null, "baseId": "IOST", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 12581951.96041666}}, "created": null, "info": {"symbol": "IOSTUSD", "status": "BREAK", "baseAsset": "IOST", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "12581951.96041666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "MATICETH", "lowercaseId": "maticeth", "symbol": "MATIC/ETH", "base": "MATIC", "quote": "ETH", "settle": null, "baseId": "MATIC", "quoteId": "ETH", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-07, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 1e-07, "max": 1000.0}, "cost": {"min": 0.005, "max": null}, "market": {"min": 0.0, "max": 299.75541666}}, "created": null, "info": {"symbol": "MATICETH", "status": "TRADING", "baseAsset": "MATIC", "baseAssetPrecision": "8", "quoteAsset": "ETH", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000010", "maxPrice": "1000.00000000", "tickSize": "0.00000010"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "0.00500000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "299.75541666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SOLETH", "lowercaseId": "soleth", "symbol": "SOL/ETH", "base": "SOL", "quote": "ETH", "settle": null, "baseId": "SOL", "quoteId": "ETH", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 92141578.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 0.005, "max": null}, "market": {"min": 0.0, "max": 68.9477875}}, "created": null, "info": {"symbol": "SOLETH", "status": "TRADING", "baseAsset": "SOL", "baseAssetPrecision": "8", "quoteAsset": "ETH", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "92141578.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "0.00500000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "68.94778750", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "FLOKI8USDT", "lowercaseId": "floki8usdt", "symbol": "FLOKI8/USDT", "base": "FLOKI8", "quote": "USDT", "settle": null, "baseId": "FLOKI8", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-08, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 46116860414.0}, "price": {"min": 1e-08, "max": 1.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 46116860414.0}}, "created": null, "info": {"symbol": "FLOKI8USDT", "status": "BREAK", "baseAsset": "FLOKI8", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "2", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000001", "maxPrice": "1.00000000", "tickSize": "0.00000001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "46116860414.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "46116860414.00000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "FLOKI8USD", "lowercaseId": "floki8usd", "symbol": "FLOKI8/USD", "base": "FLOKI8", "quote": "USD", "settle": null, "baseId": "FLOKI8", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-08, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 46116860414.0}, "price": {"min": 1e-08, "max": 1.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 46116860414.0}}, "created": null, "info": {"symbol": "FLOKI8USD", "status": "BREAK", "baseAsset": "FLOKI8", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "2", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000001", "maxPrice": "1.00000000", "tickSize": "0.00000001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "46116860414.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "46116860414.00000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ARBUSDT", "lowercaseId": "arbusdt", "symbol": "ARB/USDT", "base": "ARB", "quote": "USDT", "settle": null, "baseId": "ARB", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 18030.15166666}}, "created": null, "info": {"symbol": "ARBUSDT", "status": "TRADING", "baseAsset": "ARB", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "18030.15166666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ARBUSD", "lowercaseId": "arbusd", "symbol": "ARB/USD", "base": "ARB", "quote": "USD", "settle": null, "baseId": "ARB", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 2198003.07715277}}, "created": null, "info": {"symbol": "ARBUSD", "status": "BREAK", "baseAsset": "ARB", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "2198003.07715277", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "FLOKIUSDT", "lowercaseId": "flokiusdt", "symbol": "FLOKI/USDT", "base": "FLOKI", "quote": "USDT", "settle": null, "baseId": "FLOKI", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-08, "cost": null, "base": 0.01, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 46116860414.0}, "price": {"min": 1e-08, "max": 1.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 700230642.95}}, "created": null, "info": {"symbol": "FLOKIUSDT", "status": "TRADING", "baseAsset": "FLOKI", "baseAssetPrecision": "2", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "2", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000001", "maxPrice": "1.00000000", "tickSize": "0.00000001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00", "maxQty": "46116860414.00", "stepSize": "1.00"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00", "maxQty": "700230642.95", "stepSize": "0.00"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "FLOKIUSD", "lowercaseId": "flokiusd", "symbol": "FLOKI/USD", "base": "FLOKI", "quote": "USD", "settle": null, "baseId": "FLOKI", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-08, "cost": null, "base": 0.01, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 46116860414.0}, "price": {"min": 1e-08, "max": 1.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 46116860414.0}}, "created": null, "info": {"symbol": "FLOKIUSD", "status": "BREAK", "baseAsset": "FLOKI", "baseAssetPrecision": "2", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "2", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000001", "maxPrice": "1.00000000", "tickSize": "0.00000001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00", "maxQty": "46116860414.00", "stepSize": "1.00"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00", "maxQty": "46116860414.00", "stepSize": "0.00"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "XECUSDT", "lowercaseId": "xecusdt", "symbol": "XEC/USDT", "base": "XEC", "quote": "USDT", "settle": null, "baseId": "XEC", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-08, "cost": null, "base": 0.01, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 46116860414.0}, "price": {"min": 1e-08, "max": 1.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 297146265.97}}, "created": null, "info": {"symbol": "XECUSDT", "status": "TRADING", "baseAsset": "XEC", "baseAssetPrecision": "2", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "2", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000001", "maxPrice": "1.00000000", "tickSize": "0.00000001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00", "maxQty": "46116860414.00", "stepSize": "1.00"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00", "maxQty": "297146265.97", "stepSize": "0.00"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "XECUSD", "lowercaseId": "xecusd", "symbol": "XEC/USD", "base": "XEC", "quote": "USD", "settle": null, "baseId": "XEC", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-08, "cost": null, "base": 0.01, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 46116860414.0}, "price": {"min": 1e-08, "max": 1.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 1262004788.3}}, "created": null, "info": {"symbol": "XECUSD", "status": "BREAK", "baseAsset": "XEC", "baseAssetPrecision": "2", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "2", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000001", "maxPrice": "1.00000000", "tickSize": "0.00000001"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00", "maxQty": "46116860414.00", "stepSize": "1.00"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00", "maxQty": "1262004788.30", "stepSize": "0.00"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BLURUSDT", "lowercaseId": "blurusdt", "symbol": "BLUR/USDT", "base": "BLUR", "quote": "USDT", "settle": null, "baseId": "BLUR", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 14613.76625}}, "created": null, "info": {"symbol": "BLURUSDT", "status": "TRADING", "baseAsset": "BLUR", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "14613.76625000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "BLURUSD", "lowercaseId": "blurusd", "symbol": "BLUR/USD", "base": "BLUR", "quote": "USD", "settle": null, "baseId": "BLUR", "quoteId": "USD", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 423028.58638888}}, "created": null, "info": {"symbol": "BLURUSD", "status": "BREAK", "baseAsset": "BLUR", "baseAssetPrecision": "8", "quoteAsset": "USD", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "423028.58638888", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ANKRUSDT", "lowercaseId": "ankrusdt", "symbol": "ANKR/USDT", "base": "ANKR", "quote": "USDT", "settle": null, "baseId": "ANKR", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 212566.41958333}}, "created": null, "info": {"symbol": "ANKRUSDT", "status": "TRADING", "baseAsset": "ANKR", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "212566.41958333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "DAIUSDT", "lowercaseId": "daiusdt", "symbol": "DAI/USDT", "base": "DAI", "quote": "USDT", "settle": null, "baseId": "DAI", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 804.20333333}}, "created": null, "info": {"symbol": "DAIUSDT", "status": "TRADING", "baseAsset": "DAI", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "900000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "804.20333333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.05", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.95", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "DASHUSDT", "lowercaseId": "dashusdt", "symbol": "DASH/USDT", "base": "DASH", "quote": "USDT", "settle": null, "baseId": "DASH", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.001, "price": 0.01, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.001, "max": 90000.0}, "price": {"min": 0.01, "max": 100000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 210.846975}}, "created": null, "info": {"symbol": "DASHUSDT", "status": "TRADING", "baseAsset": "DASH", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.01000000", "maxPrice": "100000.00000000", "tickSize": "0.01000000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.00100000", "maxQty": "90000.00000000", "stepSize": "0.00100000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "210.84697500", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "HBARUSDT", "lowercaseId": "hbarusdt", "symbol": "HBAR/USDT", "base": "HBAR", "quote": "USDT", "settle": null, "baseId": "HBAR", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 582473.65416666}}, "created": null, "info": {"symbol": "HBARUSDT", "status": "TRADING", "baseAsset": "HBAR", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "582473.65416666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ICXUSDT", "lowercaseId": "icxusdt", "symbol": "ICX/USDT", "base": "ICX", "quote": "USDT", "settle": null, "baseId": "ICX", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 13971.83416666}}, "created": null, "info": {"symbol": "ICXUSDT", "status": "TRADING", "baseAsset": "ICX", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "13971.83416666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "IOTAUSDT", "lowercaseId": "iotausdt", "symbol": "IOTA/USDT", "base": "IOTA", "quote": "USDT", "settle": null, "baseId": "IOTA", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 57053.74583333}}, "created": null, "info": {"symbol": "IOTAUSDT", "status": "TRADING", "baseAsset": "IOTA", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "900000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "57053.74583333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "RVNUSDT", "lowercaseId": "rvnusdt", "symbol": "RVN/USDT", "base": "RVN", "quote": "USDT", "settle": null, "baseId": "RVN", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 258880.42916666}}, "created": null, "info": {"symbol": "RVNUSDT", "status": "TRADING", "baseAsset": "RVN", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "258880.42916666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "WAVESUSDT", "lowercaseId": "wavesusdt", "symbol": "WAVES/USDT", "base": "WAVES", "quote": "USDT", "settle": null, "baseId": "WAVES", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 90000.0}, "price": {"min": 0.001, "max": 10000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 1623.56208333}}, "created": null, "info": {"symbol": "WAVESUSDT", "status": "BREAK", "baseAsset": "WAVES", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "10000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "90000.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "1623.56208333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "XNOUSDT", "lowercaseId": "xnousdt", "symbol": "XNO/USDT", "base": "XNO", "quote": "USDT", "settle": null, "baseId": "XNO", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 2077.32364016}}, "created": null, "info": {"symbol": "XNOUSDT", "status": "TRADING", "baseAsset": "XNO", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "2077.32364016", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "XTZUSDT", "lowercaseId": "xtzusdt", "symbol": "XTZ/USDT", "base": "XTZ", "quote": "USDT", "settle": null, "baseId": "XTZ", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 900000.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 5064.14666666}}, "created": null, "info": {"symbol": "XTZUSDT", "status": "TRADING", "baseAsset": "XTZ", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "900000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "5064.14666666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ZILUSDT", "lowercaseId": "zilusdt", "symbol": "ZIL/USDT", "base": "ZIL", "quote": "USDT", "settle": null, "baseId": "ZIL", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 9000000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 311844.83583333}}, "created": null, "info": {"symbol": "ZILUSDT", "status": "TRADING", "baseAsset": "ZIL", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "9000000.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "311844.83583333", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ORBSUSDT", "lowercaseId": "orbsusdt", "symbol": "ORBS/USDT", "base": "ORBS", "quote": "USDT", "settle": null, "baseId": "ORBS", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 9000000.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 61940.4125}}, "created": null, "info": {"symbol": "ORBSUSDT", "status": "TRADING", "baseAsset": "ORBS", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "9000000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "61940.41250000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "CUDOSUSDT", "lowercaseId": "cudosusdt", "symbol": "CUDOS/USDT", "base": "CUDOS", "quote": "USDT", "settle": null, "baseId": "CUDOS", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": false, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 1e-06, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 1e-06, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 284063.77666666}}, "created": null, "info": {"symbol": "CUDOSUSDT", "status": "BREAK", "baseAsset": "CUDOS", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00000100", "maxPrice": "1000.00000000", "tickSize": "0.00000100"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "284063.77666666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ADXUSDT", "lowercaseId": "adxusdt", "symbol": "ADX/USDT", "base": "ADX", "quote": "USDT", "settle": null, "baseId": "ADX", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 4154.31666666}}, "created": null, "info": {"symbol": "ADXUSDT", "status": "TRADING", "baseAsset": "ADX", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "4154.31666666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "FORTUSDT", "lowercaseId": "fortusdt", "symbol": "FORT/USDT", "base": "FORT", "quote": "USDT", "settle": null, "baseId": "FORT", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 17927.925}}, "created": null, "info": {"symbol": "FORTUSDT", "status": "TRADING", "baseAsset": "FORT", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "900000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "17927.92500000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "SUIUSDT", "lowercaseId": "suiusdt", "symbol": "SUI/USDT", "base": "SUI", "quote": "USDT", "settle": null, "baseId": "SUI", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.1, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.1, "max": 92141578.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 44040.20041666}}, "created": null, "info": {"symbol": "SUIUSDT", "status": "TRADING", "baseAsset": "SUI", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.10000000", "maxQty": "92141578.00000000", "stepSize": "0.10000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "44040.20041666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "ONGUSDT", "lowercaseId": "ongusdt", "symbol": "ONG/USDT", "base": "ONG", "quote": "USDT", "settle": null, "baseId": "ONG", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 0.0001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 900000.0}, "price": {"min": 0.0001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 6768.0}}, "created": null, "info": {"symbol": "ONGUSDT", "status": "TRADING", "baseAsset": "ONG", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00010000", "maxPrice": "1000.00000000", "tickSize": "0.00010000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "900000.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "6768.00000000", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "GUSDT", "lowercaseId": "gusdt", "symbol": "G/USDT", "base": "G", "quote": "USDT", "settle": null, "baseId": "G", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 1.0, "price": 1e-05, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 1.0, "max": 92141578.0}, "price": {"min": 1e-05, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 190198.16666666}}, "created": null, "info": {"symbol": "GUSDT", "status": "TRADING", "baseAsset": "G", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00001000", "maxPrice": "1000.00000000", "tickSize": "0.00001000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "1.00000000", "maxQty": "92141578.00000000", "stepSize": "1.00000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "190198.16666666", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}, {"id": "RENDERUSDT", "lowercaseId": "renderusdt", "symbol": "RENDER/USDT", "base": "RENDER", "quote": "USDT", "settle": null, "baseId": "RENDER", "quoteId": "USDT", "settleId": null, "type": "spot", "spot": true, "margin": false, "swap": false, "future": false, "option": false, "index": null, "active": true, "contract": false, "linear": null, "inverse": null, "subType": null, "taker": 0.001, "maker": 0.001, "contractSize": null, "expiry": null, "expiryDatetime": null, "strike": null, "optionType": null, "precision": {"amount": 0.01, "price": 0.001, "cost": null, "base": 1e-08, "quote": 1e-08}, "limits": {"leverage": {"min": null, "max": null}, "amount": {"min": 0.01, "max": 92141578.0}, "price": {"min": 0.001, "max": 1000.0}, "cost": {"min": 1.0, "max": null}, "market": {"min": 0.0, "max": 4733.64229166}}, "created": null, "info": {"symbol": "RENDERUSDT", "status": "TRADING", "baseAsset": "RENDER", "baseAssetPrecision": "8", "quoteAsset": "USDT", "quotePrecision": "8", "quoteAssetPrecision": "8", "baseCommissionPrecision": "8", "quoteCommissionPrecision": "8", "orderTypes": ["LIMIT", "LIMIT_MAKER", "MARKET", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"], "icebergAllowed": true, "ocoAllowed": true, "quoteOrderQtyMarketAllowed": true, "allowTrailingStop": true, "cancelReplaceAllowed": true, "isSpotTradingAllowed": true, "isMarginTradingAllowed": false, "filters": [{"filterType": "PRICE_FILTER", "minPrice": "0.00100000", "maxPrice": "1000.00000000", "tickSize": "0.00100000"}, {"filterType": "PERCENT_PRICE", "multiplierUp": "5", "multiplierDown": "0.2", "avgPriceMins": "5"}, {"filterType": "LOT_SIZE", "minQty": "0.01000000", "maxQty": "92141578.00000000", "stepSize": "0.01000000"}, {"filterType": "MIN_NOTIONAL", "minNotional": "1.00000000", "applyToMarket": true, "avgPriceMins": "5"}, {"filterType": "ICEBERG_PARTS", "limit": "10"}, {"filterType": "MARKET_LOT_SIZE", "minQty": "0.00000000", "maxQty": "4733.64229166", "stepSize": "0.00000000"}, {"filterType": "TRAILING_DELTA", "minTrailingAboveDelta": "10", "maxTrailingAboveDelta": "2000", "minTrailingBelowDelta": "10", "maxTrailingBelowDelta": "2000"}, {"filterType": "PERCENT_PRICE_BY_SIDE", "bidMultiplierUp": "1.25", "bidMultiplierDown": "0.2", "askMultiplierUp": "5", "askMultiplierDown": "0.75", "avgPriceMins": "5"}, {"filterType": "MAX_NUM_ORDERS", "maxNumOrders": "200"}, {"filterType": "MAX_NUM_ALGO_ORDERS", "maxNumAlgoOrders": "5"}], "permissions": ["SPOT"], "defaultSelfTradePreventionMode": "EXPIRE_MAKER", "allowedSelfTradePreventionModes": ["EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"]}, "tierBased": true, "percentage": true, "feeSide": "get"}]'),
}
